/**
	\file
	\brief Functions for matching photometry files.

	Set of functions defined in this module allows user to 
	find matching files, i.e. to find crossreferences in two sets
	of stars by their positions.
	
	\author David Motl <dmotl@volny.cz>
	
	\par Copying
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation, version 2.
	
	$Id: cmpack_match.h,v 1.3 2013/01/12 20:45:40 dmotl Exp $
*/
#ifndef _CMPACK_MATCH_H_INCLUDED
#define _CMPACK_MATCH_H_INCLUDED

#include "cmpack_console.h"
#include "cmpack_phtfile.h"
#include "cmpack_catfile.h"

/********************   Data type definitions   ********************************/

/**
	\brief Matching configuration context
	\details This private data structure holds the configuration parameters for 
	CCD-frame conversion.
*/
typedef struct _CmpackMatch CmpackMatch;

/**
	\brief Matching method
*/
typedef enum _CmpackMatchMethod 
{
	CMPACK_MATCH_STANDARD,
	CMPACK_MATCH_AUTO,
	CMPACK_MATCH_SPARSE_FIELDS
} CmpackMatchMethod;

/************************   Public functions   ********************************/

#ifdef __cplusplus
extern "C" {
#endif

/**
	\brief Make new matching context
	\details The function allocates memory with matching context and returns a 
	new reference to it. The reference counter is set to one. The caller is 
	responsible to call cmpack_match_destroy() when it is no longer needed.
	\return pointer to a new reference
*/
CmpackMatch *cmpack_match_init(void);

/**
	\brief Make a new reference to the matching context
	\details The function makes a new reference to the context and returns a 
	pointer to it. The reference counter is incremented by one. The caller 
	is responsible to call cmpack_match_destroy() when the reference is 
	no longer needed.
	\return pointer to a new reference
*/
CmpackMatch *cmpack_match_reference(CmpackMatch *ctx);

/**
	\brief Release a reference to the matching context
	\details The function releases a reference to the context. 
	The reference counter is decreased by one and when it was the 
	last reference to the context, the context is freed and all memory 
	allocated in the context is reclaimed.
*/
void cmpack_match_destroy(CmpackMatch *ctx);

/**
	\brief Attach console to the matching context
	\details The function connects a matching context with
	a console context. The console is designed to print the information
	during the data processing. The functions makes its own reference
	to the console. Only one console can be attached to a single context, 
	if another console is attached to the single context, the original 
	one is released. Set console to NULL to release a reference to the 
	console that is currently attached to the context.
	\param[in] ctx			matching computation context
	\param[in] con			console context
*/
void cmpack_match_set_console(CmpackMatch *ctx, CmpackConsole *con);

/**
	\brief Set max. number of stars used for matching
	\param[in] ctx			matching context
	\param[in] nstars		number of stars
*/
void cmpack_match_set_maxstars(CmpackMatch *ctx, int nstars);

/**
	\brief Get max. number of stars used for matching
	\param[in] ctx			matching context
	\return number of stars
*/
int cmpack_match_get_maxstars(CmpackMatch *ctx);

/**
	\brief Set number of vertices of polygons
	\param[in] ctx			matching context
	\param[in] vertices		number of vertices
*/
void cmpack_match_set_vertices(CmpackMatch *ctx, int vertices);

/**
	\brief Get number of vertices of polygons
	\param[in] ctx			matching context
	\return number of vertices
*/
int cmpack_match_get_vertices(CmpackMatch *ctx);

/**
	\brief Set clipping threshold
	\param[in] ctx			matching context
	\param[in] threshold	threshold 
*/
void cmpack_match_set_threshold(CmpackMatch *ctx, double threshold);

/**
	\brief Get clipping threshold
	\param[in] ctx			matching context
	\return	current clipping threshold 
*/
double cmpack_match_get_threshold(CmpackMatch *ctx);

/**
	\brief Set matching method
	\param[in] ctx			matching context
	\param[in] method		(0=Standard, 1=Auto, 2=Sparse_fields)
*/
void cmpack_match_set_method(CmpackMatch *ctx, CmpackMatchMethod method);

/**
	\brief Get matching method
	\param[in] ctx			matching context
	\return (0=Standard, 1=Auto, 2=Sparse_fields)
*/
CmpackMatchMethod cmpack_match_get_method(CmpackMatch *ctx);

/**
	\brief Set Max. offset for 'sparse fields' method
	\param[in] ctx			matching context
	\param[in] maxoffset	offset in pixels
*/
void cmpack_match_set_maxoffset(CmpackMatch *ctx, double maxoffset);

/**
	\brief Get Max. offset for 'sparse fields' method
	\param[in] ctx			matching context
	\return offset in pixels
*/
double cmpack_match_get_maxoffset(CmpackMatch *ctx);

/**
	\brief Load reference frame from the photometry file
	\details The function reads photometry file into the context.
	\param[in] ctx			matching context
	\param[in] reffile		reference file context
	\return zero on success or error code on failure
*/
int cmpack_match_readref_pht(CmpackMatch *ctx, CmpackPhtFile *reffile);

/**
	\brief Load reference frame from the catalog file
	\details The function reads catalog file into the context.
	\param[in] ctx			matching context
	\param[in] reffile		reference file context
	\return zero on success or error code on failure
*/
int cmpack_match_readref_cat(CmpackMatch *ctx, CmpackCatFile *reffile);

/**
	\brief Match photometry file to the reference file
	\details The function reads the intput file, finds crossreferences to the
	reference frame and set the reference IDs.
	\param[in] ctx			matching context
	\param[in] file			photometry file context
	\param[out] mstars		number of stars matched
	\return zero on success or error code on failure
*/
int cmpack_match(CmpackMatch *ctx, CmpackPhtFile *file, int *mstars);

/**
	\brief Get offset of the last frame
	\details The function retrieves the offset of the last frame
	\param[in] ctx			matching context
	\param[out] offset_x	offset in X axis
	\param[out] offset_y	offset in Y axis
	\return zero on success or error code on failure
*/
int cmpack_match_get_offset(CmpackMatch *ctx, double *offset_x, double *offset_y);

#ifdef __cplusplus
}
#endif

#endif
