/**
	\file
	\brief Functions for the making mag-dev graph

	Set of functions defined in this module allows user to 
	make a mag-dev graph, that is used to find variable stars
	in a viewfield.
	
	\author David Motl <dmotl@volny.cz>
	
	\par Copying
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation, version 2.
	
	$Id: cmpack_mfind.h,v 1.1.1.1 2012/08/12 16:59:17 dmotl Exp $
*/
#ifndef _CMPACK_MFIND_H_INCLUDED
#define _CMPACK_MFIND_H_INCLUDED

#include "cmpack_console.h"
#include "cmpack_table.h"
#include "cmpack_fset.h"

/********************   Private data structures   ********************************/

/**
	\brief Munifind context
	\details This private data structure holds information needed to build
	the mag-dev graph.
*/
typedef struct _CmpackMuniFind CmpackMuniFind;

/**********************   Public constants   ***********************************/

#define CMPACK_MFIND_AUTO	(-1)		/**< Automatically select comparison star */

/**
	\brief Output modifiers
*/
typedef enum _CmpackMFindFlags
{
	CMPACK_MFIND_DEFAULT	= 0		/**< Default behavior */
} CmpackMFindFlags;

/********************   Public functions   ********************************/

#ifdef __cplusplus
extern "C" {
#endif

/**
	\brief Make new Munifind context
	\details The function allocates memory with Munifind context and returns a 
	new reference to it. The reference counter is set to one. The caller is 
	responsible to call cmpack_konv_destroy() when it is no longer needed.
	\return pointer to a new reference
*/
CmpackMuniFind *cmpack_mfind_init(void);

/**
	\brief Make a new reference to the Munifind context
	\details The function makes a new reference to the context and returns a 
	pointer to it. The reference counter is incremented by one. The caller 
	is responsible to call cmpack_konv_destroy() when the reference is 
	no longer needed.
	\return pointer to a new reference
*/
CmpackMuniFind *cmpack_mfind_reference(CmpackMuniFind *ctx);

/**
	\brief Release a reference to the Munifind context
	\details The function releases a reference to the context. 
	The reference counter is decreased by one and when it was the 
	last reference to the context, the context is freed and all memory 
	allocated in the context is reclaimed.
*/
void cmpack_mfind_destroy(CmpackMuniFind *ctx);

/**
	\brief Attach console to the context
	\details The function connects a Munifind context with
	a console context. The console is designed to print the information
	during the data processing. The functions makes its own reference
	to the console. Only one console can be attached to a single context, 
	if another console is attached to the single context, the original 
	one is released. Set console to NULL to release a reference to the 
	console that is currently attached to the context.
	\param[in] ctx			munifind context
	\param[in] con			console context
*/
void cmpack_mfind_set_console(CmpackMuniFind *ctx, CmpackConsole *con);

/**
	\brief Set the aperture identifier
	\details The function sets the identifier of the aperture which
	is used to read the measurements. The default value is zero.
	\param[in] ctx			munifind context
	\param[in] aperture		aperture aperture
*/
void cmpack_mfind_set_aperture(CmpackMuniFind *ctx, int aperture);

/**
	\brief Set the comparison star 
	\details The function sets the comparison star which is used to compute
	the brightness of the objects. If this parameter is set to CMPACK_MFIND_AUTO, 
	the program will choose a suitable comparison star (default). When the cmpack_mfind 
	function is finished, you can tell the used comparison star by means of the 
	cmpack_mfind_get_comparison() function. If you set this parameter to positive 
	value, it must be a valid star identifier and this star will be used as a comparison 
	star. 
	\param[in] ctx			munifind context
	\param[in] star_id		comparison star identifier or CMPACK_MFIND_AUTO
*/
void cmpack_mfind_set_comparison(CmpackMuniFind *ctx, int star_id);

/**
	\brief Get comparison star used 
	\details If the autodetection mode is in effect (see the cmpack_mfind_set_comparison()
	function), this function returns the identifier or the comparison star which was 
	actually used during computation.
	\param[in] ctx			munifind context
	\return identifier of the comparison star or negative value on error
*/
int cmpack_mfind_get_comparison(CmpackMuniFind *ctx);

/**
	\brief Set clipping threshold
	\details The function sets the clipping threshold for the
	Munifind tool.
	\param[in] ctx			munifind context
	\param[in] threshold	threshold in %
*/
void cmpack_mfind_set_threshold(CmpackMuniFind *ctx, double threshold);

/**
	\brief Make mag-dev graph from frame set
	\details The result is stored to a given table
	\param[in] ctx			munifind context
	\param[in] fset			frameset context
	\param[out] table		table with results
	\param[in] flags		flags
	\return zero on success or error code on failure.
*/
int cmpack_mfind(CmpackMuniFind *ctx, CmpackFrameSet *fset, 
	CmpackTable **table, CmpackMFindFlags flags);

#ifdef __cplusplus
}
#endif

#endif
