/**************************************************************

amass.c (C-Munipack project)
Air-mass coefficient computation
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_amass.h"
#include "trajd.h"

/**********************   PUBLIC FUNCTIONS   ***************************/

/* Compute airmass coefficient X. ra is in hours; dec, lon, lat in degrees */
int cmpack_airmass(double jd, double ra, double dec, double lon, double lat, 
	double *airmass, double *alt)
{
	double x = -1, lmst, har, h;

	/* Local mean sidereal time */
	lmst = cmpack_siderealtime(jd) + (lon/15.0);
	/* Hour angle */
	har = (lmst - ra)/12.0*M_PI;
	/* aparent altitude in degrees above horizon */
	lat = lat/180.0*M_PI;
	dec = dec/180.0*M_PI;
	h = asin(sin(dec)*sin(lat) + cos(dec)*cos(har)*cos(lat))/M_PI*180.0;
	if (alt)
		*alt = h;

	/* Is the star above the horizon? */
	if (h>=0) {
		/* Pickering's formula */
		x = 1.0 / sin((h/180.0*M_PI + 244/(165+47*pow(h, 1.1))/180.0*M_PI));
		if (airmass)
			*airmass = x;
	} else {
		if (airmass)
			*airmass = -1.0;
	}
	return CMPACK_ERR_OK;
}

/* Append air mass coefficient to the specified file */
int cmpack_airmass_table(CmpackTable *tab, const char *objname, double ra, 
	double dec, const char *location, double lon, double lat, CmpackConsole *con, 
	unsigned flags)
{
	int		res, no_airmass, no_altitude, jd_column, x_column, a_column;
	char	msg[MAXLINE];
	double	jd, x, alt;

	no_airmass	= (flags & CMPACK_AMASS_NOAIRMASS)!=0;
	no_altitude	= (flags & CMPACK_AMASS_NOALTITUDE)!=0;

	/* Print parameters */
	if (is_debug(con)) {
		printout(con, 1, "Configuration parameters:");
		printpard(con, "R.A.", 1, ra, 3);
		printpard(con, "Dec.", 1, dec, 3);
		printpard(con, "Lon.", 1, lon, 3);
		printpard(con, "Lat.", 1, lat, 3);
  	}

	jd_column = cmpack_tab_find_column(tab, "JD");
	if (jd_column<0) 
		jd_column = cmpack_tab_find_column(tab, "JDGEO");
	if (jd_column<0) {
		printout(con, 0, "Missing column with Julian date");
		return CMPACK_ERR_KEY_NOT_FOUND;
	}

	x_column = -1;
	if (!no_airmass) {
		x_column = cmpack_tab_find_column(tab, "AIRMASS");
		if (x_column<0)
			x_column = cmpack_tab_add_column_dbl(tab, "AIRMASS", AMASS_PRECISION, 0.0, 1e99, -1.0);
	}

	a_column = -1;
	if (!no_altitude) {
		a_column = cmpack_tab_find_column(tab, "ALTITUDE");
		if (a_column<0)
			a_column = cmpack_tab_add_column_dbl(tab, "ALTITUDE", ALT_PRECISION, -90.0, 90.0, 99.0);
	}

	if (objname)
		cmpack_tab_pkys(tab, "OBJECT", objname);
	else
		cmpack_tab_dkey(tab, "OBJECT");
	cmpack_ratostr(ra, msg, 256);
	cmpack_tab_pkys(tab, "RA", msg);
	cmpack_dectostr(dec, msg, 256);
	cmpack_tab_pkys(tab, "DEC", msg);
	
	if (location)
		cmpack_tab_pkys(tab, "LOCATION", location);
	else
		cmpack_tab_dkey(tab, "LOCATION");
	cmpack_lontostr(lon, msg, 256);
	cmpack_tab_pkys(tab, "LONGITUDE", msg);
	cmpack_lattostr(lat, msg, 256);
	cmpack_tab_pkys(tab, "LATITUDE", msg);
	
	res = cmpack_tab_rewind(tab);
	while (res==0) {
		cmpack_tab_gtdd(tab, jd_column, &jd);
		if (jd<=0) {
			printout(con, 0, "Invalid Julian date of observation");
			return CMPACK_ERR_INVALID_JULDAT;
		}
		if (cmpack_airmass(jd, ra, dec, lon, lat, &x, &alt)!=0) {
			x = -1.0;
			alt = -99.9;
		}
		if (is_debug(con)) {
			sprintf(msg, "%.5f -> Alt. = %.3f, X = %.3f", jd, alt, x); 
			printout(con, 1, msg);
		}
		if (x_column>=0)
			cmpack_tab_ptdd(tab, x_column, x);
		if (a_column>=0)
			cmpack_tab_ptdd(tab, a_column, alt);
		res = cmpack_tab_next(tab);
	}

	/* Normal return */
	return 0;
}


/* Open output table, start reading sequence */
int cmpack_airmass_curve(CmpackFrameSet *fset, CmpackTable **ptable, const char *objname, 
	double ra, double declination, const char *location, double longitude, double latitude,
	CmpackAMassFlags flags, CmpackConsole *con)
{	
	int		res, frameids, no_airmass, no_altitude;
	int		id_column, jd_column, a_column, x_column;
	char	msg[MAXLINE];
	CmpackTable *table;
	CmpackFrameInfo frame;

	*ptable = NULL;
	
	frameids    = (flags & CMPACK_AMASS_FRAME_IDS)!=0;
	no_airmass	= (flags & CMPACK_AMASS_NOAIRMASS)!=0;
	no_altitude	= (flags & CMPACK_AMASS_NOALTITUDE)!=0;

	/* Create table */
	table = cmpack_tab_init(CMPACK_TABLE_AIRMASS);

	id_column = x_column = a_column = -1;
	if (frameids)
		id_column = cmpack_tab_add_column_int(table, "FRAME", 0, INT_MAX, -1);
	jd_column = cmpack_tab_add_column_dbl(table, "JD", JD_PRECISION, 1e6, 1e99, 0.0);
	if (!no_airmass)
		x_column  = cmpack_tab_add_column_dbl(table, "AIRMASS", JD_PRECISION, 0.0, 1e99, -1.0);
	if (!no_altitude)
		a_column  = cmpack_tab_add_column_dbl(table, "ALTITUDE", ALT_PRECISION, -90.0, 90.0, 99.0);

	/* Process frames */
	res = cmpack_fset_rewind(fset);
	while (res==0) {
		cmpack_tab_append(table);
		cmpack_fset_get_frame(fset, CMPACK_FI_ID | CMPACK_FI_JULDAT | CMPACK_FI_HELCOR | CMPACK_FI_AIRMASS_ALT | CMPACK_FI_FILENAME, &frame);
		/* Check frame header */
		if (frameids && frame.frame_id<0) {
			if (frame.filename) 
				sprintf(msg, "%s: Invalid frame identifier", frame.filename);
			else
				strcpy(msg, "Invalid frame identifier");
			printout(con, 0, msg);
		}
		if (frame.juldat<=0) {
			if (frame.filename) 
				sprintf(msg, "%s: Invalid Julian date of observation", frame.filename);
			else
				strcpy(msg, "Invalid Julian date of observation");
			printout(con, 0, msg);
		}
		res = cmpack_airmass(frame.juldat, ra, declination, longitude, latitude, 
			&frame.airmass, &frame.altitude);
		if (res!=0)
			return res;

		if (id_column>=0)
			cmpack_tab_ptdi(table, id_column, frame.frame_id);
		if (jd_column>=0)
			cmpack_tab_ptdd(table, jd_column, frame.juldat);
		if (a_column>=0)
			cmpack_tab_ptdd(table, a_column, frame.altitude);
		if (x_column>=0)
			cmpack_tab_ptdd(table, x_column, frame.airmass);
		/* Debug printout */
		if (is_debug(con)) {
			if (frame.filename)
				sprintf(msg, "FRAME: %d (%s)", frame.frame_id, frame.filename);
			else
				sprintf(msg, "FRAME: %d", frame.frame_id);
			printout(con, 1, msg);
			sprintf(msg, "JD: %.7f", frame.juldat); 
			printout(con, 1, msg);
			sprintf(msg, "AIR MASS: %.7f", frame.airmass); 
			printout(con, 1, msg);
			sprintf(msg, "ALTITUDE: %.7f", frame.altitude); 
			printout(con, 1, msg);
		}
		res = cmpack_fset_next(fset);
	}

	*ptable = table;
	return CMPACK_ERR_OK;
}

/* Append air mass coefficient to the specified file */
int cmpack_airmass_fset(CmpackFrameSet *fset, const char *objname, double ra, 
	double dec, const char *location, double lon, double lat, CmpackConsole *con)
{
	int		res;
	char	msg[MAXLINE];
	double	x, alt;
	CmpackFrameInfo info;
	CmpackFrameSetInfo data;

	/* Print parameters */
	if (is_debug(con)) {
		printout(con, 1, "Configuration parameters:");
		printpard(con, "R.A.", 1, ra, 3);
		printpard(con, "Dec.", 1, dec, 3);
		printpard(con, "Lon.", 1, lon, 3);
		printpard(con, "Lat.", 1, lat, 3);
  	}

	data.objcoords.ra_valid = 1;
	data.objcoords.ra = ra;
	data.objcoords.dec_valid = 1;
	data.objcoords.dec = dec;
	data.objcoords.designation = (char*)objname;
	data.location.lon_valid = 1;
	data.location.lat = lat;
	data.location.lat_valid = 1;
	data.location.lon = lon;
	data.location.designation = (char*)location;
	cmpack_fset_set_info(fset, CMPACK_FS_OBJECT | CMPACK_FS_LOCATION, &data);

	res = cmpack_fset_rewind(fset);
	while (res==0) {
		cmpack_fset_get_frame(fset, CMPACK_FI_JULDAT, &info);
		info.airmass = -1.0;
		info.altitude = -99.9;
		if (info.juldat<=0) {
			printout(con, 0, "Invalid Julian date of observation");
		} else {
			if (cmpack_airmass(info.juldat, ra, dec, lon, lat, &x, &alt)==0) {
				info.airmass = x;
				info.altitude = alt;
				if (is_debug(con)) {
					sprintf(msg, "%.5f -> Alt. = %.3f, X = %.3f", info.juldat, alt, x); 
					printout(con, 1, msg);
				}
			}
		}
		cmpack_fset_set_frame(fset, CMPACK_FI_AIRMASS_ALT, &info);
		res = cmpack_fset_next(fset);
	}

	/* Normal return */
	return 0;
}
