/**************************************************************

find.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <math.h>

#include "comfun.h"
#include "cmpack_common.h"
#include "phot.h"
#include "phfun.h"
#include "find.h"

#define D(x,y)    d[(x)+((y)*rowwidth)]
#define G(x,y)    g[(x)+((y)*nbox)]
#define H(x,y)    h[(x)+((y)*ncol)]
#define SKIP(x,y) skip[(x)+((y)*nbox)]

/* Find stars */
int Find(CmpackPhot *kc, CmpackPhotFrame *frame)
{
	int left, top, right, bottom, ncol, nrow, nhalf, rowwidth, nbox, pixels, i, j, jx, jy, n, failed;
	int lastro, lastcl;
	double *d;		/* image */
	double *g;		/* subarray with Gaussian curve */
	double *h;		/* central heights of the best fitting Gaussian function for each pixel */
	char *skip;		/* which pixels will be skipped */
	double radius, wt, p, sigsq, sumg, sumgsq, jsq, rsq, denom, sgop, relerr, sgd, sd, sg, sgsq;
	double skymod, readns, phpadu, hmin, datum, dat, height, sharpness, temp;
	double sumgd, sumd, sdgdx, sdgdxs, sddgdx, sgdgdx, dgdx, hx, skylvl, dx, xcen, hy, dy, ycen, roundness;
	CmpackPhotStar *rec;

	ClearStarList(frame);

	top = frame->border.top, left = frame->border.left;
	bottom = cmpack_image_height(frame->image) - frame->border.bottom - 1;
	right = cmpack_image_width(frame->image) - frame->border.right - 1;
	if (left>right || top>bottom) {
		printout(kc->con, 0, "The borders are too large. Nothing is left from the source frame.");
		return CMPACK_ERR_INVALID_PAR;
	}

	skymod = frame->skymod;
	if (frame->fravg<=0) {
		printout(kc->con, 0, "Number of frames averaged must be greater than zero");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (frame->frsum<=0) {
		printout(kc->con, 0, "Number of frames summed must be greater than zero");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Get the source image */
	d = (double*)cmpack_image_data(frame->image);
	rowwidth = cmpack_image_width(frame->image);
	
	/* SECTION 1 - PREPARE SUBARRAY WITH CIRCULAR GAUSSIAN FUNCTION */

	radius = fmax(2.001,0.637*kc->fwhm);
	nhalf  = (int)radius;		/* number of pixels between the central pixel (exclusive)
										and the edge of the box (inclusive) */
	nbox = 2*nhalf + 1;			/* length of the side of the subarray */
	sigsq  = (kc->fwhm/2.35482)*(kc->fwhm/2.35482);
	radius = radius*radius;
	sumg = 0.0;                  /* sum of the values of the Gaussian function */
	sumgsq = 0.0;                /* sum of the squares of the values of the Gaussian function */
	pixels = 0;                  /* number of pixels in the circular area */

	g = (double*) cmpack_malloc(nbox*nbox*sizeof(double));
	skip = (char *) cmpack_malloc(nbox*nbox*sizeof(char));
	for (j=0; j<nbox; j++) {
		jsq = (j - nhalf)*(j - nhalf);
		for (i=0; i<nbox; i++) {
			rsq = (i - nhalf)*(i - nhalf) + jsq;
			G(i,j) = exp(-0.5*rsq/sigsq);
			if (rsq <= radius) {
				SKIP(i,j) = 0;
				sumg += G(i,j);
				sumgsq += G(i,j)*G(i,j);
				pixels++;
			} else {
				SKIP(i,j) = 1;
			}
		}
	}
	denom = sumgsq - (sumg*sumg)/pixels;    /* denominator of the fraction defining h */
	sgop = sumg/pixels;                     /* contains [G]/n */

	relerr = sqrt(1.0/denom);
	readns = kc->readns*kc->readns*(1.0*frame->frsum/frame->fravg);
	phpadu = kc->gain*frame->fravg;
	hmin   = sqrt(readns + fmax(0.0,skymod)/phpadu);
	hmin   = 0.01*((int)(100.0*kc->thresh*relerr*hmin+0.5));
	readns = sqrt(readns);

	frame->relerr = relerr;
	frame->phpadu = phpadu;
	frame->noise  = readns;
	frame->thres  = hmin;

	/* SECTION 2 - Compute central heights of the best fitting Gaussian function */

	ncol = right-left+1, nrow = bottom-top+1;
	h = (double*) cmpack_malloc(ncol*nrow*sizeof(double));
	for (jy=0; jy<nrow; jy++) {
		for (jx=0; jx<ncol; jx++) {
			sgd = 0.0;
			sd  = 0.0;
			sgsq = sumgsq;
			sg  = sumg;
			p   = pixels;

			/* Compute the local brightness enhancement for the pixel [jx,jy] */
			for (j=0; j<nbox; j++) {
				int iy = jy - nhalf + j;
				for (i=0; i<nbox; i++) {
					int ix = jx - nhalf + i;
					if (!SKIP(i,j)) {
						wt = G(i,j);
						if (iy>=0 && iy<nrow && ix>=0 && ix<ncol) {
							dat = D(left+ix,top+iy);
							if (frame->datalo<dat && dat<frame->datahi) {
								sgd  += wt*dat;
								sd   += dat;
							} else {
								sgsq -= wt*wt;
								sg   -= wt;
								p--;
							}
						} else {
							sgsq -= wt*wt;
							sg   -= wt;
							p--;
						}
					}
				}
			}

			if (p>1) {
	           /* Compute central height of the best fitting Gaussian function */
	           if (p<pixels) {
		           sgsq = sgsq - (sg*sg)/p;
                   if (sgsq!=0.0) 
					   sgd = (sgd - sg*sd/p)/sgsq;
                   else 
					   sgd = 0.0;
               } else {
                   sgd = (sgd - sgop*sd)/denom;
               }
           } else {
               sgd = 0.0;
           }
           H(jx,jy) = sgd;   /* central height of the best fitting Gaussian function */
		}
	}
	SKIP(nhalf,nhalf) = 1;
	pixels--;

	/* SECTION 3 - Search for local maxima in the convolved brightness data */

	lastro = nrow - nhalf - 1;       /* last row */
	lastcl = ncol - nhalf - 1;       /* last column */
	for (jy=nhalf; jy<=lastro; jy++) {
		for (jx=nhalf; jx<=lastcl; jx++) {

			/* Test the height: must be above a threshold */
			height = H(jx,jy);
			if (height<hmin) continue;  /* Test failed */

			/* the vicinity of this pixel: must be above each pixel within a radius equal to 1.5 sigma */
			failed = 0;
			for (j=0; j<nbox; j++) {
				for (i=0; i<nbox; i++) {
			        if (!SKIP(i,j)) {
						if (height<H(jx-nhalf+i,jy-nhalf+j)) { 
							failed = 1; 
							break; 
						}
					}
				}
				if (failed) break;
			}
			if (failed) continue;  /* Test failed - try next pixel */

			/* Derive the shape indices */
			sharpness = 0.0;
			datum = D(left+jx,top+jy);
			if (frame->datalo<datum && datum<frame->datahi) {
				p = 0.0;
				for (j=0; j<nbox; j++) {
					temp = 0.0;
					for (i=0; i<nbox; i++) {
						if (!SKIP(i,j)) {
							dat = D(left+jx-nhalf+i, top+jy-nhalf+j);
							if (frame->datalo<dat && dat<frame->datahi) {
								temp += dat;
								p    += 1.0;
							}
						}
					}
                    sharpness += temp;
				}
				sharpness = (datum-sharpness/p)/height;
				if ((sharpness<kc->shrplo)||(sharpness>kc->shrphi)) continue;    /* Test failed - try next pixel */
			}

			sumgd = 0.0;                  /* sum of the Gaussian times the data */
			sumgsq = 0.0;                 /* sum of the Gaussian squared */
			sumg = 0.0;                   /* sum of the Gaussian */
			sumd = 0.0;                   /* sum of the data */
			sdgdx = 0.0;                  /* sum of d(Gaussian)/dx */
			sdgdxs = 0.0;                 /* sum of {d(Gaussian)/dx}^2 */
			sddgdx = 0.0;                 /* sum of data times d(Gaussian)/dx */
			sgdgdx = 0.0;                 /* sum of Gaussian times d(Gaussian)/dx */
			p = 0.0;
			n = 0;
			for (i=0; i<nbox; i++) {
				sg = 0.0;                  /* temporary sum of the Gaussian */
				sd = 0.0;                  /* temporary sum of the data */
				for (j=0; j<nbox; j++) {
					wt = nhalf - fabs(j - nhalf) + 1;
					dat = D(left+jx-nhalf+i, top+jy-nhalf+j);
					if (frame->datalo<dat && dat<frame->datahi) {
						sd += (dat - skymod)*wt;
						sg += G(i,j)*wt;
					}
				}
				if (sg > 0.0) {
					wt = nhalf - fabs(i - nhalf) + 1;
					sumgd  += wt*sg*sd;
					sumgsq += wt*sg*sg;
					sumg   += wt*sg;
					sumd   += wt*sd;
					dgdx    = sg*(nhalf - i);
					sdgdxs += wt*dgdx*dgdx;
					sdgdx  += wt*dgdx;
					sddgdx += wt*sd*dgdx;
					sgdgdx += wt*sg*dgdx;
					p      += wt;
					n++;
				}
			}

			if (n<=2) continue;   /* we need at least three points */
			hx = (sumgd - sumg*sumd/p)/(sumgsq - (sumg*sumg)/p);
			if (hx<=0.0) continue;
			skylvl = (sumd - hx*sumg)/p;
			dx = (sgdgdx - (sddgdx - sdgdx*(hx*sumg + skylvl*p)))/(hx*sdgdxs/sigsq);
			/* compute the first-order correction to the x-centroid of the star */
			xcen = jx + dx/(1.0 + fabs(dx));

			if (xcen < 0.5 || xcen > ncol - 0.5) {
				/* if the best estimate of the star's center falls outside the lc->image, reject it */
				continue; 
			}

			/* Compute the height of the y-marginal Gaussian distribution */
			sumgd  = 0.0;
			sumgsq = 0.0;
			sumg   = 0.0;
			sumd   = 0.0;
			sdgdx  = 0.0;
			sdgdxs = 0.0;
			sddgdx = 0.0;
			sgdgdx = 0.0;
			p      = 0.0;
			n      = 0;
			for (j=0; j<nbox; j++) {
				sg = sd = 0.0;
				for (i=0; i<nbox; i++) {
					wt = nhalf - fabs(i - nhalf) + 1;
					dat = D(left+jx-nhalf+i, top+jy-nhalf+j);
					if (frame->datalo<dat && dat<frame->datahi) {
						sd += (dat - skymod)*wt;
						sg += G(i,j)*wt;
					}
				}
				if (sg>0.0) {
					wt = nhalf - fabs(j - nhalf) + 1;
					sumgd += wt*sg*sd;
					sumgsq+= wt*sg*sg;
					sumg  += wt*sg;
					sumd  += wt*sd;
					dgdx   = sg*(nhalf - j);
					sdgdx += wt*dgdx;
					sdgdxs+= wt*dgdx*dgdx;
					sddgdx+= wt*sd*dgdx;
					sgdgdx+= wt*sg*dgdx;
					p     += wt;
					n++;
				}
			}

			if (n<=2) continue;   /* we need at least three points */
			hy = (sumgd - sumg*sumd/p)/(sumgsq - (sumg*sumg)/p);
			if (hy<=0.0) continue;
			skylvl = (sumd - hy*sumg)/p;
			dy = (sgdgdx - (sddgdx - sdgdx*(hy*sumg + skylvl*p)))/(hy*sdgdxs/sigsq);
			ycen = jy + dy/(1.0 + fabs(dy));

			if (ycen < 0.5 || ycen > nrow - 0.5) {
				/* if the best estimate of the star's center falls outside the lc->image, reject it */
				continue;    
			}

			/* Derive roundness of object */
			roundness = 2.0*(hx - hy)/(hx + hy);
			if ((roundness<kc->rndlo)||(roundness>kc->rndhi)) continue;   /* Fails roundness test */

			/* Prints it's parameters to the monitor and output file */
			AddStar(frame, &rec);
			rec->xmax      = left + jx;
			rec->ymax      = top + jy;
			rec->xcen      = left + xcen + 0.5;
			rec->ycen      = top + ycen + 0.5;
			rec->height    = datum;
			rec->sharpness = sharpness;
			rec->roundness = roundness;

			/* continue with the next pixel */
		}
	}

	/* Section 4 - frees alocated memory */

	cmpack_free(g);
	cmpack_free(h);
	cmpack_free(skip);
	return 0;
}
