/**************************************************************

list.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_lcurve.h"

/* Star attributes */
typedef struct _StarInfo
{
	int id;					/**< Star identifier */
	int index;				/**< Frameset's object index */
	int valid;				/**< Nonzero if magnitude and error are valid */
	double mag, err;		/**< Magnitude and error */
} StarInfo;

/* List of selected stars */
typedef struct _StarList
{
	StarInfo *list;			/**< List of star identifiers */
	int	count;				/**< Number of items */
} StarList;

/* List process context */
struct _CmpackLCurve
{
	int refcnt;				/**< Reference counter */
	CmpackConsole *con;		/**< Console */
	int aperture;			/**< Aperture identifier */
	StarList var;			/**< Measurements of variable stars */
	StarList comp;			/**< Measurements of comparison stars */
	StarList chk;			/**< Measurements of check stars */
};

/*********************   LOCAL FUNCTIONS   ****************************/

/* Set list of stars */
static void stars_set(StarList *tab, const int *list, int count)
{
	int i;

	if (tab->list) {
		cmpack_free(tab->list);
		tab->list = NULL;
		tab->count = 0;
	}
	if (list && count>0) {
		tab->list = (StarInfo*)cmpack_calloc(count, sizeof(StarInfo));
		for (i=0; i<count; i++)
			tab->list[i].id = list[i];
		tab->count = count;
	}
}

/* Get list of stars */
static void stars_get(StarList *tab, int **plist, int *pcount)
{
	int i, *list;

	if (tab->list) {
		list = cmpack_malloc(tab->count*sizeof(int));
		for (i=0; i<tab->count; i++)
			list[i] = tab->list[i].id;
		*plist = list;
		*pcount = tab->count;
	} else {
		*plist = NULL;
		*pcount = 0;
	}
}

static void star_get(StarInfo *info, CmpackFrameSet *fset, int ap_index)
{
	CmpackPhtData data;

	info->valid = 0;
	if (info->index>=0) {
		data.mag_valid = 0;
		cmpack_fset_get_data(fset, info->index, ap_index, &data);
		if (data.mag_valid) {
			info->valid = data.mag_valid;
			info->mag = data.magnitude;
			info->err = data.mag_error;
		}
	}
}

/* Process one frame */
static void diff(CmpackLCurve *lc, CmpackTable *table, int first_col, const StarInfo *var, 
	int var_count, const StarInfo *comp, int comp_count, const StarInfo *chk, int chk_count,
	const char *filename)
{
	char	msg[MAXLINE];
	int		i, n, j, col, var_ok, comp_ok;
	double	cerr, cmag;

	/* Check variable star */
	var_ok = var[0].valid;
	if (!var_ok) {
		if (filename)
			sprintf(msg, "%s: The variable star(s) have undefined value of mag. or std. err.", filename);
		else
			strcpy(msg, "The variable star(s) have undefined value of mag. or std. err.");
		printout(lc->con, 0, msg);
	}

	/* Comparison star */
	if (comp_count==1) {
		if (comp[0].valid) {
			cmag = comp[0].mag;
			cerr = comp[0].err;
			comp_ok = 1;
		} else {
			cerr = cmag = 0.0;
			comp_ok = 0;
		}
	} else {
		cmag = cerr = 0.0; 
		n = 0;
		for (i=0; i<comp_count; i++) {
			if (comp[i].valid) {
				cmag += pow(10.0, -0.4*comp[i].mag);
				cerr += comp[i].err;
				n++;
			}
		}
		if (n==comp_count) {
			cmag = -2.5*log10(cmag/n);
			cerr = (cerr/n)/sqrt((double)n);
			comp_ok = 1;
		} else {
			cerr = cmag = 0.0;
			comp_ok = 0;
		}
	}
	if (!comp_ok) {
		if (filename)
			sprintf(msg, "%s: The comparison star(s) have undefined value of mag. or std. err.", filename);
		else
			strcpy(msg, "The comparison star(s) have undefined value of mag. or std. err.");
		printout(lc->con, 0, msg);
	}
	if (is_debug(lc->con)) {
		sprintf(msg, "Comparison star: %.3f +- %.3f mag", cmag, cerr);
		printout(lc->con, 1, msg);
	}

	col = first_col;

	/* V-C */
	if (var_ok && comp_ok) {
		cmpack_tab_ptdd(table, col, var[0].mag - cmag);
		cmpack_tab_ptdd(table, col+1, sqrt(var[0].err*var[0].err + cerr*cerr));
	}
	col+=2;

	/* V-K1, V-K2, ... */
	for (j=0; j<lc->chk.count; j++) {
		if (var_ok && chk[j].valid) {
			cmpack_tab_ptdd(table, col, var[0].mag - chk[j].mag);
			cmpack_tab_ptdd(table, col+1, sqrt(var[0].err*var[0].err + chk[j].err*chk[j].err));
		}
		col+=2;
	}

	/* C1-K1, C1-K2, ... */
	for (i=0; i<comp_count; i++) {
		for (j=0; j<chk_count; j++) {
			if (comp[i].valid && chk[j].valid) {
				cmpack_tab_ptdd(table, col, comp[i].mag - chk[j].mag);
				cmpack_tab_ptdd(table, col+1, sqrt(comp[i].err*comp[i].err + chk[j].err*chk[j].err));
			}
			col+=2;
		}
	}

	/* K1-K2, ... */
	for (i=0; i<chk_count; i++) {
		for (j=i+1; j<chk_count; j++) {
			if (chk[i].valid && chk[j].valid) {
				cmpack_tab_ptdd(table, col, chk[i].mag - chk[j].mag);
				cmpack_tab_ptdd(table, col+1, sqrt(chk[i].err*chk[i].err + chk[j].err*chk[j].err));
			}
			col+=2;
		}
	}
}

/* Process single frame */
static void inst(CmpackLCurve *lc, CmpackTable *table, int first_col, const StarInfo *var, 
	int var_count, const StarInfo *comp, int comp_count, const StarInfo *chk, int chk_count)
{
	int i, col = first_col;

	/* Variables */
	for (i=0; i<var_count; i++) {
		if (var[i].valid) {
			cmpack_tab_ptdd(table, col, var[i].mag);
			cmpack_tab_ptdd(table, col+1, var[i].err);
		}
		col+=2;
	}

	/* Comparison stars */
	for (i=0; i<comp_count; i++) {
		if (comp[i].valid) {
			cmpack_tab_ptdd(table, col, comp[i].mag);
			cmpack_tab_ptdd(table, col+1, comp[i].err);
		}
		col+=2;
	}

	/* Check stars */
	for (i=0; i<chk_count; i++) { 
		if (chk[i].valid) {
			cmpack_tab_ptdd(table, col, chk[i].mag);
			cmpack_tab_ptdd(table, col+1, chk[i].err);
		}
		col+=2;
	}
}

/********************   PUBLIC FUNCTIONS   **************************/

/* Make new listing context */
CmpackLCurve *cmpack_lcurve_init(void)
{
	CmpackLCurve *pc = (CmpackLCurve*)cmpack_calloc(1, sizeof(CmpackLCurve));
	pc->refcnt = 1;
	return pc;
}

/* Increment the reference counter */
CmpackLCurve *cmpack_lcurve_reference(CmpackLCurve *ctx)
{
	ctx->refcnt++;
	return ctx;
}

/* Decrement reference counter / detroy the instance */
void cmpack_lcurve_destroy(CmpackLCurve *ctx)
{
	ctx->refcnt--;
	if (ctx->refcnt==0) {
		cmpack_free(ctx->var.list);
		cmpack_free(ctx->comp.list);
		cmpack_free(ctx->chk.list);
		if (ctx->con) {
			cmpack_con_destroy(ctx->con);
			ctx->con = NULL;
		}
		cmpack_free(ctx);
	}
}

/* Attach console */
void cmpack_lcurve_set_console(CmpackLCurve *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		if (ctx->con) 
			cmpack_con_destroy(ctx->con);
		ctx->con = con;
		if (ctx->con) 
			cmpack_con_reference(ctx->con);
	}
}

/* Set aperture index */
void cmpack_lcurve_set_aperture(CmpackLCurve *proc, int aperture)
{
	proc->aperture = aperture;
}

/* Get aperture index */
int cmpack_lcurve_get_aperture(CmpackLCurve *proc)
{
	return proc->aperture;
}

/* Set list of variable stars */
void cmpack_lcurve_set_var(CmpackLCurve *lc, const int *items, int nitems)
{
	stars_set(&lc->var, items, nitems);
}

/* Set list of variable stars */
void cmpack_lcurve_get_var(CmpackLCurve *lc, int **items, int *nitems)
{
	stars_get(&lc->var, items, nitems);
}

/* Set list of comparison stars */
void cmpack_lcurve_set_comp(CmpackLCurve *lc, const int *items, int nitems)
{
	stars_set(&lc->comp, items, nitems);
}

/* Set list of comparison stars */
void cmpack_lcurve_get_comp(CmpackLCurve *lc, int **items, int *nitems)
{
	stars_get(&lc->comp, items, nitems);
}

/* Set list of check stars */
void cmpack_lcurve_set_check(CmpackLCurve *lc, const int *items, int nitems)
{
	stars_set(&lc->chk, items, nitems);
}

/* Set list of check stars */
void cmpack_lcurve_get_check(CmpackLCurve *lc, int **items, int *nitems)
{
	stars_get(&lc->chk, items, nitems);
}

/* Open output table, start reading sequence */
int cmpack_lcurve(CmpackLCurve *lc, CmpackFrameSet *fset, CmpackTable **ptable, 
	CmpackLCurveFlags flags)
{	
	int res, i, j, k, col, inst_mag, ap_index, frameids;
	int airmass, altitude, helcorr, jdhel, hjdonly;
	int	id_column, vc_column, jd_column, hjd_column, hc_column, a_column, x_column;
	char buf[80], msg[MAXLINE];
	CmpackTable *table;
	CmpackPhtAperture aperture;
	CmpackFrameInfo frame;
	StarInfo *var, *comp, *chk;

	*ptable = NULL;
	
	inst_mag = (flags & CMPACK_LCURVE_INSTMAG)!=0;
	frameids = (flags & CMPACK_LCURVE_FRAME_IDS)!=0;
	airmass  = (flags & CMPACK_LCURVE_AIRMASS)!=0;
	altitude = (flags & CMPACK_LCURVE_ALTITUDE)!=0;
	helcorr  = (flags & CMPACK_LCURVE_HELCORR)!=0;
	jdhel    = (flags & CMPACK_LCURVE_HJD)!=0;
	hjdonly  = (flags & CMPACK_LCURVE_HJDONLY)!=0;

	/* Check aperture */
	ap_index = cmpack_fset_find_aperture(fset, lc->aperture);
	if (ap_index<0) {
		printout(lc->con, 0, "Invalid aperture identifier");
		return CMPACK_ERR_INVALID_PAR;
	}
	aperture.id = -1;
	cmpack_fset_get_aperture(fset, ap_index, CMPACK_PA_ID, &aperture);

	/* Check stars */
	if (!inst_mag) {
		if (lc->var.count!=1) {
			printout(lc->con, 0, "Exactly one variable star must be selected");
			return CMPACK_ERR_INVALID_PAR;
		}
		if (lc->comp.count<1) {
			printout(lc->con, 0, "At least one comparison star must be selected");
			return CMPACK_ERR_INVALID_PAR;
		}
	} else {
		int nstar = lc->var.count + lc->comp.count + lc->chk.count;
		if (nstar<1) {
			printout(lc->con, 0, "At least one star must be selected");
			return CMPACK_ERR_INVALID_PAR;
		}
	}
	var = lc->var.list;
	for (i=0; i<lc->var.count; i++)
		var[i].index = cmpack_fset_find_object(fset, var[i].id);
	comp = lc->comp.list;
	for (i=0; i<lc->comp.count; i++)
		comp[i].index = cmpack_fset_find_object(fset, comp[i].id);
	chk = lc->chk.list;
	for (i=0; i<lc->chk.count; i++)
		chk[i].index = cmpack_fset_find_object(fset, chk[i].id);

	/* Print parameters */
	if (is_debug(lc->con)) {
		sprintf(msg, "Aperture: %d", aperture.id);
		printout(lc->con, 1, msg);
	}

	/* Create table */
	table = cmpack_tab_init(CMPACK_TABLE_LIGHTCURVE);
	cmpack_tab_pkyi(table, "Aperture", aperture.id);
	cmpack_fset_rewind(fset);
	if (cmpack_fset_get_frame(fset, CMPACK_FI_FILTER, &frame)==0) {
		if (frame.filter && frame.filter[0]!='\0')
			cmpack_tab_pkys(table, "Filter", frame.filter);
	}

	id_column = -1;
	if (frameids)
		id_column = cmpack_tab_add_column_int(table, "FRAME", 0, INT_MAX, -1);

	jd_column = hjd_column = -1;
	if (!jdhel || !hjdonly)
		jd_column = cmpack_tab_add_column_dbl(table, "JD", JD_PRECISION, 1e6, 1e99, 0.0);
	if (jdhel)
		hjd_column = cmpack_tab_add_column_dbl(table, "JDHEL", JD_PRECISION, 1e6, 1e99, 0.0);

	k = 1;
	vc_column = -1;
	if (!inst_mag) {
		/* V-C */
		sprintf(buf, "V-C");
		vc_column = cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
		sprintf(buf, "s%d", k++);
		cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
		/* V1-K1, V1-K2, ... */
		for (j=0; j<lc->chk.count; j++) {
			sprintf(buf, "V-K%d", j+1);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
			sprintf(buf, "s%d", k++);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
		}
		/* C1-K1, C1-K2, ... */
		for (i=0; i<lc->comp.count; i++) {
			for (j=0; j<lc->chk.count; j++) {
				sprintf(buf, "C%d-K%d", i+1, j+1);
				cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
				sprintf(buf, "s%d", k++);
				cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
			}
		}
		/* K1-K2, ... */
		for (i=0; i<lc->chk.count; i++) {
			for (j=i+1; j<lc->chk.count; j++) {
				sprintf(buf, "K%d-K%d", i+1, j+1);
				cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
				sprintf(buf, "s%d", k++);
				cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
			}
		}
	} else {
		/* V1, V2, ... */
		for (i=0; i<lc->var.count; i++) {
			sprintf(buf, "V%d", i+1);
			col = cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
			if (vc_column<0)
				vc_column = col;
			sprintf(buf, "s%d", k++);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
		}
		/* C1, C2, ... */
		for (i=0; i<lc->comp.count; i++) {
			sprintf(buf, "C%d", i+1);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
			sprintf(buf, "s%d", k++);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
		}
		/* K1-K2, ... */
		for (i=0; i<lc->chk.count; i++) {
			sprintf(buf, "K%d", i+1);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
			sprintf(buf, "s%d", k++);
			cmpack_tab_add_column_dbl(table, buf, MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
		}
	}

	hc_column = x_column = a_column = -1;
	if (helcorr)
		hc_column = cmpack_tab_add_column_dbl(table, "HELCOR", JD_PRECISION, -1.0, 1.0, 99.0);
	if (airmass)
		x_column = cmpack_tab_add_column_dbl(table, "AIRMASS", JD_PRECISION, 0.0, 1e99, -1.0);
	if (altitude)
		a_column = cmpack_tab_add_column_dbl(table, "ALTITUDE", ALT_PRECISION, -90.0, 90.0, 99.0);

	/* Process frames */
	res = cmpack_fset_rewind(fset);
	while (res==0) {
		cmpack_tab_append(table);
		cmpack_fset_get_frame(fset, CMPACK_FI_ID | CMPACK_FI_JULDAT | CMPACK_FI_HELCOR | CMPACK_FI_AIRMASS_ALT | CMPACK_FI_FILENAME, &frame);
		/* Check frame header */
		if (frameids && frame.frame_id<0) {
			if (frame.filename) 
				sprintf(msg, "%s: Invalid frame identifier", frame.filename);
			else
				strcpy(msg, "Invalid frame identifier");
			printout(lc->con, 0, msg);
		}
		if (frame.juldat<=0) {
			if (frame.filename) 
				sprintf(msg, "%s: Invalid Julian date of observation", frame.filename);
			else
				strcpy(msg, "Invalid Julian date of observation");
			printout(lc->con, 0, msg);
		}
		if ((helcorr || jdhel) && !frame.valid_helcor) {
			if (frame.filename)
				sprintf(msg, "%s: Invalid value of heliocentric correction observation", frame.filename);
			else
				strcpy(msg, "Invalid value of heliocentric correction observation");
			printout(lc->con, 0, msg);
		}
		if (altitude && !frame.altitude<-90) {
			if (frame.filename)
				sprintf(msg, "%s: Invalid value of altitude", frame.filename);
			else
				strcpy(msg, "Invalid value of altitude");
			printout(lc->con, 0, msg);
		}
		if (airmass && !frame.airmass<0) {
			if (frame.filename)
				sprintf(msg, "%s: Invalid value of air mass coefficient", frame.filename);
			else
				strcpy(msg, "Invalid value of air mass coefficient");
			printout(lc->con, 0, msg);
		}
		if (id_column>=0)
			cmpack_tab_ptdi(table, id_column, frame.frame_id);
		if (jd_column>=0)
			cmpack_tab_ptdd(table, jd_column, frame.juldat);
		if (hc_column>=0) 
			cmpack_tab_ptdd(table, hc_column, (frame.valid_helcor ? frame.helcor : -99.9));
		if (hjd_column>=0)
			cmpack_tab_ptdd(table, hjd_column, (frame.valid_helcor ? frame.juldat + frame.helcor : 0));
		if (a_column>=0)
			cmpack_tab_ptdd(table, a_column, frame.altitude);
		if (x_column>=0)
			cmpack_tab_ptdd(table, x_column, frame.airmass);
		/* Read magnitudes for selected stars */
		for (j=0; j<lc->var.count; j++) 
			star_get(&var[j], fset, ap_index);
		for (j=0; j<lc->comp.count; j++)
			star_get(&comp[j], fset, ap_index);
		for (j=0; j<lc->chk.count; j++)
			star_get(&chk[j], fset, ap_index);
		/* Debug printout */
		if (is_debug(lc->con)) {
			if (frame.filename)
				sprintf(msg, "FRAME: %d (%s)", frame.frame_id, frame.filename);
			else
				sprintf(msg, "FRAME: %d", frame.frame_id);
			printout(lc->con, 1, msg);
			sprintf(msg, "JD: %.7f", frame.juldat); 
			printout(lc->con, 1, msg);
			for (i=0; i<lc->var.count; i++) {
				sprintf(msg, "Variable star #%d:", i+1); 
				printout(lc->con, 1, msg);
				if (var[i].valid) {
					sprintf(msg, "\tMagnitude: %.3f +- %.3f mag", var[i].mag, var[i].err);
					printout(lc->con, 1, msg);
				}
			}
			for (i=0; i<lc->comp.count; i++) {
				sprintf(msg, "Comparison star #%d:", i+1); 
				printout(lc->con, 1, msg);
				if (comp[i].valid) {
					sprintf(msg, "\tMagnitude: %.3f +- %.3f mag", comp[i].mag, comp[i].err);
					printout(lc->con, 1, msg);
				}
			}
			for (i=0; i<lc->chk.count; i++) {
				sprintf(msg, "Check star #%d:", i+1); 
				printout(lc->con, 1, msg);
				if (chk[i].valid) {
					sprintf(msg, "\tMagnitude: %.3f +- %.3f mag", chk[i].mag, chk[i].err);
					printout(lc->con, 1, msg);
				}
			}
		}
		if (!inst_mag) 
			diff(lc, table, vc_column, var, lc->var.count, comp, lc->comp.count, chk, lc->chk.count, frame.filename);
		else
			inst(lc, table, vc_column, var, lc->var.count, comp, lc->comp.count, chk, lc->chk.count);
		res = cmpack_fset_next(fset);
	}

	*ptable = table;
	return CMPACK_ERR_OK;
}
