/**************************************************************

match.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/
#define _NO_OLDNAMES

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_catfile.h"
#include "cmpack_match.h"
#include "cmpack_phtfile.h"
#include "matfun.h"
#include "matread.h"
#include "matwrite.h"
#include "matsolve.h"
#include "matsimple.h"
#include "matstack.h"
#include "match.h"

/**********************   LOCAL FUNCTIONS   ******************************/

/* Clean context */
static void match_clear(CmpackMatch *lc)
{
	/* Free allocated memory */
	cmpack_free(lc->x1);
	cmpack_free(lc->y1);
	if (lc->con) {
		cmpack_con_destroy(lc->con);
		lc->con = NULL;
	}
}

/*********************   PUBLIC FUNCTIONS   *******************************/

/* Context initialization */
CmpackMatch *cmpack_match_init(void)
{
	CmpackMatch *lc = (CmpackMatch*)cmpack_calloc(1, sizeof(CmpackMatch));
	lc->refcnt    = 1;
	lc->maxstar   = 10;
	lc->nstar     = 5;
	lc->clip      = 2.5;
	lc->method    = CMPACK_MATCH_STANDARD;
	lc->maxoffset = 5.0;
	return lc;
}

/* Increment the reference counter */
CmpackMatch *cmpack_match_reference(CmpackMatch *ctx)
{
	ctx->refcnt++;
	return ctx;
}

/* Decrement reference counter / detroy the instance */
void cmpack_match_destroy(CmpackMatch *ctx)
{
	ctx->refcnt--;
	if (ctx->refcnt==0) {
		match_clear(ctx);
		cmpack_free(ctx);
	}
}

/* Attach console */
void cmpack_match_set_console(CmpackMatch *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		if (ctx->con) 
			cmpack_con_destroy(ctx->con);
		ctx->con = con;
		if (ctx->con) 
			cmpack_con_reference(ctx->con);
	}
}

/* Set max. number of stars used for matching */
void cmpack_match_set_maxstars(CmpackMatch *ctx, int nstars)
{
	ctx->maxstar = nstars;
}

/* Get max. number of stars used for matching */
int cmpack_match_get_maxstars(CmpackMatch *ctx)
{
	return ctx->maxstar;
}

/* Set number of vertices of polygons */
void cmpack_match_set_vertices(CmpackMatch *ctx, int vertices)
{
	ctx->nstar = vertices;
}

/* Get number of vertices of polygons */
int cmpack_match_get_vertices(CmpackMatch *ctx)
{
	return ctx->nstar;
}

/* Set clipping threshold */
void cmpack_match_set_threshold(CmpackMatch *ctx, double threshold)
{
	ctx->clip = threshold;
}

/* Get clipping threshold */
double cmpack_match_get_threshold(CmpackMatch *ctx)
{
	return ctx->clip;
}

/* Set matching method */
void cmpack_match_set_method(CmpackMatch *ctx, CmpackMatchMethod method)
{
	ctx->method = method;
}

/* Get matching method */
CmpackMatchMethod cmpack_match_get_method(CmpackMatch *ctx)
{
	return ctx->method;
}

/* Set Max. offset for 'sparse fields' method */
void cmpack_match_set_maxoffset(CmpackMatch *ctx, double maxoffset)
{
	ctx->maxoffset = maxoffset;
}

/* Get Max. offset for 'sparse fields' method */
double cmpack_match_get_maxoffset(CmpackMatch *ctx)
{
	return ctx->maxoffset;
}

/* Read reference file */
int cmpack_match_readref_pht(CmpackMatch *lc, CmpackPhtFile *pht)
{
	char msg[MAXLINE];

	/* Check source file */
	if (!pht) {
		printout(lc->con, 0, "Invalid reference file context");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Read reference stars */
	ReadRef(lc, pht);  
	sprintf(msg,"No. of stars   : %d",lc->c1);
	printout(lc->con, 1, msg);

  	/* Check reference file */
	if (lc->method==CMPACK_MATCH_STANDARD) {
		if (lc->c1<lc->nstar) {
	        printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	} else {
		if (lc->c1<1) {
			printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	}

  	/* Normal return */
  	return 0;
}

/* Read reference file */
int cmpack_match_readref_cat(CmpackMatch *lc, CmpackCatFile *cat)
{
	char msg[MAXLINE];

	/* Check source file */
	if (!cat) {
		printout(lc->con, 0, "Invalid reference file context");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Read reference stars */
	ReadCat(lc, cat);
	sprintf(msg,"No. of stars   : %d",lc->c1);
	printout(lc->con, 1, msg);

  	/* Check reference file */
	if (lc->method==CMPACK_MATCH_STANDARD) {
		if (lc->c1<lc->nstar) {
	        printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	} else {
		if (lc->c1<1) {
			printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	}

  	/* Normal return */
  	return 0;
}

static int match(CmpackMatch *lc, CmpackPhtFile *infile, CmpackPhtFile *outfile)
{
	char	msg[MAXLINE];
	double	x, y;
	int		res, max2;
	CmpackMatchFrame frame;

	memset(&frame, 0, sizeof(CmpackMatchFrame));
	/* Make copy of reference frame */
	frame.c1 = lc->c1;
	frame.x1 = (double*) cmpack_malloc(frame.c1*sizeof(double));
	memcpy(frame.x1, lc->x1, frame.c1*sizeof(double));
	frame.y1 = (double*) cmpack_malloc(frame.c1*sizeof(double));
	memcpy(frame.y1, lc->y1, frame.c1*sizeof(double));
	frame.width1 = lc->width;
	frame.height1 = lc->height;
	/* Alloc buffers */
	frame.idr1 = (struct _CmpackMatchObject*) cmpack_malloc(lc->maxstar*sizeof(struct _CmpackMatchObject));
	frame.id1 = (int*) cmpack_malloc(lc->maxstar*sizeof(int));
	frame.idr2 = (struct _CmpackMatchObject*) cmpack_malloc(lc->maxstar*sizeof(struct _CmpackMatchObject));
	frame.id2 = (int*) cmpack_malloc(lc->maxstar*sizeof(int));
	max2 = (lc->nstar*(lc->nstar-1)*(lc->nstar-2))/3+1;
	frame.dev = (double*)cmpack_malloc(max2*sizeof(double));
	frame.k2 = (double*)cmpack_malloc(max2*sizeof(double));

	StInit(&frame.stack);

	/* Check parameters */
	if (lc->nstar<=2) {
		printout(lc->con, 0,"Number of identification stars muse be greater than 2");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->nstar>=20) {
		printout(lc->con, 0,"Number of identification stars muse be less than 20");
		return CMPACK_ERR_INVALID_PAR;
	}	  
	if (lc->maxstar<lc->nstar) {
		printout(lc->con, 0,"Number of stars used muse be greater or equal to number of identification stars");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->maxstar>=1000) {
		printout(lc->con, 0,"Number of stars used for matching muse be less than 1000");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->clip<=0) {
		printout(lc->con, 0,"Clipping factor must be greater than zero");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->maxoffset<=0) {
		printout(lc->con, 0,"Max. position offset muse be greater than zero");
		return CMPACK_ERR_INVALID_PAR;
	}
        
	/* Read input file */
    ReadSrc(&frame, infile);
	sprintf(msg,"Number of stars: %d", frame.c2);
	printout(lc->con, 1, msg);

    /* Find coincidences */
	switch (lc->method) 
	{
	case CMPACK_MATCH_STANDARD:
		/* Always use standard method */
		if (frame.c1>=lc->nstar && frame.c2>=lc->nstar) {
			res = Solve(lc, &frame);
		} else {
			printout(lc->con, 0, "Too few stars in source file!");
			res = CMPACK_ERR_FEW_POINTS_SRC;
		}
	    break;
	case CMPACK_MATCH_AUTO:
		/* Use standard method if we have enough stars */
		if (frame.c1>=lc->nstar && frame.c2>=lc->nstar) {
		    res = Solve(lc, &frame);
		} else
		if (frame.c1>=1 && frame.c2>=1) {
		    res = Simple(lc, &frame);
		} else {
			printout(lc->con, 0, "Too few stars in source file!");
			res = CMPACK_ERR_FEW_POINTS_SRC;
		}
		break;
	case CMPACK_MATCH_SPARSE_FIELDS:
		/* Always use method for sparse fields */
		if (frame.c1>=1 && frame.c2>=1) {
			res = Simple(lc, &frame);
		} else {
			printout(lc->con, 0, "Too few stars in source file!");
			res = CMPACK_ERR_FEW_POINTS_SRC;
		}
		break;
	default:
		printout(lc->con, 0, "Unsupported matching method");
		res = CMPACK_ERR_INVALID_PAR;
    }

	if (res==0) {
		lc->matched_stars = frame.mstar;
		if (frame.mstar>0) {
			sprintf(msg,"Result         : %d matched (%.0f%%)", frame.mstar, (100.0*frame.mstar)/frame.c2);
			printout(lc->con, 1, msg);
			x = 0.5*frame.width2;
			y = 0.5*frame.height2;
			cmpack_matrix_transform_point(&frame.trafo, &x, &y);
			lc->offset[0] = x - 0.5*frame.width1;
			lc->offset[1] = y - 0.5*frame.height1;
		} else {
			printout(lc->con, 1,"Result     : Coincidences not found!");
			lc->offset[0] = lc->offset[1] = 0;
		}
		MatWrite(lc, &frame, outfile);
	}

    /* Normal return */
	StClear(&frame.stack);
	cmpack_free(frame.x1);
	cmpack_free(frame.y1);
	cmpack_free(frame.i2);
	cmpack_free(frame.x2);
	cmpack_free(frame.y2);
	cmpack_free(frame.idr1);
	cmpack_free(frame.idr2);
	cmpack_free(frame.id1);
	cmpack_free(frame.id2);
	cmpack_free(frame.k2);
	cmpack_free(frame.dev);
    return res;
}

int cmpack_match(CmpackMatch *lc, CmpackPhtFile *file, int *mstars)
{
	int res;

	if (mstars)
		*mstars = 0;
	lc->matched_stars = 0;
	lc->offset[0] = lc->offset[1] = 0;

	if (!file) {
		printout(lc->con, 0,"Invalid photometry file context");
		return CMPACK_ERR_INVALID_PAR;
	}
	res = match(lc, file, file);
	if (res==0 && mstars)
		*mstars = lc->matched_stars;
	return res;
}

int cmpack_match_get_offset(CmpackMatch *lc, double *offset_x, double *offset_y)
{
	if (!lc) 
		return CMPACK_ERR_INVALID_CONTEXT;
	if (lc->matched_stars==0)
		return CMPACK_ERR_UNDEF_VALUE;

	if (offset_x)
		*offset_x = lc->offset[0];
	if (offset_y)
		*offset_y = lc->offset[1];
	return 0;
}
