/**************************************************************

mfind.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "cmpack_common.h"
#include "cmpack_mfind.h"
#include "console.h"
#include "comfun.h"

/* Default threshold in % */
#define DEF_THRESHOLD		60.0

/* Clipping threshold */
#define SIGMA				3.0

/* Max. number of stars used in automatic finding of best comparison star */
#define RSTARS				30

/* Master-flat frame context */
struct _CmpackMuniFind
{
	int refcnt;					/**< Reference counter */
	CmpackConsole *con;			/**< Console */
	int aperture;				/**< Aperture index */
	int req_comp;				/**< <0 = Autodetection, >=0 comparison star */
	int act_comp;				/** Used comparison star identifier */
	double threshold;			/**< Required image data format */
};

typedef struct _ObjectRec
{
	int object_id, npoints;
	double meanmag, stddev;
} ObjectRec;

static int compare_fn(const void *a, const void *b)
{
	double ma = ((ObjectRec*)a)->meanmag, mb = ((ObjectRec*)b)->meanmag;
	return (ma < mb ? -1 : (ma > mb ? 1 : 0));
}

/********************   PUBLIC FUNCTIONS   ******************************/

/* Initializes the context */
CmpackMuniFind *cmpack_mfind_init(void)
{
	CmpackMuniFind *lc = (CmpackMuniFind*)cmpack_calloc(1, sizeof(CmpackMuniFind));
	lc->refcnt = 1;
	lc->req_comp = lc->act_comp = CMPACK_MFIND_AUTO;
	lc->threshold = DEF_THRESHOLD;
	return lc;
}

/* Create and initialize the context */
CmpackMuniFind *cmpack_mfind_reference(CmpackMuniFind *ctx)
{
	ctx->refcnt++;
	return ctx;
}

/* Create and initialize the context */
void cmpack_mfind_destroy(CmpackMuniFind *ctx)
{
	ctx->refcnt--;
	if (ctx->refcnt==0) {
		if (ctx->con) {
			cmpack_con_destroy(ctx->con);
			ctx->con = NULL;
		}
		cmpack_free(ctx);
	}
}

/* Attach console */
void cmpack_mfind_set_console(CmpackMuniFind *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		if (ctx->con) 
			cmpack_con_destroy(ctx->con);
		ctx->con = con;
		if (ctx->con) 
			cmpack_con_reference(ctx->con);
	}
}

/* Set aperture */
void cmpack_mfind_set_aperture(CmpackMuniFind *lc, int aperture)
{
	lc->aperture = aperture;
}

/* Set comparison star */
void cmpack_mfind_set_comparison(CmpackMuniFind *lc, int star_id)
{
	lc->req_comp = (star_id >= 0 ? star_id : CMPACK_MFIND_AUTO);
}

/* Set output image data format */
int cmpack_mfind_get_comparison(CmpackMuniFind *lc)
{
	return lc->act_comp;
}

/* Set output image data format */
void cmpack_mfind_set_threshold(CmpackMuniFind *lc, double threshold)
{
	lc->threshold = threshold;
}

/* Make mag-dev graph from read-all file */
int cmpack_mfind(CmpackMuniFind *lc, CmpackFrameSet *fset, CmpackTable **ptable,
	CmpackMFindFlags flags)
{
	char msg[256];
	int res, c, v, count, ap_index, nobjects, nframes, comp_star_index, count2;
	int i, j, vframes, valid, npoints, nmeasur;
	int nreferr, nrequired, nwritten, nbadpoints, cmp_npoints;
	double *data, sum, mean, stddev, q0;
	CmpackTable *table;
	CmpackCatObject obj;
	CmpackPhtAperture aperture;
	CmpackPhtData info, cinfo, vinfo;
	ObjectRec *tmp;

	*ptable = NULL;

	/* Check parameters */
	if (lc->threshold<0 || lc->threshold>100) {
		printout(lc->con, 0, "Invalid value of threshold"); 
		return CMPACK_ERR_INVALID_PAR;
	} 

	/* Print configuration parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Munifind parameters:");
		if (lc->req_comp!=CMPACK_MFIND_AUTO)
			printpari(lc->con, "Comparison star", 1, lc->req_comp);
		else
			printpars(lc->con, "Comparison star", 1, "autodetection");
		printpard(lc->con, "Threshold", 1, lc->threshold, 1);
	}

	/* Check number of frames */
	nframes = cmpack_fset_frame_count(fset);
	if (nframes<2) {
		printout(lc->con, 0, "Not enough frames in the input data"); 
		return CMPACK_ERR_INVALID_PAR;
	} 

	/* Check number of objects */
	nobjects = cmpack_fset_object_count(fset);
	if (nobjects<2) {
		printout(lc->con, 0, "Not enough objects in the input data"); 
		return CMPACK_ERR_INVALID_PAR;
	} 

	/* Check aperture */
	ap_index = cmpack_fset_find_aperture(fset, lc->aperture);
	if (ap_index<0) {
		printout(lc->con, 0, "Invalid aperture identifier"); 
		return CMPACK_ERR_AP_NOT_FOUND;
	} 
	aperture.id = -1;
	cmpack_fset_get_aperture(fset, ap_index, CMPACK_PA_ID, &aperture);

	/* Check comparison star */
	if (lc->req_comp == CMPACK_MFIND_AUTO) {
		double *fmag = (double*)cmpack_malloc(nframes*sizeof(double));
		double *sdev = (double*)cmpack_calloc(nobjects, sizeof(double));
		int    *scnt = (int*)cmpack_calloc(nobjects, sizeof(int));
		int		*sok = (int*)cmpack_calloc(nobjects, sizeof(int));
		/* Compute the maximum number of valid frames */
		vframes = 0;
		for (i=0; i<nobjects; i++) {
			count = 0;
			res = cmpack_fset_rewind(fset);
			while (res==0) {
				info.mag_valid = 0;
				cmpack_fset_get_data(fset, i, ap_index, &info);
				if (info.mag_valid) 
					count++;
				res = cmpack_fset_next(fset);
			}
			if (vframes < count)
				vframes = count;
		}
		/* Take stars that have maximum number of valid measurements */
		while (vframes>0) {
			count2 = 0;
			memset(sok, 0, nobjects*sizeof(int));
			for (i=0; i<nobjects; i++) {
				/* Count number of measurements */
				count = 0;
				res = cmpack_fset_rewind(fset);
				while (res==0) {
					info.mag_valid = 0;
					cmpack_fset_get_data(fset, i, ap_index, &info);
					if (info.mag_valid)
						count++;
					res = cmpack_fset_next(fset);
				}
				/* If count is maximum number, select it */
				if (count>=vframes) {
					sok[i] = 1;
					count2++;
				}
			}
			/* If less than three stars was found, decrease the number of frames required */
			if (count2>2)
				break;
			vframes--;
		}
		/* Select star with lowest rank */
		for (c=0; c<nobjects; c++) {
			if (sok[c]) {
				for (v=c+1; v<nobjects; v++) {
					if (sok[v]) {
						/* Make light curve v-c */
						count = 0;
						res = cmpack_fset_rewind(fset);
						while (res==0) {
							vinfo.mag_valid = cinfo.mag_valid = 0;
							cmpack_fset_get_data(fset, c, ap_index, &cinfo);
							cmpack_fset_get_data(fset, v, ap_index, &vinfo);
							if (vinfo.mag_valid && cinfo.mag_valid)
								fmag[count++] = vinfo.magnitude - cinfo.magnitude;
							res = cmpack_fset_next(fset);
						}
						if (count>=2) {
							/* Robust mean and sample standard deviation */
							cmpack_robustmean(count, fmag, &mean, NULL);
							sum = 0.0;
							for (j=0; j<count; j++) {
								double r = fmag[j]-mean;
								sum += r*r;
							}
							stddev = sqrt(sum/(count-1));
							sdev[c] += stddev;
							scnt[c]++;
							sdev[v] += stddev;
							scnt[v]++;
						}
					}
				}
			}
		}
		/* Find lowest */
		j = -1; q0 = 99.9e9;
		for (i=0; i<nobjects; i++) {
			if (sok[i] && scnt[i]>0) {
				sdev[i] /= scnt[i];
				if (sdev[i] < q0) {
					j = i;
					q0 = sdev[i];
				}
			}
		}
		comp_star_index = j;
		cmpack_free(fmag);
		cmpack_free(sdev);
		cmpack_free(scnt);
		/* Find object identifier for the comparison star*/
		cmpack_free(sok);
		obj.id = -1;
		cmpack_fset_get_object(fset, comp_star_index, CMPACK_OM_ID, &obj);
		lc->act_comp = obj.id;
	} else {
		/* Comparison star is specified by a caller */
		lc->act_comp = lc->req_comp;
		comp_star_index = cmpack_fset_find_object(fset, lc->req_comp);
	}

	/* Check comparison star */
	cmp_npoints = 0;
	if (comp_star_index>=0) {
		res = cmpack_fset_rewind(fset);
		while (res==0) {
			info.mag_valid = 0;
			cmpack_fset_get_data(fset, comp_star_index, ap_index, &info);
			if (info.mag_valid) 
				cmp_npoints++;
			res = cmpack_fset_next(fset);
		}
	}
	if (cmp_npoints<2) {
		printout(lc->con, 0, "Invalid comparison star"); 
		return CMPACK_ERR_REF_NOT_FOUND;
	}

	/* Compute minimum required measurements */
	nreferr = nframes - cmp_npoints;
	nrequired = (int)(0.01 * lc->threshold * cmp_npoints);
	if (nrequired<2) 
		nrequired = 2;

	/* computing average and std. dev. for each star */
	data = (double*)cmpack_malloc(nframes*sizeof(double));
	tmp = (ObjectRec*)cmpack_malloc(nobjects*sizeof(ObjectRec));
	nwritten = nbadpoints = nmeasur = 0;
	for (i=0; i<nobjects; i++) {
		valid = 0;
		mean = stddev = 0.0;
		obj.id = -1;
		cmpack_fset_get_object(fset, i, CMPACK_OM_ID, &obj);
		if (i != comp_star_index) {
			/* Make light curve i - comp_star_index */
			count = 0;
			res = cmpack_fset_rewind(fset);
			while (res==0) {
				vinfo.mag_valid = cinfo.mag_valid = 0;
				cmpack_fset_get_data(fset, i, ap_index, &vinfo);
				cmpack_fset_get_data(fset, comp_star_index, ap_index, &cinfo);
				if (cinfo.mag_valid && vinfo.mag_valid)
					data[count++] = vinfo.magnitude - cinfo.magnitude;
				res = cmpack_fset_next(fset);
			}
			if (count>=nrequired) {
				/* Robust mean + sample standard deviation */
				cmpack_robustmean(count, data, &mean, NULL);
				sum = 0.0;
				for (j=0; j<count; j++) {
					double r = data[j]-mean;
					sum += r*r;
				}
				stddev = sqrt(sum/(count-1));
				valid = 1;
			}
			npoints = count;
		} else {
			npoints = cmp_npoints;
		}
		nbadpoints += (cmp_npoints - npoints);
		/* Store data to table */
		if (valid) {
			tmp[nwritten].object_id = obj.id;
			tmp[nwritten].meanmag = mean;
			tmp[nwritten].stddev = stddev;
			tmp[nwritten].npoints = npoints;
			nwritten++;
		}
	}
	cmpack_free(data);

	/* Sort records in tmp by mean magnitude */
	qsort(tmp,nwritten,sizeof(ObjectRec),compare_fn);

	/* Init table */
	table = cmpack_tab_init(CMPACK_TABLE_MUNIFIND);
	cmpack_tab_pkyi(table, "Reference star", lc->act_comp);
	cmpack_tab_pkyi(table, "Aperture", aperture.id);
	cmpack_tab_add_column_int(table, "INDEX", 0, INT_MAX, -1);
	cmpack_tab_add_column_dbl(table, "MEAN_MAG", MAG_PRECISION, -99.0, 99.0, INVALID_MAG);
	cmpack_tab_add_column_dbl(table, "STDEV", MAG_PRECISION, 0.0, 9.0, INVALID_ERR);
	cmpack_tab_add_column_int(table, "GOODPOINTS", 0, INT_MAX, 0);

	/* Create table data */
	for (i=0; i<nwritten; i++) {
		cmpack_tab_append(table);
		cmpack_tab_ptdi(table, 0, tmp[i].object_id);
		cmpack_tab_ptdd(table, 1, tmp[i].meanmag);
		cmpack_tab_ptdd(table, 2, tmp[i].stddev);
		cmpack_tab_ptdi(table, 3, tmp[i].npoints); 
	}
	cmpack_free(tmp);
	
	/* Print messages */
	sprintf(msg,"Aperture:                           %d (%.4f pxl)", aperture.id, aperture.radius);
	printout(lc->con, 1, msg);
	sprintf(msg,"Number of analyzed stars:           %d (w/o reference star)", nobjects-1);
	printout(lc->con, 1, msg);
	sprintf(msg,"Number of source files:             %d", nframes);
	printout(lc->con, 1, msg);
	sprintf(msg,"Total number of data points:        %d", nmeasur);
	printout(lc->con, 1, msg);
	sprintf(msg,"Reference star used:                %d", lc->act_comp);
	printout(lc->con, 1, msg);
	sprintf(msg,"Reference star errors:              %d (%.0f%%)", nreferr, (nmeasur>0 ? 100.0*nreferr/nmeasur : 100.0));
	printout(lc->con, 1, msg);
	sprintf(msg,"Number of good points required:     %d", nrequired);
	printout(lc->con, 1, msg);
	sprintf(msg,"Number of stars written to output:  %d (%.0f%%)", nwritten, 100.0*nwritten/(nobjects-1));
	printout(lc->con, 1, msg);
	sprintf(msg,"Number of points ruled out:         %d (%.0f%%)", nbadpoints, (nmeasur>0 ? 100.0*nbadpoints/nmeasur : 100.0));
	printout(lc->con, 1, msg);
	
	*ptable = table;
    return 0;
}
