/**************************************************************

catalog_class.cpp (C-Munipack project)
Catalog file class interface
Copyright (C) 2011 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>

#include "catalog_class.h"

//-------------------------   PRIVATE DATA   ---------------------------

static const struct {
	const gchar *label;
	CmpackColor color;
} Types[CMPACK_SELECT_COUNT] = {
	{ NULL, CMPACK_COLOR_DEFAULT },
	{ "var", CMPACK_COLOR_RED },
	{ "comp", CMPACK_COLOR_GREEN },
	{ "check", CMPACK_COLOR_BLUE }
};

//-------------------   Catalogue file class interface  -------------------------------------

//
// Constructor(s)
//
CCatalog::CCatalog():m_Handle(NULL), m_CacheFlags(0), m_RefMag(0)
{
}

//
// Destructor
//
CCatalog::~CCatalog()
{
	if (m_Handle)
		cmpack_cat_destroy(m_Handle);
}

//
// Create memory only file
//
void CCatalog::Create(void)
{
	Close();

	m_Handle = cmpack_cat_new();
	InvalidateCache();
}

//
// Copy data from a file
//
bool CCatalog::Open(const gchar *fpath, CmpackOpenMode mode, int *result)
{
	int res;

	Close();

	if (mode != CMPACK_OPEN_CREATE) {
		if (!fpath && !g_file_test(fpath, G_FILE_TEST_IS_REGULAR)) {
			if (result)
				*result = CMPACK_ERR_OPEN_ERROR;
			return false;
		}
	} else {
		if (!fpath) {
			if (result)
				*result = CMPACK_ERR_OPEN_ERROR;
			return false;
		}
	}

	char *f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	res = cmpack_cat_open(&m_Handle, f, mode, 0);
	g_free(f);
	if (res!=0) {
		if (result)
			*result = res;
		return false;
	}

	InvalidateCache();
	if (result)
		*result = 0;
	return true;
}

//
// Refresh cached data
//
void CCatalog::InvalidateCache()
{
	m_CacheFlags = 0;
}

//
// Get selection
//
const CSelectionList *CCatalog::Selections(void)
{
	if (!(m_CacheFlags & CF_SELECTIONS)) {
		m_Selections.Clear();
		if (m_Handle) {
			int count = cmpack_cat_selection_set_count(m_Handle);
			for (int i=0; i<count; i++) {
				cmpack_cat_set_current_selection_set(m_Handle, i);
				CSelection sel;
				LoadSelection(sel, CMPACK_SELECT_VAR);
				LoadSelection(sel, CMPACK_SELECT_COMP);
				LoadSelection(sel, CMPACK_SELECT_CHECK);
				if (sel.Count()>0) 
					m_Selections.Set(cmpack_cat_get_selection_set_name(m_Handle), sel);
			}
		}
		m_CacheFlags |= CF_SELECTIONS;
	}
	return &m_Selections;
}

//
// Get selection
//
const CTags *CCatalog::Tags(void)
{
	int id;
	const char *tag;

	if (!(m_CacheFlags & CF_TAGS)) {
		m_Tags.Clear();
		if (m_Handle) {
			int count = cmpack_cat_get_tag_count(m_Handle);
			for (int i=0; i<count; i++) {
				cmpack_cat_get_tag(m_Handle, i, &id, &tag);
				m_Tags.Set(id, tag);
			}
		}
		m_CacheFlags |= CF_TAGS;
	}
	return &m_Tags;
}

//
// Get object coordinates
//
const CObjectCoords *CCatalog::Object(void)
{
	if (!(m_CacheFlags & CF_OBJECT)) {
		m_Object.Clear();
		if (m_Handle) {
			const gchar *name = cmpack_cat_gkys(m_Handle, "object");
			if (name) {
				gchar *aux = g_strdup(name);
				m_Object.SetName(g_strstrip(aux));
				g_free(aux);
			}
			const gchar *ra = cmpack_cat_gkys(m_Handle, "ra2000");
			if (ra) {
				gchar *aux = g_strdup(ra);
				m_Object.SetRA(g_strstrip(aux));
				g_free(aux);
			}
			const gchar *dec = cmpack_cat_gkys(m_Handle, "dec2000");
			if (dec) {
				gchar *aux = g_strdup(dec);
				m_Object.SetDec(g_strstrip(aux));
				g_free(aux);
			}
		}
		m_CacheFlags |= CF_OBJECT;
	}
	return &m_Object;
}

//
// Get location
//
const CLocation *CCatalog::Location(void)
{
	if (!(m_CacheFlags & CF_LOCATION)) {
		m_Location.Clear();
		if (m_Handle) {
			const gchar *name = cmpack_cat_gkys(m_Handle, "observatory");
			if (name) {
				gchar *aux = g_strdup(name);
				m_Location.SetName(g_strstrip(aux));
				g_free(aux);
			}
			const gchar *lon = cmpack_cat_gkys(m_Handle, "longitude");
			if (lon) {
				gchar *aux = g_strdup(lon);
				m_Location.SetLon(g_strstrip(aux));
				g_free(aux);
			}
			const gchar *lat = cmpack_cat_gkys(m_Handle, "latitude");
			if (lat) {
				gchar *aux = g_strdup(lat);
				m_Location.SetLat(g_strstrip(aux));
				g_free(aux);
			}
		}
		m_CacheFlags |= CF_LOCATION;
	}
	return &m_Location;
}

//
// Clear data
//
bool CCatalog::Close(int *result)
{
	if (m_Handle) {
		int res = cmpack_cat_close(m_Handle);
		if (res!=0) {
			if (*result)
				*result = res;
			return false;
		}
		m_Handle = NULL;
		m_Selections.Clear();
		m_Object.Clear();
		m_Location.Clear();
		m_Tags.Clear();
		m_RefMag = 0;
		InvalidateCache();
	}
	if (result)
		*result = 0;
	return true;
}

//
// Date of observation
//
double CCatalog::JulianDate(void) const
{
	double jd;

	if (m_Handle)
		if (cmpack_cat_gkyd(m_Handle, "jd", &jd)==0)
			return jd;
	return 0.0;
}

//
// Exposure duration
//
double CCatalog::ExposureDuration(void) const
{
	double exptime;

	if (m_Handle)
		if (cmpack_cat_gkyd(m_Handle, "exptime", &exptime)==0)
			return exptime;
	return 0.0;
}

//
// Number of stars
//
int CCatalog::ObjectCount(void) const
{
	if (m_Handle)
		return cmpack_cat_nstar(m_Handle);
	return 0;
}

//
// Find object
//
int CCatalog::FindObject(int id)
{
	if (m_Handle)
		return cmpack_cat_find_star(m_Handle, id);
	return -1;
}


//
// Get star identifier
//
int CCatalog::GetObjectID(int index) const
{
	CmpackCatObject obj;

	if (m_Handle && cmpack_cat_get_star(m_Handle, index, CMPACK_OM_ID, &obj)==0)
		return obj.id;
	return -1;
}

//
// Get star identifier
//
bool CCatalog::GetObject(int star_index, unsigned mask, CmpackCatObject *obj) const
{
	if (m_Handle)
		return cmpack_cat_get_star(m_Handle, star_index, mask, obj)==0;
	return false;
}

//
// Load selection (one type)
//
void CCatalog::LoadSelection(CSelection &selection, CmpackSelectionType type1)
{
	int i, j, count, id, nstars;
	CmpackSelectionType type;
	
	nstars = cmpack_cat_get_selection_count(m_Handle);
	count = 0;
	for (i=0; i<nstars; i++) {
		cmpack_cat_get_selection(m_Handle, i, NULL, &type);
		if (type == type1) 
			count++;
	}
	if (count>0) {
		int *buf = (int*)g_malloc(count*sizeof(int));
		j = 0;
		for (i=0; i<nstars; i++) {
			cmpack_cat_get_selection(m_Handle, i, &id, &type);
			if (type == type1)
				buf[j++] = id;
		}
		selection.SetStarList(type1, buf, j);
		g_free(buf);
	}
}

//
// Save selection to a file
//
void CCatalog::SetSelections(const CSelectionList &map)
{
	if (m_Handle) {
		m_Selections = map;
		cmpack_cat_clear_selection(m_Handle);
		for (int i=0; i<map.Count(); i++) {
			const CSelection *s = map.At(i);
			if (s) {
				if (cmpack_cat_selection_set_new(m_Handle, map.Name(i), -1)==0) {
					SaveSelection(*s, CMPACK_SELECT_VAR);
					SaveSelection(*s, CMPACK_SELECT_COMP);
					SaveSelection(*s, CMPACK_SELECT_CHECK);
				}
			}
		}
		m_CacheFlags |= CF_SELECTIONS;
	}
}

//
// Save tags to a file
//
void CCatalog::SetTags(const CTags &tags)
{
	int id;
	const char *tag;

	if (m_Handle) {
		m_Tags = tags;
		cmpack_cat_clear_tags(m_Handle);
		CTags::tIter iter;
		bool ok = m_Tags.First(&iter, &id, &tag);
		while (ok) {
			cmpack_cat_update_tag(m_Handle, id, tag);
			ok = m_Tags.Next(&iter, &id, &tag);
		}
		m_CacheFlags |= CF_TAGS;
	}
}

//
// Save selection
//
void CCatalog::SaveSelection(const CSelection &selection, CmpackSelectionType type) const
{
	int i, count, *buf;

	count = selection.CountStars(type);
	if (count>0) {
		buf = (int*)g_malloc(count*sizeof(int));
		selection.GetStarList(type, buf, count);
		for (i=0; i<count; i++) 
			cmpack_cat_update_selection(m_Handle, buf[i], type);
		g_free(buf);
	}
}

//
// Chart width in pixels
//
int CCatalog::Width(void) const
{
	if (m_Handle)
		return cmpack_cat_get_width(m_Handle);
	return 0;
}

//
// Chart height in pixels
//
int CCatalog::Height(void) const
{
	if (m_Handle)
		return cmpack_cat_get_height(m_Handle);
	return 0;
}

//
// Date of observation
//
bool CCatalog::DateTime(CmpackDateTime *dt) const
{
	double jd = JulianDate();
	if (jd>0.0)
		return cmpack_decodejd(jd, dt)==0;
	return false;
}

//
// Filter name
//
const char *CCatalog::Filter(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "filter");
	return NULL;
}
void CCatalog::SetFilter(const gchar *filter)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "filter", filter, NULL);
}

//
// Observer's name
//
const gchar *CCatalog::Observer(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "observer");
	return NULL;
}
void CCatalog::SetObserver(const gchar *observer)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "observer", observer, NULL);
}

//
// Telescope
//
const gchar *CCatalog::Telescope(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "telescope");
	return NULL;
}
void CCatalog::SetTelescope(const gchar *telescope)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "telescope", telescope, NULL);
}

//
// Camera
//
const gchar *CCatalog::Camera(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "camera");
	return NULL;
}
void CCatalog::SetCamera(const gchar *camera)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "camera", camera, NULL);
}

//
// Field of view
//
const gchar *CCatalog::FieldOfView(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "fov");
	return NULL;
}
void CCatalog::SetFieldOfView(const gchar *fov)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "fov", fov, NULL);
}

//
// Orientation
//
const gchar *CCatalog::Orientation(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "orientation");
	return NULL;
}
void CCatalog::SetOrientation(const gchar *orientation)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "orientation", orientation, NULL);
}

//
// Notes
//
const gchar *CCatalog::Notes(void) const
{
	if (m_Handle)
		return cmpack_cat_gkys(m_Handle, "notes");
	return NULL;
}
void CCatalog::SetNotes(const gchar *notes)
{
	if (m_Handle)
		cmpack_cat_pkys(m_Handle, "notes", notes, NULL);
}

//
// Get header field by index
//
bool CCatalog::GetParam(int index, const char **key, const char **val, const char **com) const
{
	if (m_Handle)
		return cmpack_cat_gkyn(m_Handle, index, key, val, com)!=0;
	return false;
}

//
// Update reference magnitudes
//
double CCatalog::GetRefMag(void)
{
	int i, count;
	CmpackCatObject obj;

	if (!(m_CacheFlags & CF_REF_MAG)) {
		m_RefMag = 0.0;
		if (m_Handle) {
			int nstars = cmpack_cat_nstar(m_Handle);
			if (nstars>0) {
				double *maglist = (double*)g_malloc(nstars*sizeof(double));
				count = 0;
				for (i=0; i<nstars; i++) {
					if (GetObject(i, CMPACK_OM_MAGNITUDE, &obj) && obj.refmag_valid) 
						maglist[count++] = obj.refmagnitude;
				}
				if (count>0) 
					cmpack_robustmean(count, maglist, &m_RefMag, NULL);
				g_free(maglist);
			}
		}
		m_CacheFlags |= CF_REF_MAG;
	}
	return m_RefMag;
}

//
// Set location
//
void CCatalog::SetLocation(const CLocation &loc)
{
	if (m_Handle && m_Location!=loc) {
		m_Location = loc;
		if (m_Location.Name())
			cmpack_cat_pkys(m_Handle, "observatory", m_Location.Name(), NULL);
		else
			cmpack_cat_dkey(m_Handle, "observatory");
		if (m_Location.Lon()) 
			cmpack_cat_pkys(m_Handle, "longitude", m_Location.Lon(), NULL);
		else
			cmpack_cat_dkey(m_Handle, "longitude");
		if (m_Location.Lat()) 
			cmpack_cat_pkys(m_Handle, "latitude", m_Location.Lat(), NULL);
		else
			cmpack_cat_dkey(m_Handle, "latitude");
		m_CacheFlags |= CF_LOCATION;
	}
}

//
// Set object coordinates
//
void CCatalog::SetObject(const CObjectCoords &obj)
{
	if (m_Handle && m_Object!=obj) {
		m_Object = obj;
		if (m_Object.Name())
			cmpack_cat_pkys(m_Handle, "object", m_Object.Name(), NULL);
		else
			cmpack_cat_dkey(m_Handle, "object");
		if (m_Object.RA())
			cmpack_cat_pkys(m_Handle, "ra2000", m_Object.RA(), NULL);
		else
			cmpack_cat_dkey(m_Handle, "ra2000");
		if (m_Object.Dec())
			cmpack_cat_pkys(m_Handle, "dec2000", m_Object.Dec(), NULL);
		else
			cmpack_cat_dkey(m_Handle, "dec2000");
		m_CacheFlags |= CF_OBJECT;
	}
}

//
// Make catalogue file from photometry file
//
bool CCatalog::Assign(const CPhot &pht, int aperture, int *result)
{
	int res;

	if (!pht.Handle()) {
		if (result)
			*result = CMPACK_ERR_INVALID_PAR;
		return false;
	}

	if (!m_Handle) 
		m_Handle = cmpack_cat_new();

	res = cmpack_cat_make(m_Handle, pht.Handle(), aperture);
	if (res!=0) {
		if (result)
			*result = res;
		return false;
	}

	InvalidateCache();
	if (result)
		*result = CMPACK_ERR_OK;
	return true;
}

//
// Make deep copy of a catalogue file
//
bool CCatalog::Assign(const CCatalog &cat, int *result)
{
	if (!m_Handle)
		m_Handle = cmpack_cat_new();

	cmpack_cat_copy(m_Handle, cat.Handle());

	InvalidateCache();
	if (result)
		*result = 0;
	return true;
}

//
// Make chart data
//
CmpackChartData *CCatalog::ToChartData(int selectionIndex, bool show_tags, bool transparent)
{
	static const unsigned mask = CMPACK_OM_ID | CMPACK_OM_CENTER | CMPACK_OM_MAGNITUDE;

	gint count, width, height;
	CmpackCatObject obj;
	CmpackChartItem item;

	memset(&item, 0, sizeof(CmpackChartItem));
	
	if (m_Handle) {
		CmpackChartData *data = cmpack_chart_data_new();
		width = cmpack_cat_get_width(m_Handle);
		height = cmpack_cat_get_height(m_Handle);
		cmpack_chart_data_set_dimensions(data, width, height);
		count = cmpack_cat_nstar(m_Handle);
		cmpack_chart_data_alloc(data, count);
		gdouble ref = GetRefMag();
		for (int i=0; i<count; i++) {
			cmpack_cat_get_star(m_Handle, i, mask, &obj);
			item.x = obj.center_x;
			item.y = obj.center_y;
			item.outline = !obj.refmag_valid;
			item.param = obj.id;
			if (!transparent)
				item.d = 2.0 + MAX(0, (obj.refmag_valid ? 1.0 + 0.75*(ref - obj.refmagnitude) : 1.0));
			else
				item.d = 0;
			cmpack_chart_data_add(data, &item, sizeof(item));
		}
		const CTags *tags = (show_tags ? Tags() : NULL);
		const CSelection *selection = (selectionIndex>=0 ? Selections()->At(selectionIndex) : NULL);
		if (selection || tags) {
			int count = cmpack_chart_data_count(data);
			for (int row=0; row<count; row++) {
				int star_id = cmpack_chart_data_get_param(data, row);
				const gchar *tag = (tags ? tags->Get(star_id) : NULL);
				int i = (selection ? selection->IndexOf(star_id) : -1);
				if (i>=0) {
					int index = selection->GetIndex(i);
					CmpackSelectionType type = selection->GetType(i);
					gchar *buf = (gchar*)g_malloc((256+(tag ? strlen(tag)+1 : 0))*sizeof(gchar));
					if (index==1)
						strcpy(buf, Types[type].label);
					else
						sprintf(buf, "%s #%d", Types[type].label, index);
					if (tag) {
						strcat(buf, "\n");
						strcat(buf, tag);
					}
					cmpack_chart_data_set_tag(data, row, buf);
					cmpack_chart_data_set_color(data, row, Types[type].color);
					cmpack_chart_data_set_topmost(data, row, TRUE);
					if (transparent)
						cmpack_chart_data_set_diameter(data, row, 4.0);
					g_free(buf);
				} else if (tag) {
					cmpack_chart_data_set_tag(data, row, tag);
					cmpack_chart_data_set_color(data, row, CMPACK_COLOR_YELLOW);
					cmpack_chart_data_set_topmost(data, row, TRUE);
					if (transparent)
						cmpack_chart_data_set_diameter(data, row, 4.0);
				}
			}
		}
		return data;
	}
	return NULL;
}
