/**************************************************************

catalog_class.h (C-Munipack project)
Catalog file class interface
Copyright (C) 2011 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_CATALOG_CLASS_H
#define MUNIWIN_CATALOG_CLASS_H

#include "helper_classes.h"
#include "phot_class.h"

//
// Catalogue file class
//
class CCatalog
{
public:
	// Constructor(s)
	CCatalog();

	// Destructor
	virtual ~CCatalog();

	// Create file stored in memory
	void Create(void);

	// Load data from a file
	bool Open(const gchar *fpath, CmpackOpenMode mode, int *result = NULL);

	// Create from photometry file
	bool Assign(const CPhot &pht, int aperture_index, int *result = NULL);

	// Make deep copy of another catalog file
	bool Assign(const CCatalog &cat, int *result = NULL);

	// Free internal data
	bool Close(int *result = NULL);

	// Is the object valid?
	bool Valid(void) const
	{ return m_Handle!=NULL; }

	// File's handle
	CmpackCatFile *Handle(void) const
	{ return m_Handle; }

	// Get header field by index
	bool GetParam(int index, const char **key, const char **val, const char **com) const;
	
	// Chart width in pixels
	int Width(void) const;

	// Chart height in pixels
	int Height(void) const;

	// Date of observation
	bool DateTime(CmpackDateTime *dt) const;
	double JulianDate(void) const;

	// Exposure duration
	double ExposureDuration(void) const;

	// Frame properties
	const gchar *Filter(void) const;
	const gchar *Observer(void) const;
	const gchar *Telescope(void) const;
	const gchar *Camera(void) const;
	const gchar *FieldOfView(void) const;
	const gchar *Orientation(void) const;
	const gchar *Notes(void) const;

	void SetFilter(const gchar *filter);
	void SetObserver(const gchar *observer);
	void SetTelescope(const gchar *telescope);
	void SetCamera(const gchar *camera);
	void SetFieldOfView(const gchar *fov);
	void SetOrientation(const gchar *orientation);
	void SetNotes(const gchar *notes);

	// Number of stars
	int ObjectCount(void) const;

	// Find object
	int FindObject(int id);

	// Get star identifier
	int GetObjectID(int star_index) const;

	// Get star parameters
	bool GetObject(int star_index, unsigned mask, CmpackCatObject *obj) const;

	// Change selection map
	void SetSelections(const CSelectionList &sel);

	// Get selection map
	const CSelectionList *Selections(void);

	// Change tags
	void SetTags(const CTags &tags);

	// Get tags
	const CTags *Tags(void);

	// Change object coordinates
	void SetObject(const CObjectCoords &obj);

	// Get object coordinates
	const CObjectCoords *Object(void);

	// Change object coordinates
	void SetLocation(const CLocation &loc);

	// Get location
	const CLocation *Location(void);

	// Make chart data
	CmpackChartData *ToChartData(int selectionIndex, bool show_tags, bool transparent);

	// Make table model
	GtkTreeModel *ToTreeModel(void);

protected:
	enum tCacheFlags {
		CF_SELECTIONS	= (1<<0),
		CF_REF_MAG		= (1<<1),
		CF_OBJECT		= (1<<2), 
		CF_LOCATION		= (1<<3),
		CF_TAGS			= (1<<4)
	};

	CmpackCatFile	*m_Handle;
	unsigned		m_CacheFlags;
	double			m_RefMag;
	CSelectionList	m_Selections;
	CTags			m_Tags;
	CObjectCoords	m_Object;
	CLocation		m_Location;

	// Refresh cached data
	void InvalidateCache(void);

	// Compute reference magnitude for given aperture
	double GetRefMag(void);

	// Load selection
	void LoadSelection(CSelection &selection, CmpackSelectionType type);

	// Save selection
	void SaveSelection(const CSelection &selection, CmpackSelectionType type) const;

private:
	// Disable copy constructor and assigment operator
	CCatalog(const CCatalog &cat);
	CCatalog &operator=(const CCatalog &orig);
};

#endif
