/**************************************************************

chart_dlg.cpp (C-Munipack project)
The 'Plot chart' dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>

#include "utils.h"
#include "main.h"
#include "ctxhelp.h"
#include "configuration.h"
#include "export_dlgs.h"
#include "export_options.h"

#define MIN_WIDTH			1
#define MIN_HEIGHT			1
#define MAX_WIDTH			8192
#define MAX_HEIGHT			8192

//-------------------------   SAVE CHART DIALOG   --------------------------------

//
// Constructor
//
CChartExportDlg::CChartExportDlg(GtkWindow *pParent):m_Parent(pParent), m_ImageWidth(0), 
	m_ImageHeight(0), m_JpegQuality(80)
{
	GtkFileFilter *filters[2];

	// Dialog with buttons
	m_pDlg = gtk_file_chooser_dialog_new(NULL, GTK_WINDOW(pParent),
		GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, 
		GTK_RESPONSE_ACCEPT, GTK_STOCK_HELP, GTK_RESPONSE_HELP, NULL);
	gtk_file_chooser_standard_tooltips(GTK_FILE_CHOOSER(m_pDlg));
	g_signal_connect(G_OBJECT(m_pDlg), "response", G_CALLBACK(response_dialog), this);
	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(m_pDlg), true);

	// File filters
	filters[0] = gtk_file_filter_new();
	gtk_file_filter_add_pixbuf_formats(filters[0]);
	gtk_file_filter_set_name(filters[0], "Images");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);
	filters[1] = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filters[1], "*");
	gtk_file_filter_set_name(filters[1], "All files");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[1]);
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);

	// Dialog icon
	gchar *icon = get_icon_file("muniwin");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);

	// Chart size widget
	m_Widget = new CExportChartOptions();
	gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(m_pDlg), m_Widget->Handle());
	m_Widget->RegisterCallback(WidgetCallback, this);
}

//
// Destructor
//
CChartExportDlg::~CChartExportDlg()
{
	gtk_widget_destroy(m_pDlg);
	delete m_Widget;
}

//
// Execute the dialog
//
bool CChartExportDlg::Execute(CmpackChartData *chart, CmpackImageData *image, const gchar *name,
	bool invertImage, bool rowsUpwards)
{
	bool retval = true;

	int width, height;
	if (chart) {
		width = cmpack_chart_data_width(chart);
		height = cmpack_chart_data_height(chart);
	} else if (image) {
		width = cmpack_image_data_width(image);
		height = cmpack_image_data_height(image);
	} else
		return false;

	gchar *filePath = Execute(name, width, height);
	if (filePath) {
		CmpackOrientation orienation = (rowsUpwards ? CMPACK_ROWS_UPWARDS : CMPACK_ROWS_DOWNWARDS);
		if (image && !chart) {
			retval = cmpack_image_data_write_to_file(image, filePath, 
				m_Format, m_ImageWidth, m_ImageHeight, m_JpegQuality)!=0;
		} else
		if (chart && image) {
			retval = cmpack_chart_data_write_to_file(chart, image, filePath, 
				m_Format, m_ImageWidth, m_ImageHeight, invertImage, orienation, m_JpegQuality)!=0;
		} else 
		if (chart && !image) {
			retval = cmpack_chart_data_write_to_file(chart, NULL, filePath, 
				m_Format, m_ImageWidth, m_ImageHeight, invertImage, orienation, m_JpegQuality)!=0;
		}
		if (!retval) 
			ShowError(m_Parent, "Failed to write image to a file");
		g_free(filePath);
	}
	return retval;
}

//
// Execute the dialog
//
gchar *CChartExportDlg::Execute(const gchar *name, int image_width, int image_height)
{
	m_ImageWidth = image_width;
	m_ImageHeight = image_height;

	gchar *type = CConfig::GetStr("ChartExport", "Type", DEFAULT_IMAGE_TYPE);
	m_Widget->Init(m_ImageWidth, m_ImageHeight,
		(CExportChartOptions::tSizeMode)CConfig::GetInt("ChartExport", "Mode", CExportChartOptions::RELATIVE, 0, 1),
		type,
		CConfig::GetDbl("ChartExport", "Zoom", 1.0),
		CConfig::GetInt("ChartExport", "Width", m_ImageWidth, 1, 65535),
		CConfig::GetInt("ChartExport", "JpegQuality", 80, 0, 100));
	g_free(type);

	// Restore last folder and filename
	gchar *folder = g_Project->GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = CConfig::GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = g_path_get_dirname(g_Project->Path());
	if (folder && g_file_test(folder, G_FILE_TEST_IS_DIR)) 
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(m_pDlg), folder);
	g_free(folder);

	// Set the default file name
	if (!name)
		name = "chart";
	const gchar *ext = ImageFileExtension(m_Widget->Type());
	if (!ext) 
		ext = ImageFileExtension(DEFAULT_IMAGE_TYPE);
	gchar *newname = AddFileExtension(name, ext);
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(m_pDlg), newname);
	g_free(newname);

	// Execute the dialog
	gtk_window_set_title(GTK_WINDOW(m_pDlg), "Export chart");
	if (gtk_dialog_run(GTK_DIALOG(m_pDlg)) != GTK_RESPONSE_ACCEPT)
		return false;
	gtk_widget_hide(m_pDlg);

	gchar *filePath = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_pDlg));

	m_ImageWidth = m_Widget->OutputWidth();
	m_ImageHeight = m_Widget->OutputHeight();
	m_JpegQuality = m_Widget->JpegQuality();
	m_Format = m_Widget->Type();
	if (!m_Format)
		m_Format = ImageTypeByExtension(filePath);

	// Save configuration
	CConfig::SetInt("ChartExport", "Mode", (int)m_Widget->Mode());
	CConfig::SetStr("ChartExport", "Type", m_Widget->Type());
	CConfig::SetDbl("ChartExport", "Zoom", m_Widget->Zoom());
	CConfig::SetInt("ChartExport", "Width", m_Widget->OutputWidth());
	CConfig::SetInt("ChartExport", "JpegQuality", m_JpegQuality);
	folder = g_path_get_dirname(filePath);
	g_Project->SetStr("Output", "Folder", folder);
	CConfig::SetStr("Output", "Folder", folder);
	g_free(folder);

	return filePath;
}

void CChartExportDlg::response_dialog(GtkWidget *pDlg, gint response_id, CChartExportDlg *pMe)
{
	if (!pMe->OnResponseDialog(response_id)) 
		g_signal_stop_emission_by_name(pDlg, "response");
}

bool CChartExportDlg::OnResponseDialog(gint response_id)
{
	switch (response_id)
	{
	case GTK_RESPONSE_ACCEPT:
		// Check values
		if (!CheckFormat()) {
			ShowError(GTK_WINDOW(m_pDlg), "Invalid file format.\n\n"
				"If you specify the file format by an extension, check that the file name contains a valid extension. "
				"Current version of the software supports the following formats: PNG and JPEG");
			return false;
		}
		if (m_Widget->OutputWidth()<MIN_WIDTH || m_Widget->OutputHeight()<MIN_HEIGHT) {
			ShowError(GTK_WINDOW(m_pDlg), "Invalid image size");
			return false;
		}
		if (m_Widget->OutputWidth()>MAX_WIDTH || m_Widget->OutputHeight()>MAX_HEIGHT) {
			ShowError(GTK_WINDOW(m_pDlg), "The image is too large.");
			return false;
		}
		break;

	case GTK_RESPONSE_HELP:
		// Show context help
		g_MainWnd->ShowHelp(GTK_WINDOW(m_pDlg), IDH_SAVE_CHART);
		return false;
	}
	return true;
}

void CChartExportDlg::WidgetCallback(CCBObject *sender, int message, int wparam, void* lparam, void* cb_data)
{
	CChartExportDlg *pMe = (CChartExportDlg*)cb_data;

	switch (message)
	{
	case CExportChartOptions::CB_TYPE_CHANGED:
		pMe->OnTypeChanged();
		break;
	}
}

void CChartExportDlg::OnTypeChanged(void)
{
	const gchar *ext = ImageFileExtension(m_Widget->Type());
	if (ext) {
		gchar *oldname = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_pDlg));
		if (oldname) {
			gchar *newname = SetFileExtension(oldname, ext);
			gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(m_pDlg), g_basename(newname));
			g_free(newname);
			g_free(oldname);
		}
	}
}

bool CChartExportDlg::CheckFormat(void)
{
	const gchar *type = m_Widget->Type();
	if (!type) {
		gchar *fname = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_pDlg));
		if (fname) {
			type = ImageTypeByExtension(fname);
			g_free(fname);
		}
	}
	return (type!=NULL);
}

//---------------------------   GRAPH EXPORT DIALOG   ------------------------

CGraphExportDlg::CGraphExportDlg(GtkWindow *pParent):m_Parent(pParent)
{
	GtkFileFilter *filters[2];

	// Dialog with buttons
	m_pDlg = gtk_file_chooser_dialog_new(NULL, GTK_WINDOW(pParent),
		GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, 
		GTK_RESPONSE_ACCEPT, GTK_STOCK_HELP, GTK_RESPONSE_HELP, NULL);
	gtk_file_chooser_standard_tooltips(GTK_FILE_CHOOSER(m_pDlg));
	g_signal_connect(G_OBJECT(m_pDlg), "response", G_CALLBACK(response_dialog), this);
	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(m_pDlg), true);

	// File filters
	filters[0] = gtk_file_filter_new();
	gtk_file_filter_add_pixbuf_formats(filters[0]);
	gtk_file_filter_set_name(filters[0], "Images");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);
	filters[1] = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filters[1], "*");
	gtk_file_filter_set_name(filters[1], "All files");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[1]);
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);

	// Dialog icon
	gchar *icon = get_icon_file("muniwin");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);
}

CGraphExportDlg::~CGraphExportDlg()
{
	gtk_widget_destroy(m_pDlg);
}

bool CGraphExportDlg::Execute(CmpackGraphView *widget, const gchar *graphName)
{
	// Restore last folder
	gchar *folder = g_Project->GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = CConfig::GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = g_path_get_dirname(g_Project->Path());
	if (folder && g_file_test(folder, G_FILE_TEST_IS_DIR)) 
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(m_pDlg), folder);
	g_free(folder);

	// Set default file name
	if (!graphName)
		graphName = "graph";
	gchar *newname = AddFileExtension(graphName, ImageFileExtension("png"));
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(m_pDlg), newname);
	g_free(newname);

	// Execute the dialog
	gtk_window_set_title(GTK_WINDOW(m_pDlg), "Export graph");
	if (gtk_dialog_run(GTK_DIALOG(m_pDlg)) != GTK_RESPONSE_ACCEPT)
		return false;
	gtk_widget_hide(m_pDlg);

	gchar *filePath = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_pDlg));

	// Save configuration
	folder = g_path_get_dirname(filePath);
	g_Project->SetStr("Output", "Folder", folder);
	CConfig::SetStr("Output", "Folder", folder);
	g_free(folder);

	// Make output file
	gboolean retval = cmpack_graph_view_write_to_file(widget, filePath, "png");
	g_free(filePath);
	return retval!=0;
}

void CGraphExportDlg::response_dialog(GtkWidget *pDlg, gint response_id, CGraphExportDlg *pMe)
{
	if (!pMe->OnResponseDialog(response_id)) 
		g_signal_stop_emission_by_name(pDlg, "response");
}

bool CGraphExportDlg::OnResponseDialog(gint response_id)
{
	switch (response_id)
	{
	case GTK_RESPONSE_HELP:
		// Show context help
		g_MainWnd->ShowHelp(GTK_WINDOW(m_pDlg), IDH_SAVE_GRAPH);
		return false;
	}
	return true;
}

//---------------------------   TABLE EXPORT DIALOG   ------------------------

CTableExportDlg::CTableExportDlg(GtkWindow *pParent):m_Parent(pParent)
{
	GtkFileFilter *filters[2];

	// Dialog with buttons
	m_pDlg = gtk_file_chooser_dialog_new(NULL, GTK_WINDOW(pParent),
		GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, 
		GTK_RESPONSE_ACCEPT, GTK_STOCK_HELP, GTK_RESPONSE_HELP, NULL);
	gtk_file_chooser_standard_tooltips(GTK_FILE_CHOOSER(m_pDlg));
	g_signal_connect(G_OBJECT(m_pDlg), "response", G_CALLBACK(response_dialog), this);
	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(m_pDlg), true);

	// File filters
	filters[0] = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filters[0], "*.csv");
	gtk_file_filter_set_name(filters[0], "CSV files");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);
	filters[1] = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filters[1], "*");
	gtk_file_filter_set_name(filters[1], "All files");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[1]);
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);

	// Dialog icon
	gchar *icon = get_icon_file("muniwin");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);
}

CTableExportDlg::~CTableExportDlg()
{
	gtk_widget_destroy(m_pDlg);
}

void CTableExportDlg::Execute(const CPhot &file, int apertureIndex, const gchar *name,
	int sort_column_id, GtkSortType sort_order)
{
	gchar *filePath = Execute(name);
	if (filePath) {
		GError *error = NULL;
		if (!file.ExportTable(apertureIndex, filePath, "csv", sort_column_id, sort_order, &error)) {
			if (error) {
				ShowError(m_Parent, error->message);
				g_error_free(error);
			}
		}
		g_free(filePath);
	}
}

void CTableExportDlg::Execute(CTable &table, const gchar *name,
	int sort_column_id, GtkSortType sort_order)
{
	gchar *filePath = Execute(name);
	if (filePath) {
		GError *error = NULL;
		if (!table.Export(filePath, "csv", sort_column_id, sort_order, &error)) {
			if (error) {
				ShowError(m_Parent, error->message);
				g_error_free(error);
			}
		}
		g_free(filePath);
	}
}

gchar *CTableExportDlg::Execute(const gchar *name)
{
	// Restore last folder
	gchar *folder = g_Project->GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = CConfig::GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = g_path_get_dirname(g_Project->Path());
	if (folder && g_file_test(folder, G_FILE_TEST_IS_DIR)) 
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(m_pDlg), folder);
	g_free(folder);

	// Set default file name
	if (!name)
		name = "table";
	gchar *newname = AddFileExtension(name, TableFileExtension("csv"));
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(m_pDlg), newname);
	g_free(newname);

	// Execute the dialog
	gtk_window_set_title(GTK_WINDOW(m_pDlg), "Export table");
	if (gtk_dialog_run(GTK_DIALOG(m_pDlg)) != GTK_RESPONSE_ACCEPT)
		return NULL;
	gtk_widget_hide(m_pDlg);

	gchar *filePath = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_pDlg));

	// Save configuration
	folder = g_path_get_dirname(filePath);
	g_Project->SetStr("Output", "Folder", folder);
	CConfig::SetStr("Output", "Folder", folder);
	g_free(folder);

	return filePath;
}

void CTableExportDlg::response_dialog(GtkWidget *pDlg, gint response_id, CTableExportDlg *pMe)
{
	if (!pMe->OnResponseDialog(response_id)) 
		g_signal_stop_emission_by_name(pDlg, "response");
}

bool CTableExportDlg::OnResponseDialog(gint response_id)
{
	switch (response_id)
	{
	case GTK_RESPONSE_HELP:
		// Show context help
		g_MainWnd->ShowHelp(GTK_WINDOW(m_pDlg), IDH_SAVE_TABLE);
		return false;
	}
	return true;
}

//---------------------------   EXPORT PHOTOMETRY FILE   ------------------------

CPhotExportDlg::CPhotExportDlg(GtkWindow *pParent):m_Parent(pParent)
{
	GtkFileFilter *filters[2];

	// Dialog with buttons
	m_pDlg = gtk_file_chooser_dialog_new(NULL, GTK_WINDOW(pParent),
		GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, 
		GTK_RESPONSE_ACCEPT, GTK_STOCK_HELP, GTK_RESPONSE_HELP, NULL);
	gtk_file_chooser_standard_tooltips(GTK_FILE_CHOOSER(m_pDlg));
	g_signal_connect(G_OBJECT(m_pDlg), "response", G_CALLBACK(response_dialog), this);
	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(m_pDlg), true);

	// File filters
	filters[0] = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filters[0], "*.csv");
	gtk_file_filter_set_name(filters[0], "CSV files");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);
	filters[1] = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filters[1], "*");
	gtk_file_filter_set_name(filters[1], "All files");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(m_pDlg), filters[1]);
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(m_pDlg), filters[0]);

	// Dialog icon
	gchar *icon = get_icon_file("muniwin");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);
}

CPhotExportDlg::~CPhotExportDlg()
{
	gtk_widget_destroy(m_pDlg);
}

void CPhotExportDlg::Execute(const CPhot &file, int apertureIndex, const gchar *name,
	int sort_column_id, GtkSortType sort_order)
{
	gchar *filePath = Execute(name);
	if (filePath) {
		GError *error = NULL;
		if (!file.ExportTable(apertureIndex, filePath, "csv", sort_column_id, sort_order, &error)) {
			if (error) {
				ShowError(m_Parent, error->message);
				g_error_free(error);
			}
		}
		g_free(filePath);
	}
}

gchar *CPhotExportDlg::Execute(const gchar *name)
{
	// Restore last folder
	gchar *folder = g_Project->GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = CConfig::GetStr("Output", "Folder", NULL);
	if (!folder)
		folder = g_path_get_dirname(g_Project->Path());
	if (folder && g_file_test(folder, G_FILE_TEST_IS_DIR)) 
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(m_pDlg), folder);
	g_free(folder);

	// Set default file name
	if (!name)
		name = "table";
	gchar *newname = AddFileExtension(name, TableFileExtension("csv"));
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(m_pDlg), newname);
	g_free(newname);

	// Execute the dialog
	gtk_window_set_title(GTK_WINDOW(m_pDlg), "Export photometry file");
	if (gtk_dialog_run(GTK_DIALOG(m_pDlg)) != GTK_RESPONSE_ACCEPT)
		return false;
	gtk_widget_hide(m_pDlg);

	gchar *filePath = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_pDlg));

	// Save configuration
	folder = g_path_get_dirname(filePath);
	g_Project->SetStr("Output", "Folder", folder);
	CConfig::SetStr("Output", "Folder", folder);
	g_free(folder);

	return filePath;
}

void CPhotExportDlg::response_dialog(GtkWidget *pDlg, gint response_id, CPhotExportDlg *pMe)
{
	if (!pMe->OnResponseDialog(response_id)) 
		g_signal_stop_emission_by_name(pDlg, "response");
}

bool CPhotExportDlg::OnResponseDialog(gint response_id)
{
	switch (response_id)
	{
	case GTK_RESPONSE_HELP:
		// Show context help
		g_MainWnd->ShowHelp(GTK_WINDOW(m_pDlg), IDH_SAVE_TABLE);
		return false;
	}
	return true;
}
