/**************************************************************

selection.h (C-Munipack project)
Table of selected stars
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_HELPER_CLASSES_H
#define MUNIWIN_HELPER_CLASSES_H

#include <gtk/gtk.h>
#include <cmunipack.h>

#include "callback.h"
#include "cmpack_widgets.h"

#define MAX_TAG_SIZE 1023

// Date time formats
enum tDateFormat {
	JULIAN_DATE,
	GREGORIAN_DATE
};

// File types
enum tFileType
{
	TYPE_UNKNOWN	= 0,
	TYPE_IMAGE		= (1<<0),
	TYPE_PHOT		= (1<<1),
	TYPE_CAT		= (1<<2),
	TYPE_TABLE		= (1<<3),
	TYPE_FRAMESET	= (1<<4),
	TYPE_PROJECT    = (1<<5),
	TYPE_PROFILE	= (1<<6),
	TYPE_ALL		= 0xFFFF
};

// Autodetect file type
tFileType FileType(const char *path);

// Project type to recent group
const gchar *FileTypeRecentGroup(tFileType type);

//
// Console - receives messages from C-Munipack library
//
class CConsole
{
public:
	// Constructor
	CConsole(void) {}

	// Destructor
	virtual ~CConsole(void) {}
	
	// Callback function that calls instance's OnPrint
	static void OutputProc(const char *text, void *user_data);

	// Calls OnPrint
	void Print(const char *text);
	
protected:
	// Must be overriden, message processing
	virtual void OnPrint(const char *text) = 0;
};


//
// Selection of stars
//
class CSelection
{
public:
	// Constructor(s)
	CSelection();
	CSelection(const gchar *string);
	CSelection(const CSelection &other);
	
	// Destructor
	virtual ~CSelection();

	// Clear data
	void Clear();

	// Begin/end update
	void BeginUpdate(void);
	void EndUpdate(void);

	// Assignment operator
	CSelection &operator=(const CSelection &other);

	// Select a star
	int Select(int id, CmpackSelectionType type);

	// Convert to string
	gchar *toString(void) const;

	// Unselect a star
	void Unselect(CmpackSelectionType type);

	// Find record that matches specified star identifier
	// Returns the index of the record
	int IndexOf(int id) const;

	// Get number of selected stars
	int Count(void) const
	{ return m_Count; }

	// Get number of stars of specified type
	int CountStars(CmpackSelectionType type) const;

	// Enumerate selected stars
	int GetId(int i) const;
	CmpackSelectionType GetType(int i) const;
	int GetIndex(int i) const;

	// Get identifiers of selected star
	int GetStarList(CmpackSelectionType type, int *list, int maxcount) const;
	char *GetStarList(CmpackSelectionType type) const;

	// Set list of selected stars
	void SetStarList(CmpackSelectionType type, const int *list, int count);
	void SetStarList(CmpackSelectionType type, const char *string);

protected:
	struct tStarRec
	{
		int id;
		CmpackSelectionType type;
		int index;
	};

	struct tReindexInfo
	{
		int			var_index;
		int			comp_index;
		int			chk_index;
	};

	int m_Count, m_Capacity;
	tStarRec *m_List;
	bool m_Changed;
	int m_LockCount;

	// Search star by its id
	int Search(int id) const;

	// Refresh selection indices
	void Reindex(void);

	// Callback emission is postponed if the object is locked
	void Changed(void);

	// Comparison operators
	friend bool operator==(const CSelection&, const CSelection&);
};

bool operator==(const CSelection &a, const CSelection &b);
inline bool operator!=(const CSelection &a, const CSelection &b) { return !operator==(a,b); }

//
// List of selections
//
class CSelectionList
{
public:
	// Constructor(s)
	CSelectionList(void):m_List(NULL) {}
	CSelectionList(const CSelectionList &other);

	// Destructor
	virtual ~CSelectionList();

	// Clear data
	void Clear();

	// Add a new selection (takes ownership of the instance)
	void Set(const gchar *name, const CSelection &value);

	// Remove a selection from the list (caller must delete it)
	void Remove(const gchar *name);
	void RemoveAt(int index);

	// Get number of records
	int Count(void) const;

	// Get record by index
	const gchar *Name(int index) const;
	const CSelection *At(int index) const;

	// Get first/last record
	int IndexOf(const gchar *name) const;
	int IndexOf(const CSelection &value) const;

	// Find selection by name
	const CSelection *Value(const gchar *name) const;

	// Assignment operator (makes deep copy)
	CSelectionList &operator=(const CSelectionList&);

private:
	struct tData {
		gchar *name;
		CSelection *data;
		tData *next;
	};

	tData *m_List;

	tData *findInsertPos(const gchar *name) const;
	tData *findData(const gchar *name) const;
};


//
// List of object tags
//
class CTags
{
private:
	struct tData
	{
		int id;
		gchar *tag;
		tData *next;
	};

public:
	// Constructor(s)
	CTags(void):m_List(NULL) {}
	CTags(const CTags &orig);
	
	// Destructor
	virtual ~CTags();

	// Clear data
	void Clear();

	// Assignment operator (makes deep copy)
	CTags &operator=(const CTags&);

	// Set a tag
	void Set(int id, const gchar *tag);

	// Remove a tag
	void Remove(int id);

	// Get a tag
	const gchar *Get(int id) const;

	// Number of tags
	int Count(void) const;

	// Enumeration
	struct tIter { struct CTags::tData *ptr; };
	bool First(tIter *iter, int *id, const gchar **tag) const;
	bool Next(tIter *iter, int *id, const gchar **tag) const;
	
private:
	tData *m_List;
};


//
// File information
//
class CFileInfo
{
public:
	// Constructors
	CFileInfo();
	CFileInfo(const gchar *filepath);
	CFileInfo(const CFileInfo &orig);

	// Destructor
	virtual ~CFileInfo();

	// Assignment operator
	CFileInfo &operator=(const CFileInfo &orig);

	// Clear data
	void Clear(void);

	// Is the data valid?
	bool Valid() const
	{ return m_FullPath!=NULL; }

	// Comparison operators
	int Compare(const CFileInfo &other) const;

	// File path
	const gchar *FullPath() const
	{ return m_FullPath; }

	// File name
	const gchar *Name() const
	{ return (m_FullPath ? g_basename(m_FullPath) : NULL); }

	// Load from project file
	void Load(GKeyFile *cfg, const gchar *section, const gchar *prefix);

	// Save to project file
	void Save(GKeyFile *cfg, const gchar *section, const gchar *prefix) const;

private:
	gchar		*m_FullPath;
	time_t		m_MTime;
};


//
// Frame information
//
class CFrameInfo
{
public:
	// Constructors
	CFrameInfo();
	CFrameInfo(const CFrameInfo &orig);

	// Destructor
	virtual ~CFrameInfo();

	// Assignment operator
	CFrameInfo &operator=(const CFrameInfo &orig);

	// Clear data
	void Clear(void);

	// Initialize data
	bool Init(const gchar *filepath, int *result = 0);

	// Is the data valid?
	bool Valid() const
	{ return m_FullPath!=NULL; }

	// File path
	const gchar *FullPath() const
	{ return m_FullPath; }

	// File name
	const gchar *Name() const
	{ return (m_FullPath ? g_basename(m_FullPath) : NULL); }

	// Julian date
	double JulDat() const
	{ return m_JulDat; }

	// Exposure time
	double ExpTime() const
	{ return m_ExpTime; }

	// CCD temperature
	double CCDTemp() const
	{ return m_CCDTemp; }

	// Color filter name
	const gchar *Filter() const
	{ return m_Filter; }
	
	// Object designation
	const gchar *Object() const
	{ return m_Object; }

	// Number of subframes averaged
	int AvgFrames() const
	{ return m_AvgFrames; }

	// Number of subframes summed
	int SumFrames() const
	{ return m_SumFrames; }

	// File format
	CmpackFormat Format() const
	{ return m_Format; }

private:
	gchar	*m_FullPath;
	gchar	*m_Filter, *m_Object;
	double	m_JulDat, m_ExpTime, m_CCDTemp;
	int		m_AvgFrames, m_SumFrames;
	CmpackFormat m_Format;
};


//
// Object coordinates
//
class CObjectCoords
{
public:
	// Constructor(s)
	CObjectCoords();
	CObjectCoords(const CObjectCoords &orig);
	
	// Destructor
	virtual ~CObjectCoords();

	// assignment operator
	CObjectCoords &operator=(const CObjectCoords &orig);

	// Set object designation
	void SetName(const char *name);

	// Assign coordinates
	void Assign(const CmpackObjCoords *obj);

	// Set object coordinates
	void SetRA(const char *ra);
	void SetDec(const char *dec);

	// Set source
	void SetSource(const char *source);

	// Set comment
	void SetRemarks(const char *remarks);

	// Clear data
	void Clear();

	// Object designation
	const char *Name(void) const
	{ return m_Name; }

	// Object coordinates
	const char *RA(void) const
	{ return m_RA; }
	const char *Dec(void) const
	{ return m_Dec; }

	// Source
	const char *Source(void) const
	{ return m_Source; }

	// Remarks
	const char *Remarks(void) const
	{ return m_Remarks; }

	// Check validity of coordinates
	bool Valid(void) const;

protected:
	char	*m_Name;
	char	*m_RA, *m_Dec;
	char	*m_Source, *m_Remarks;

	// Comparison operators
	friend bool operator==(const CObjectCoords&, const CObjectCoords&);
};

// Comparison operators
bool operator==(const CObjectCoords &a, const CObjectCoords &b);
inline bool operator!=(const CObjectCoords &a, const CObjectCoords &b) { return !operator==(a,b); }


//
// Observer coordinates
//
class CLocation
{
public:
	// Constructor(s)
	CLocation();
	CLocation(const CLocation &orig);
	
	// Destructor
	virtual ~CLocation();

	// assignment operator
	CLocation &operator=(const CLocation &orig);

	// Assign location
	void Assign(const CmpackLocation *location);

	// Set location name
	void SetName(const char *name);

	// Set observer's coordinates
	void SetLon(const char *lon);
	void SetLat(const char *lat);

	// Set comment
	void SetComment(const char *remarks);

	// Clear data
	void Clear();

	// Location name
	const char *Name(void) const
	{ return m_Name; }

	// Observer's coordinates
	const char *Lon(void) const
	{ return m_Lon; }
	const char *Lat(void) const
	{ return m_Lat; }

	// Remarks
	const char *Comment(void) const
	{ return m_Com; }

	// Check validity of coordinates
	bool Valid(void) const;

protected:
	char	*m_Name;
	char	*m_Lon, *m_Lat;
	char	*m_Com;

	// Comparison operators
	friend bool operator==(const CLocation&, const CLocation&);
};

// Comparison operators
bool operator==(const CLocation &a, const CLocation &b);
inline bool operator!=(const CLocation &a, const CLocation &b) { return !operator==(a,b); }


//
// Aperture parameters
//
class CAperture 
{
public:
	// Constructors
	CAperture():m_Id(0), m_Radius(0) {}
	CAperture(int id, double radius):m_Id(id), m_Radius(radius) {}

	// Destructor
	virtual ~CAperture(void) {}

	// Get identifier
	int Id(void) const
	{ return m_Id; }

	// Get radius in pixels
	double Radius(void) const
	{ return m_Radius; }

	// Is the aperture valid?
	bool Valid(void) const
	{ return m_Radius>0; }

private:
	int		m_Id;				// Aperture identifier
	double	m_Radius;			// Radius in pixels
};

//
// Table of apertures
//
class CApertures
{
public:
	// Constructor(s)
	CApertures();
	CApertures(const CApertures &orig);
	
	// Destructor
	virtual ~CApertures();

	// assignment operator
	CApertures &operator=(const CApertures &orig);

	// Get number of apertures
	int Count(void) const
	{ return m_Count; }

	// Returns true if the table is empty
	bool Empty(void) const
	{ return m_Count==0; }

	// Clear the table
	void Clear(void);

	// Add an aperture, if the aperture with the same id is 
	// in the table, it changes its parameters
	void Add(const CAperture &item);

	// Find aperture by id 
	int Find(int id) const;

	// Return index of smallest aperture
	int FindSmallest(void) const;

	// Get aperture by index
	const CAperture *Get(int index) const;

	// Get aperture by index
	int GetId(int index) const;

	// Get aperture by index
	double GetRadius(int index) const;

protected:
	int			m_Count, m_Capacity;
	CAperture	**m_List;
};

#endif
