/**************************************************************

phot_class.h (C-Munipack project)
Photometry file class interface
Copyright (C) 2011 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_PHOT_CLASS_H
#define MUNIWIN_PHOT_CLASS_H

#include "helper_classes.h"
#include "file_classes.h"

//
// Photometry file class
//
class CPhot
{
public:
	// Table columns (also sorting columns)
	enum tColumnId {
		COL_ID, COL_REF_ID, COL_POS_X, COL_POS_Y, COL_FWHM, COL_SKY, 
		COL_SKY_DEV, COL_MAG_VALID, COL_MAG, COL_MAG_ERR, COL_REMARKS, 
		COL_COUNT
	};

	// Additional columns for sorting
	enum tSortColumnId {
		SORT_INT = COL_COUNT, SORT_NOISE, SORT_SNR
	};

	// Constructor(s)
	CPhot();
	
	// Destructor
	virtual ~CPhot();

	// Copy data from a file
	bool Open(const gchar *fpath, CmpackOpenMode mode, unsigned flags = 0,
		int *result = NULL);

	// Make deep copy from a photometry file
	bool Assign(const CPhot &pht, int *result = NULL);

	// Close file, free internal data
	bool Close(int *result = NULL);

	// Is the object valid?
	bool Valid(void) const
	{ return m_Handle!=NULL; }

	// File's handle
	CmpackPhtFile *Handle(void) const
	{ return m_Handle; }

	// Get header field by index
	bool GetParams(unsigned mask, CmpackPhtInfo &info) const;

	// Date of observation
	bool DateTime(CmpackDateTime *dt) const;
	double JulianDate(void) const;

	// Exposure duration
	double ExposureDuration(void) const;

	// Has the file been matched?
	bool Matched(void) const;

	// Number of stars
	int ObjectCount(void) const;

	// Find object
	int FindObject(int id);

	// Find object by reference identifier
	int FindObjectRefID(int ref_id);

	// Get star identifier
	int GetObjectID(int star_index);

	// Get reference indentifier
	int GetObjectRefID(int star_index);

	// Get position of a star
	bool GetObjectPos(int star_index, double *x, double *y);

	// Get star attributes
	bool GetObjectParams(int star_index, unsigned mask, CmpackPhtObject *obj);

	// Get measurement
	bool GetMagnitude(int star_index, int ap_index, double *mag, double *err);

	// Get measurement with a code indicating a reason why the brightness is not valid
	bool GetMagnitudeAndCode(int star_index, int ap_index, CmpackPhtData &res, CmpackError &code);

	// Number of apertures
	int ApertureCount(void) const;

	// Find aperture
	int FindAperture(int id) const;

	// Get table of apertures
	const CApertures *Apertures(void);

	// Make chart data
	CmpackChartData *ToChartData(int aperture_index, bool gray_not_matched, bool transparent = false);

	// Make table model
	GtkTreeModel *ToTreeModel(int aperture_index) const;

	// Set view parameters
	void SetView(GtkTreeView *view) const;

	// Make table model
	bool ExportTable(int aperture_index, const gchar *filepath, const gchar *format, int sort_column_id,
		GtkSortType sort_order, GError **error) const;

protected:
	enum tCacheFlags {
		CF_APERTURES	= (1<<0)
	};

	struct tRefMag
	{
		bool valid;
		double mag;
	};
	
	CmpackPhtFile	*m_Handle;
	int				m_CacheFlags;
	CApertures		m_Apertures;
	int				m_RefCount;
	tRefMag			*m_RefMag;

	// Compute reference magnitude for given aperture
	double GetRefMag(int ap_index);

	// Refresh cached data
	void InvalidateCache(void);

private:
	// Disable copy constructor and assigment operator
	CPhot(const CPhot&);
	CPhot &operator=(const CPhot &orig);

	void ExportTable(CCSVWriter &file, int aperture_index, int sort_column_id, GtkSortType sort_order) const;
};

#endif
