/**************************************************************

project.cpp (C-Munipack project)
Table of input frames
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib/gstdio.h>

#include "project.h"
#include "utils.h"
#include "ccdfile_class.h"
#include "configuration.h"
#include "main.h"

//-------------------------------   PRIVATE DATA TYPES   -----------------------------------

struct tGetFileInfo
{
	int			id;
	GtkTreePath *path;
};

//-------------------------------   HELPER FUNCTIONS   -----------------------------------

static gboolean FindFile(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	int fileid;
	tGetFileInfo *pData = (tGetFileInfo*)data;

	gtk_tree_model_get(model, iter, FRAME_ID, &fileid, -1);
	if (fileid == pData->id) {
		pData->path = gtk_tree_path_copy(path);
		return TRUE;
	}
	return FALSE;
}

static gboolean MakeStatus(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, CProject::tStatus *st)
{
	unsigned state;

	gtk_tree_model_get(model, iter, FRAME_STATE, &state, -1);
	if (state & CFILE_CONVERSION)
		st->converted++;
	if (state & CFILE_PHOTOMETRY)
		st->photometred++;
	if (state & CFILE_MATCHING)
		st->matched++;
	st->files++;

	return FALSE;
}

// Return true if the given buffer can be a project file
static bool checkProjectFile(GKeyFile *file, tProjectType *type)
{
	if (g_key_file_has_group(file, "CMUNIPACK-PROJECT")) {
		int revision = g_key_file_get_integer(file, "CMUNIPACK-PROJECT", "Revision", NULL);
		if (revision==1) {
			if (type) {
				gchar *str = g_key_file_get_string(file, "Profile", "Type", NULL);
				*type = StrToProjectType(str);
				g_free(str);
			}
			return true;
		}
	}
	return false;
}

// Check that the file is writable
static bool checkWritable(const gchar *fpath)
{
	FILE *f = open_file(fpath, "a");
	if (f) {
		fclose(f);
		return true;
	}
	return false;
}

// Ask user if we can open project in read-only mode
static bool confirmReadOnly(GtkWindow *parent, const gchar *fpath)
{
	GtkWidget *dialog = gtk_message_dialog_new(parent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT),
		GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE, "A project is already opened in another application. Do you want to open it anyway?");
	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), "The project file \"%s\" is already opened in another application. "
		"You can answer yes to open the project in the read-only mode. You will be able to make outputs (i.e. a light curve), but you won't be allowed to do any action that would "
		"change the frames or the photometry files.", fpath);
	gtk_window_set_title(GTK_WINDOW (dialog), "Question");
	gtk_dialog_add_button(GTK_DIALOG (dialog), "_No, thanks.", GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button(GTK_DIALOG (dialog), "_Yes, open in read-only mode", GTK_RESPONSE_ACCEPT);
	gtk_dialog_set_alternative_button_order(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT, GTK_RESPONSE_CANCEL, -1);
	gtk_dialog_set_default_response(GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL);
	int response = gtk_dialog_run(GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
	return (response == GTK_RESPONSE_ACCEPT);
}

//-------------------------------   PROJECT CLASS   -----------------------------------

// Constructor
CProject::CProject(void):m_Name(NULL), m_Path(NULL), m_DataDir(NULL), m_ReadOnly(true),
	m_FreeID(1), m_RefType(REF_UNDEFINED), m_RefID(-1), m_Ref(NULL), m_Observer(NULL)
{
	m_Frames = gtk_list_store_new(FRAME_NCOLS, G_TYPE_INT, G_TYPE_UINT, G_TYPE_UINT, 
		G_TYPE_DOUBLE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, 
		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_DOUBLE, 
		G_TYPE_DOUBLE, G_TYPE_DOUBLE, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT,
		G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_DOUBLE, G_TYPE_DOUBLE);
	m_Params = g_key_file_new();
	m_FileLock = new FileLock();
}

// Destructor
CProject::~CProject(void)
{
	if (m_Ref)
		gtk_tree_row_reference_free(m_Ref);
	g_free(m_Name);
	g_free(m_Path);
	g_free(m_DataDir);
	g_object_unref(m_Frames);
	g_key_file_free(m_Params);
	g_free(m_Observer);
	delete m_FileLock;
}

// Lock data
void CProject::Lock(void) const
{
	gdk_threads_lock();
}

// Unlock data
void CProject::Unlock(void) const
{
	gdk_threads_unlock();
}

void CProject::SetLastSelection(const CSelection &sel)
{
	Lock();
	m_Selection = sel;
	Unlock();
}

// Set obejct's coordinates
void CProject::SetObjectCoords(const CObjectCoords &obj)
{
	Lock();
	m_Coords = obj;
	Unlock();
}

// Set reference observer's name
void CProject::SetObserver(const gchar *name)
{
	Lock();
	g_free(m_Observer);
	m_Observer = (name!=NULL ? g_strdup(name) : NULL);
	Unlock();
}

// Set reference observer's coordinates
void CProject::SetLocation(const CLocation &obs)
{
	Lock();
	m_Location = obs;
	Unlock();
}

// Set reference observer's coordinates
void CProject::SetApertures(const CApertures &list)
{
	Lock();
	m_Apertures = list;
	Unlock();
}

// Close project
void CProject::Close(void)
{
	Lock();
	if (m_Path) 
		Save();
	m_FileLock->Release();
	reset();
	m_Profile.Clear();
	g_free(m_Name);
	m_Name = NULL;
	g_free(m_Path);
	m_Path = NULL;
	g_free(m_DataDir);
	m_DataDir = NULL;
	Unlock();
}

// Open existing project
bool CProject::Open(const gchar *path, tOpenMode mode, GError **error, GtkWindow *parent)
{
	g_return_val_if_fail(error == NULL || *error == NULL, false);

	Lock();
	Close();

	// Acquire file lock
	m_ReadOnly = true;
	if (mode!=OPEN_READONLY) {
		if (m_FileLock->Acquire(path)) {
			m_ReadOnly = !checkWritable(path);
			if (m_ReadOnly && mode==OPEN_CREATE) {
				gchar *dirpath = g_path_get_dirname(path);
				g_set_error(error, g_AppError, 0, "You do not have permission to create a file in the directory: %s", dirpath);
				g_free(dirpath);
				Unlock();
				return false;
			} 
		} else {
			if (mode==OPEN_CREATE || !parent || !confirmReadOnly(parent, path)) {
				g_set_error(error, g_AppError, 0, "Access denied.");
				Unlock();
				return false;
			}
		}
	}
	if (mode!=OPEN_CREATE) {
		// Read a project file
		if (!read(path, false, error)) {
			m_FileLock->Release();
			Unlock();
			return false;
		}
	} else {
		// Set parameters to default values
		reset();
		m_Profile.Clear();
	}

	// Make sure that data directory exists
	gchar *datadir = g_strdup_printf("%s-files", path);
	if (!datadir || !force_directory(datadir)) {
		g_set_error(error, g_AppError, 0, "Failed to create the directory: %s", datadir);
		g_free(datadir);
		m_FileLock->Release();
		Unlock();
		return false;
	}

	if (mode==OPEN_CREATE) {
		// Write a new project file
		if (!write(path, error)) {
			m_FileLock->Release();
			Unlock();
			return false;
		}
	}

	// Set name, file name and path
	m_Name = StripFileExtension(g_basename(path));
	m_DataDir = datadir;
	m_Path = g_strdup(path);

	CConfig::SetStr("Projects", "Last", m_Path);
	CConfig::AddProjectToRecentList(ProjectType(), m_Path);
	Unlock();
	return true;
}

// Rename project
bool CProject::Rename(const gchar *newname, GError **error)
{
	char buf[MAX_PROJECT_NAME+32];

	g_return_val_if_fail(error == NULL || *error == NULL, false);

	// Check new file name
	if (!newname || newname[0]=='\0') {
		g_set_error(error, g_AppError, 0, "%s", "Please, specify a project name.");
		return false;
	}
	if (!CheckFileBaseName(newname)) {
		g_set_error(error, g_AppError, 0, "%s", "A project name must contain characters that are allowed in file name.\nDo not use: / \\ ? % * : | \" < and >");
		return false;
	}

	Lock();

	if (!m_Path || !m_Name) {
		// Cannot rename closed project
		g_set_error(error, g_AppError, 0, "%s", "Cannot rename closed project.");
		Unlock();
		return false;
	}
	if (ComparePaths(newname, m_Name)==0) {
		// Special case: same name
		Save();
		Unlock();
		return false;
	}

	sprintf(buf, "%s.%s", newname, FILE_EXTENSION_PROJECT);
	gchar *basedir = g_path_get_dirname(m_Path);
	gchar *fpath = g_build_filename(basedir, buf, NULL);
	g_free(basedir);

	if (g_file_test(fpath, G_FILE_TEST_EXISTS)) {
		// Project exists already
		g_set_error(error, g_AppError, 0, "A project file \"%s\" exists already. Please delete it first.", fpath);
		g_free(fpath);
		Unlock();
		return false;
	}
	gchar *datadir = g_strdup_printf("%s-files", fpath);
	if (g_file_test(datadir, G_FILE_TEST_EXISTS)) {
		// Data directory already exists
		g_set_error(error, g_AppError, 0, "A file or directory \"%s\" exists already. Please delete it first.", datadir);
		g_free(datadir);
		g_free(fpath);
		Unlock();
		return false;
	}

	// Acquire new lock
	FileLock *dstlock = new FileLock();
	if (!dstlock->Acquire(fpath)) {
		g_set_error(error, g_AppError, 0, "Access denined.");
		g_free(fpath);
		g_free(datadir);
		delete dstlock;
		Unlock();
		return false;
	}

	// Move project file
	if (g_rename(m_Path, fpath)!=0) {
		g_set_error(error, g_AppError, 0, "Failed to rename the project file \"%s\" to \"%s\".", m_Path, fpath);
		g_free(fpath);
		g_free(datadir);
		delete dstlock;
		Unlock();
		return false;
	}
	// Move data directory
	if (g_rename(m_DataDir, datadir)!=0) {
		g_set_error(error, g_AppError, 0, "Failed to rename the directory \"%s\" to \"%s\".", m_DataDir, datadir);
		g_free(fpath);
		g_free(datadir);
		delete dstlock;
		Unlock();
		return false;
	}

	// Set name, file name and path
	g_free(m_Name);
	m_Name = StripFileExtension(g_basename(fpath));
	g_free(m_DataDir);
	m_DataDir = datadir;
	g_free(m_Path);
	m_Path = fpath;
	delete m_FileLock;
	m_FileLock = dstlock;

	CConfig::SetStr("Projects", "Last", fpath);
	CConfig::AddProjectToRecentList(ProjectType(), m_Path);
	Unlock();
	return true;
}

// Save project as
bool CProject::Export(const gchar *dstpath, GError **error)
{
	g_return_val_if_fail(error == NULL || *error == NULL, false);

	Lock();

	if (!m_Path || !m_Name) {
		// Cannot make copy of closed project
		g_set_error(error, g_AppError, 0, "%s", "Cannot make copy of closed project.");
		Unlock();
		return false;
	}
	if (ComparePaths(dstpath, m_Path)==0) {
		// Special case: same path
		g_set_error(error, g_AppError, 0, "%s", "Cannot overwrite itself.");
		Unlock();
		return false;
	}

	// Aquire new lock
	FileLock dstlock;
	if (!dstlock.Acquire(dstpath)) {
		g_set_error(error, g_AppError, 0, "%s", "Access denined.");
		Unlock();
		return false;
	}
	
	// Copy project file
	if (!write(dstpath, error)) {
		Unlock();
		return false;
	}

	// Copy data files
	gchar *datadir = g_strdup_printf("%s-files", dstpath);
	if (!CopyFiles(datadir, error)) {
		g_free(datadir);
		Unlock();
		return false;
	}

	Unlock();
	return true;
}

// Flush project to disk
void CProject::Save(void) const
{
	Lock();
	if (!m_ReadOnly && m_Path) {
		// Save project file
		write(m_Path, NULL);
	}
	Unlock();
}

// Import data from old C-Munipack
bool CProject::Import(const gchar *srcdir, GError **error)
{
	bool retval = true;
	int state;
	gchar *ftemp, *fphot;
	GtkTreeIter iter;

	g_return_val_if_fail(error == NULL || *error == NULL, false);

	Lock();
	if (!m_Path || !m_Name) {
		// Cannot make copy of closed project
		g_set_error(error, g_AppError, 0, "%s", "Cannot import data into a closed project.");
		Unlock();
		return false;
	}

	reset();

	// Read a project file
	gchar *fpath = g_build_filename(srcdir, "project.ini", NULL);
	if (!read(fpath, true, error)) {
		g_free(fpath);
		Unlock();
		return false;
	}
	g_free(fpath);

	// Make sure that the data directory exists
	if (!force_directory(m_DataDir)) {
		g_set_error(error, g_AppError, 0, "Failed to create the directory: %s", m_DataDir);
		Unlock();
		return false;
	}

	// Copy images and photometry files
	gboolean ok = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter);
	while (ok && retval) {
		gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter,	FRAME_STATE, &state, 
			FRAME_TEMPFILE, &ftemp, FRAME_PHOTFILE, &fphot, -1);
		if (retval && ftemp && (state & CFILE_CONVERSION)!=0) {
			gchar *srcpath = g_build_filename(srcdir, ftemp, NULL);
			gchar *dstpath = g_build_filename(m_DataDir, ftemp, NULL);
			retval = copy_file(srcpath, dstpath, false, error);
			g_free(srcpath);
			g_free(dstpath);
		}
		if (retval && fphot && (state & CFILE_PHOTOMETRY)!=0) {
			gchar *srcpath = g_build_filename(srcdir, fphot, NULL);
			gchar *dstpath = g_build_filename(m_DataDir, fphot, NULL);
			retval = copy_file(srcpath, dstpath, false, error);
			g_free(srcpath);
			g_free(dstpath);
		}
		g_free(ftemp);
		g_free(fphot);
		ok = gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter);
	}

	// Copy temporary correction and reference files
	if (retval && m_TempBiasFile.Valid()) {
		gchar *srcpath = g_build_filename(srcdir, g_basename(m_TempBiasFile.FullPath()), NULL);
		gchar *dstpath = g_build_filename(m_DataDir, g_basename(m_TempBiasFile.FullPath()), NULL);
		if (copy_file(srcpath, dstpath, false, error))
			m_TempBiasFile = CFileInfo(dstpath);
		else
			m_TempBiasFile.Clear();
		g_free(dstpath);
		g_free(srcpath);
	}
	if (retval && m_TempDarkFile.Valid()) {
		gchar *srcpath = g_build_filename(srcdir, g_basename(m_TempDarkFile.FullPath()), NULL);
		gchar *dstpath = g_build_filename(m_DataDir, g_basename(m_TempDarkFile.FullPath()), NULL);
		if (copy_file(srcpath, dstpath, false, error)) 
			m_TempDarkFile = CFileInfo(dstpath);
		else
			m_TempDarkFile.Clear();
		g_free(dstpath);
		g_free(srcpath);
	}
	if (retval && m_OrigFlatFile.Valid() && m_TempFlatFile.Valid()) {
		gchar *srcpath = g_build_filename(srcdir, g_basename(m_TempFlatFile.FullPath()), NULL);
		gchar *dstpath = g_build_filename(m_DataDir, g_basename(m_TempFlatFile.FullPath()), NULL);
		if (copy_file(srcpath, dstpath, false, error)) 
			m_TempFlatFile = CFileInfo(dstpath);
		else
			m_TempFlatFile.Clear();
		g_free(dstpath);
		g_free(srcpath);
	}
	if (retval && m_RefType==REF_CATALOG_FILE && m_TempCatFile.Valid()) {
		gchar *srcpath = g_build_filename(srcdir, g_basename(m_TempCatFile.FullPath()), NULL);
		gchar *dstpath = g_build_filename(m_DataDir, g_basename(m_TempCatFile.FullPath()), NULL);
		if (copy_file(srcpath, dstpath, false, error))
			m_TempCatFile = CFileInfo(dstpath);
		else
			m_TempCatFile.Clear();
		g_free(dstpath);
		g_free(srcpath);
	}

	// If fails, clear project
	if (!retval)
		reset();

	Unlock();
	return retval;
}

// Load project file
bool CProject::read(const gchar *cfg_path, bool import, GError **error)
{
	int i, count, state, id, nstars, mstars, avg_frames, sum_frames;
	double juldat, exptime, ccdtemp, tcorr, radius, offset_x, offset_y;
	unsigned flags;
	char *str, file[64], key[64], val[64], *ftemp, *fpath, tag[MAX_TAG_SIZE+1];
	char *filter, *result, *fphot, *fbias, *fdark, *fflat;
	GtkTreeIter iter;
	GError *err = NULL;

	reset();
	m_Profile.Clear();

	// Load file
	GKeyFile *cfg = g_key_file_new();
	if (!g_key_file_load_from_file(cfg, cfg_path, G_KEY_FILE_NONE, error)) {
		g_key_file_free(cfg);
		return false;
	}

	if (!import) {
		// Check file format
		gchar *start = g_key_file_get_start_group(cfg);
		if (StrCmp0(start, "CMUNIPACK-PROJECT")!=0) {
			g_set_error(error, g_AppError, 0, "The file is not a valid project file: %s", cfg_path);
			g_free(start);
			g_key_file_free(cfg);
			return false;
		}
		g_free(start);
		// Check file version
		int revision = g_key_file_get_integer(cfg, "CMUNIPACK-PROJECT", "Revision", NULL);
		if (revision!=1) {
			g_set_error(error, g_AppError, 0, "Unsupported revision of the project file: %d", revision);
			g_key_file_free(cfg);
			return false;
		}
	} else {
		// Old project does not support revisions, use one of mandatory keys
		GError *e = NULL;
		g_key_file_get_integer(cfg, "PROJECT", "files", &e);
		if (e) {
			g_set_error(error, g_AppError, 0, "The file is not a valid project file: %s", cfg_path);
			g_error_free(e);
		}
	}

	// Input files
	count = g_key_file_get_integer(cfg, "PROJECT", "files", NULL);
	for (i=1; i<=count; i++) {
		sprintf(file, "FILE%d", i);
		fpath = g_key_file_get_string(cfg, file, "path", NULL);
		id = g_key_file_get_integer(cfg, file, "fileid", NULL);
		if (fpath && id>0) {
			gtk_list_store_append(m_Frames, &iter);
			juldat = g_key_file_get_double(cfg, file, "juldat", NULL);
			state = g_key_file_get_integer(cfg, file, "state", NULL);
			flags = (unsigned)g_key_file_get_integer(cfg, file, "flags", NULL);
			tcorr = g_key_file_get_double(cfg, file, "tcorr", NULL);
			exptime = g_key_file_get_double(cfg, file, "exptime", &err);
			if (err) {
				exptime = -1.0;
				g_error_free(err);
				err = NULL;
			}
			ccdtemp = g_key_file_get_double(cfg, file, "ccdtemp", &err);
			if (err) {
				ccdtemp = -999.9;
				g_error_free(err);
				err = NULL;
			}
			avg_frames = g_key_file_get_integer(cfg, file, "avg_frames", NULL);
			avg_frames = MIN(1, avg_frames);
			sum_frames = g_key_file_get_integer(cfg, file, "sum_frames", NULL);
			sum_frames = MIN(1, sum_frames);
			ftemp = g_key_file_get_string(cfg, file, "temp", NULL);
			fphot = g_key_file_get_string(cfg, file, "phot", NULL);
			fbias = g_key_file_get_string(cfg, file, "bias", NULL);
			fdark = g_key_file_get_string(cfg, file, "dark", NULL);
			fflat = g_key_file_get_string(cfg, file, "flat", NULL);
			filter = g_key_file_get_string(cfg, file, "filter", NULL);
			result = g_key_file_get_string(cfg, file, "result", NULL);
			nstars = g_key_file_get_integer(cfg, file, "nstars", NULL);
			mstars = g_key_file_get_integer(cfg, file, "mstars", NULL);
			offset_x = g_key_file_get_double(cfg, file, "offsetx", NULL);
			offset_y = g_key_file_get_double(cfg, file, "offsety", NULL);
			state &= (~CFILE_THUMBNAIL);
			sprintf(val, "%d", id);
			gtk_list_store_set(m_Frames, &iter, FRAME_STATE, state,
				FRAME_FLAGS, flags, FRAME_ID, id, FRAME_JULDAT, juldat, 
				FRAME_REPORT, result, FRAME_ORIGFILE, fpath, FRAME_BIASFILE, fbias, 
				FRAME_DARKFILE, fdark, FRAME_FLATFILE, fflat, FRAME_FILTER, filter, 
				FRAME_TIMECORR, tcorr, FRAME_EXPTIME, exptime, FRAME_CCDTEMP, ccdtemp, 
				FRAME_STARS, nstars, FRAME_MSTARS, mstars, FRAME_TEMPFILE, ftemp, 
				FRAME_OFFSET_X, offset_x, FRAME_OFFSET_Y, offset_y, 
				FRAME_PHOTFILE, fphot, FRAME_AVGFRAMES, avg_frames, 
				FRAME_SUMFRAMES, sum_frames, FRAME_STRINGID, val, -1);
			if (id>=m_FreeID)
				m_FreeID = id+1;
			g_free(fflat);
			g_free(fdark);
			g_free(fbias);
			g_free(filter);
			g_free(result);
			g_free(fpath);
			g_free(ftemp);
			g_free(fphot);
		}
	}

	// Apertures
	count = g_key_file_get_integer(cfg, "APERTURES", "count", NULL);
	for (int i=1; i<=count; i++) {
		sprintf(key, "aperture%d", i);
		str = g_key_file_get_string(cfg, "APERTURES", key, NULL);
		if (sscanf(str, " %9d; %32lf ", &id, &radius)==2)
			m_Apertures.Add(CAperture(id, radius));
		g_free(str);
	}
	
	// Selected stars
	str = g_key_file_get_string(cfg, "STARS", "var", NULL);
	m_Selection.SetStarList(CMPACK_SELECT_VAR, str);
	g_free(str);
	str = g_key_file_get_string(cfg, "STARS", "comp", NULL);
	m_Selection.SetStarList(CMPACK_SELECT_COMP, str);
	g_free(str);
	str = g_key_file_get_string(cfg, "STARS", "check", NULL);
	m_Selection.SetStarList(CMPACK_SELECT_CHECK, str);
	g_free(str);

	// Saved object selections
	int index = 1;
	count = g_key_file_get_integer(cfg, "SelectionList", "count", NULL);
	for (int i=1; i<=count; i++) {
		sprintf(key, "item%d", i);
		gchar *str = g_key_file_get_string(cfg, "SelectionList", key, NULL);
		sprintf(key, "name%d", i);
		gchar *name = g_key_file_get_string(cfg, "SelectionList", key, NULL);
		if (name)
			name = g_strstrip(name);
		CSelection stars(str);
		if (stars.Count()>0 && !m_SelectionList.Value(name)) {
			while (!name) {
				gchar aux[128];
				sprintf(aux, "Selection %d", index);
				if (!m_SelectionList.Value(aux)) 
					name = g_strstrip(g_strdup(aux));
				index++;
			}
			m_SelectionList.Set(name, stars);
		}
		g_free(str);
		g_free(name);
	}
	
	// Object tags
	count = g_key_file_get_integer(cfg, "TAGS", "count", NULL);
	for (int i=1; i<=count; i++) {
		sprintf(key, "tag%d", i);
		str = g_key_file_get_string(cfg, "TAGS", key, NULL);
		if (sscanf(str, " %9d; %1023[^\n] ", &id, tag)==2) {
			if (id>=0 && *tag!='\0') {
				gchar *aux = g_strdup(tag);
				m_Tags.Set(id, g_strstrip(aux));
				g_free(aux);
			}
		}
		g_free(str);
	}

	// Object coordinates
	str = g_key_file_get_string(cfg, "OBJECT", "name", NULL);
	if (str)
		m_Coords.SetName(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "OBJECT", "ra", NULL);
	if (str)
		m_Coords.SetRA(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "OBJECT", "dec", NULL);
	if (str)
		m_Coords.SetDec(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "OBJECT", "source", NULL);
	if (str)
		m_Coords.SetSource(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "OBJECT", "remarks", NULL);
	if (str)
		m_Coords.SetRemarks(g_strstrip(str));
	g_free(str);

	// Observer's name
	m_Observer = g_key_file_get_string(cfg, "OBSERVER", "name", NULL);
	if (m_Observer)
		m_Observer = g_strstrip(m_Observer);

	// Reference location
	str = g_key_file_get_string(cfg, "LOCATION", "name", NULL);
	if (str)
		m_Location.SetName(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "LOCATION", "lon", NULL);
	if (str)
		m_Location.SetLon(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "LOCATION", "lat", NULL);
	if (str)
		m_Location.SetLat(g_strstrip(str));
	g_free(str);
	str = g_key_file_get_string(cfg, "LOCATION", "remarks", NULL);
	if (str)
		m_Location.SetComment(g_strstrip(str));
	g_free(str);

	// Common
	m_RefType = (tReferenceType)g_key_file_get_integer(cfg, "PROJECT", "reftype", NULL);

	// Reference frame
	m_RefID = g_key_file_get_integer(cfg, "PROJECT", "refid", NULL);
	GtkTreePath *path = GetFilePath(m_RefID);
	if (path) {
		m_Ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(m_Frames), path);
		gtk_tree_path_free(path);
	}

	m_OrigBiasFile.Load(cfg, "BIAS", "orig");
	m_TempBiasFile.Load(cfg, "BIAS", "temp");
	m_OrigDarkFile.Load(cfg, "DARK", "orig");
	m_TempDarkFile.Load(cfg, "DARK", "temp");
	m_OrigFlatFile.Load(cfg, "FLAT", "orig");
	m_TempFlatFile.Load(cfg, "FLAT", "temp");
	m_OrigCatFile.Load(cfg, "REF", "orig");
	m_TempCatFile.Load(cfg, "REF", "temp");

	// Profile
	m_Profile.Load(cfg);

	// User settings
	gchar **keys = g_key_file_get_keys(cfg, "ProjectSettings", NULL, NULL);
	if (keys) {
		for (int j=0; keys[j]!=NULL; j++) {
			gchar *value = g_key_file_get_value(cfg, "ProjectSettings", keys[j], NULL);
			if (value && value[0]!='\0') {
				char *key = strchr(keys[j], '.');
				if (key) {
					*key++ = '\0';
					g_key_file_set_value(m_Params, keys[j], key, value);
				}
			}
			g_free(value);
		}
		g_strfreev(keys);
	}
	
	g_key_file_free(cfg);
	return true;
}

// Save project to file
bool CProject::write(const gchar *cfg_path, GError **error) const
{
	gboolean ok;
	gint i, id, count, avg_frames, sum_frames, nstars, mstars, state;
	gchar file[64], *fpath, *ftemp, *fbias, *fdark, *fflat;
	gchar *str, *filter, *result, *fphot, *data;
	unsigned flags;
	const gchar *cstr, *tag;
	gsize datalen;
	GtkTreeIter iter;
	gdouble juldat, ccdtemp, exptime, tcorr, offset_x, offset_y;
	GKeyFile *cfg = g_key_file_new();

	// Identification
	g_key_file_set_integer(cfg, "CMUNIPACK-PROJECT", "Revision", 1);

	// Profile
	m_Profile.Save(cfg);

	// Dialog settings
	gchar **grps = g_key_file_get_groups(m_Params, NULL);
	if (grps) {
		for (int i=0; grps[i]!=NULL; i++) {
			gchar **keys = g_key_file_get_keys(m_Params, grps[i], NULL, NULL);
			if (keys) {
				for (int j=0; keys[j]!=NULL; j++) {
					gchar *value = g_key_file_get_value(m_Params, grps[i], keys[j], NULL);
					if (value && value[0]!='\0') {
						char key[256];
						sprintf(key, "%s.%s", grps[i], keys[j]);
						g_key_file_set_value(cfg, "ProjectSettings", key, value);
					}
					g_free(value);
				}
				g_strfreev(keys);
			}
		}
		g_strfreev(grps);
	}

	// Input files
	count = 0;
	ok = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter);
	while (ok) {
		sprintf(file, "FILE%d", ++count);
		gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter, FRAME_ORIGFILE, &fpath, FRAME_ID, &id, 
			FRAME_STATE, &state, FRAME_FLAGS, &flags, FRAME_JULDAT, &juldat,
			FRAME_REPORT, &result, FRAME_BIASFILE, &fbias, FRAME_DARKFILE, &fdark, 
			FRAME_FLATFILE, &fflat, FRAME_FILTER, &filter, FRAME_TIMECORR, &tcorr, 
			FRAME_EXPTIME, &exptime, FRAME_CCDTEMP, &ccdtemp, FRAME_STARS, &nstars, 
			FRAME_MSTARS, &mstars, FRAME_TEMPFILE, &ftemp, FRAME_OFFSET_X, &offset_x, 
			FRAME_OFFSET_Y, &offset_y, FRAME_PHOTFILE, &fphot, FRAME_AVGFRAMES, &avg_frames, 
			FRAME_SUMFRAMES, &sum_frames, -1);
		g_key_file_set_string(cfg, file, "path", fpath);
		g_key_file_set_integer(cfg, file, "fileid", id);
		g_key_file_set_integer(cfg, file, "state", state & (~CFILE_THUMBNAIL));
		g_key_file_set_integer(cfg, file, "flags", (int)flags);
		if (ftemp)
			g_key_file_set_string(cfg, file, "temp", ftemp);
		if (fphot)
			g_key_file_set_string(cfg, file, "phot", fphot);
		if (juldat>0)
			g_key_file_set_double(cfg, file, "juldat", juldat);
		if (tcorr>0)
			g_key_file_set_double(cfg, file, "tcorr", tcorr);
		if (filter)
			g_key_file_set_string(cfg, file, "filter", filter);
		if (exptime>=0)
			g_key_file_set_double(cfg, file, "exptime", exptime);
		if (ccdtemp>-999.0)
			g_key_file_set_double(cfg, file, "ccdtemp", ccdtemp);
		if (avg_frames>1)
			g_key_file_set_integer(cfg, file, "avg_frames", avg_frames);
		if (sum_frames>1)
			g_key_file_set_integer(cfg, file, "sum_frames", sum_frames);
		if (fbias)
			g_key_file_set_string(cfg, file, "bias", fbias);
		if (fdark)
			g_key_file_set_string(cfg, file, "dark", fdark);
		if (fflat)
			g_key_file_set_string(cfg, file, "flat", fflat);
		if (nstars>0)
			g_key_file_set_integer(cfg, file, "nstars", nstars);
		if (mstars>0)
			g_key_file_set_integer(cfg, file, "mstars", mstars);
		if (offset_x!=0)
			g_key_file_set_double(cfg, file, "offsetx", offset_x);
		if (offset_y!=0)
			g_key_file_set_double(cfg, file, "offsety", offset_y);
		if (result) 
			g_key_file_set_string(cfg, file, "result", result);
		g_free(filter);
		g_free(ftemp);
		g_free(fbias);
		g_free(fdark);
		g_free(fflat);
		g_free(fpath);
		g_free(fphot);
		g_free(result);
		ok = gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter);
	}
	g_key_file_set_integer(cfg, "PROJECT", "files", count);

	// Apertures
	g_key_file_set_integer(cfg, "APERTURES", "count", m_Apertures.Count());
	for (i=0; i<m_Apertures.Count(); i++) {
		const CAperture *aper = m_Apertures.Get(i);
		gchar key[64], val[64];
		sprintf(key, "aperture%d", i+1);
		sprintf(val, "%d;%.5f", aper->Id(), aper->Radius());
		g_key_file_set_string(cfg, "APERTURES", key, val);
	}
	
	// Selected stars
	str = m_Selection.GetStarList(CMPACK_SELECT_VAR);
	if (str)
		g_key_file_set_string(cfg, "STARS", "var", str);
	g_free(str);
	str = m_Selection.GetStarList(CMPACK_SELECT_COMP);
	if (str)
		g_key_file_set_string(cfg, "STARS", "comp", str);
	g_free(str);
	str = m_Selection.GetStarList(CMPACK_SELECT_CHECK);
	if (str)
		g_key_file_set_string(cfg, "STARS", "check", str);
	g_free(str);

	// Selected stars
	count = m_SelectionList.Count();
	g_key_file_set_integer(cfg, "SelectionList", "count", count);
	for (int i=0; i<count; i++) {
		gchar key[64], *str = m_SelectionList.At(i)->toString();
		sprintf(key, "name%d", i+1);
		g_key_file_set_string(cfg, "SelectionList", key, m_SelectionList.Name(i));
		sprintf(key, "item%d", i+1);
		g_key_file_set_string(cfg, "SelectionList", key, str);
		g_free(str);
	}

	// Tags
	g_key_file_set_integer(cfg, "TAGS", "count", m_Tags.Count());
	CTags::tIter tagiter;
	ok = m_Tags.First(&tagiter, &id, &tag);
	i = 0;
	while (ok) {
		gchar key[64], val[MAX_TAG_SIZE+64];
		sprintf(key, "tag%d", ++i);
		sprintf(val, "%d;%s", id, tag);
		g_key_file_set_string(cfg, "TAGS", key, val);
		ok = m_Tags.Next(&tagiter, &id, &tag);
	}

	// Object coordinates
	cstr = m_Coords.Name();
	if (cstr)
		g_key_file_set_string(cfg, "OBJECT", "name", cstr);
	cstr = m_Coords.RA();
	if (cstr)
		g_key_file_set_string(cfg, "OBJECT", "ra", cstr);
	cstr = m_Coords.Dec();
	if (cstr)
		g_key_file_set_string(cfg, "OBJECT", "dec", cstr);
	cstr = m_Coords.Source();
	if (cstr)
		g_key_file_set_string(cfg, "OBJECT", "source", cstr);
	cstr = m_Coords.Remarks();
	if (cstr)
		g_key_file_set_string(cfg, "OBJECT", "remarks", cstr);

	// Reference observer's name
	if (m_Observer)
		g_key_file_set_string(cfg, "OBSERVER", "name", m_Observer);

	// Reference location
	cstr = m_Location.Name();
	if (cstr)
		g_key_file_set_string(cfg, "LOCATION", "name", cstr);
	cstr = m_Location.Lon();
	if (cstr)
		g_key_file_set_string(cfg, "LOCATION", "lon", cstr);
	cstr = m_Location.Lat();
	if (cstr)
		g_key_file_set_string(cfg, "LOCATION", "lat", cstr);
	cstr = m_Location.Comment();
	if (cstr)
		g_key_file_set_string(cfg, "LOCATION", "remarks", cstr);

	// Common
	g_key_file_set_integer(cfg, "PROJECT", "refid", m_RefID);
	g_key_file_set_integer(cfg, "PROJECT", "reftype", m_RefType);
	m_OrigBiasFile.Save(cfg, "BIAS", "orig");
	m_TempBiasFile.Save(cfg, "BIAS", "temp");
	m_OrigDarkFile.Save(cfg, "DARK", "orig");
	m_TempDarkFile.Save(cfg, "DARK", "temp");
	m_OrigFlatFile.Save(cfg, "FLAT", "orig");
	m_TempFlatFile.Save(cfg, "FLAT", "temp");
	m_OrigCatFile.Save(cfg, "REF", "orig");
	m_TempCatFile.Save(cfg, "REF", "temp");

	// Save content to file
	ok = FALSE;
	data = g_key_file_to_data(cfg, &datalen, error);
	if (data) {
		GIOChannel *pFile = g_io_channel_new_file(cfg_path, "w+", error);
		if (pFile) {
			g_io_channel_write_chars(pFile, data, datalen, NULL, error);
			g_io_channel_close(pFile);
			ok = TRUE;
		}
		g_free(data);
	}

	g_key_file_free(cfg);
	return ok!=0;
}

// Return true if the given buffer can be a project file
bool CProject::isProjectFile(const gchar *filepath, tProjectType *type)
{
	bool retval = false;
	GKeyFile *f = g_key_file_new();
	if (g_key_file_load_from_file(f, filepath, G_KEY_FILE_NONE, NULL)) 
		retval = checkProjectFile(f, type);
	g_key_file_free(f);
	return retval;
}

// Get number of files
int CProject::GetFileCount(void)
{ 
	int retval;
	Lock();
	retval = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(m_Frames), NULL);
	Unlock();
	return retval;
}

// Remove all frames from the project
void CProject::RemoveAllFrames(void)
{
	Lock();
	if (m_DataDir && !m_ReadOnly) {
		ClearTempFiles();
		ClearReference();
		ClearCorrections();
		gtk_list_store_clear(m_Frames);
		m_FreeID = 1;
		m_Coords.Clear();
		m_Apertures.Clear();
		g_key_file_free(m_Params);
		m_Params = g_key_file_new();
	}
	Unlock();
}

// Clear all internal parameters
void CProject::reset(void)
{
	Lock();
	ClearReference();
	ClearCorrections();
	gtk_list_store_clear(m_Frames);
	m_FreeID = 1;
	m_Coords.Clear();
	m_Apertures.Clear();
	g_key_file_free(m_Params);
	m_Params = g_key_file_new();
	Unlock();
}

void CProject::ClearCorrections(void)
{
	Lock();
	SetDbl("TimeCorr", "Seconds", 0);
	m_OrigBiasFile.Clear();
	m_TempBiasFile.Clear();
	m_OrigDarkFile.Clear();
	m_TempDarkFile.Clear();
	m_OrigFlatFile.Clear();
	m_TempFlatFile.Clear();
	Unlock();
}

void CProject::ClearObject(void)
{
	Lock();
	m_Coords.Clear();
	Unlock();
}

GtkTreePath *CProject::AddFile(const CFrameInfo &info)
{
	GtkTreePath *path = NULL;
	char buf[128];
	GtkTreeIter iter;

	Lock();
	if (info.Valid() && !m_ReadOnly) {
		int fileid = m_FreeID++;
		sprintf(buf, "%d", fileid);
		gtk_list_store_append(m_Frames, &iter);
		gtk_list_store_set(m_Frames, &iter, FRAME_ID, fileid, 
			FRAME_STRINGID, buf, FRAME_ORIGFILE, info.FullPath(), 
			FRAME_JULDAT, info.JulDat(), FRAME_EXPTIME, info.ExpTime(),
			FRAME_CCDTEMP, info.CCDTemp(), FRAME_AVGFRAMES, info.AvgFrames(),
			FRAME_SUMFRAMES, info.SumFrames(), FRAME_FILTER, info.Filter(), -1);
		path = gtk_tree_model_get_path(GTK_TREE_MODEL(m_Frames), &iter);
	}
	Unlock();
	return path;
}

// Returns true if the file is already in the project
bool CProject::FileInProject(const char *filepath)
{
	bool ok, retval = false;
	gchar *path;
	GtkTreeIter iter;

	Lock();
	ok = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter)!=0;
	while (ok && !retval) {
		gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter, FRAME_ORIGFILE, &path, -1);
		if (ComparePaths(filepath, path)==0)
			retval = true;
		g_free(path);
		ok = gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter)!=0;
	}
	Unlock();
	return retval;
}

//
// Returns true if the given frame is a reference frame
//
bool CProject::IsReferenceFrame(GtkTreePath *path)
{
	bool retval = false;

	Lock();
	GtkTreePath *refpath = GetReferencePath();
	if (refpath) {
		retval = gtk_tree_path_compare(refpath, path)==0;
		gtk_tree_path_free(refpath);
	}
	Unlock();
	return retval;
}

//
// Returns true if the given list of tree paths contains a reference frame
//
bool CProject::ContainsReferenceFrame(GList *list)
{
	bool retval = false;

	Lock();
	if (m_RefType==REF_FRAME) {
		GtkTreePath *refpath = GetReferencePath();
		if (refpath) {
			for (GList *ptr=list; ptr!=NULL; ptr=ptr->next) {
				if (gtk_tree_path_compare(refpath, (GtkTreePath*)ptr->data)==0) {
					retval = true;
					break;
				}
			}
			gtk_tree_path_free(refpath);
		}
	}
	Unlock();
	return retval;
}

//
// Remove single file
//
void CProject::RemovePath(GtkTreePath *pPath)
{
	int state;
	GtkTreeIter iter;
	gchar *ftemp, *fphot;

	Lock();
	if (m_DataDir && !m_ReadOnly && gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath)) {
		if (IsReferenceFrame(pPath)) 
			ClearReference();
		// Remove temporary files
		gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter, FRAME_STATE, &state,
			FRAME_TEMPFILE, &ftemp, FRAME_PHOTFILE, &fphot, -1);
		// Delete temporary CCD image and photometry file
		if (ftemp && (state & CFILE_CONVERSION)!=0) {
			gchar *path = g_build_filename(m_DataDir, ftemp, NULL);
			g_remove(path);
			g_free(path);
		}
		if (fphot && (state & CFILE_PHOTOMETRY)!=0) {
			gchar *path = g_build_filename(m_DataDir, fphot, NULL);
			g_remove(path);
			g_free(path);
		}
		g_free(ftemp);
		g_free(fphot);
		// Remove the frame from the tree model
		gtk_list_store_remove(m_Frames, &iter);
		// If this was the last frame, reset internal parameters
		if (gtk_tree_model_iter_n_children(GTK_TREE_MODEL(m_Frames), NULL)==0) 
			reset();
	}
	Unlock();
}

void CProject::RemoveFrame(int frame_id)
{
	Lock();
	GtkTreePath *path = GetFilePath(frame_id);
	if (path) {
		RemovePath(path);
		gtk_tree_path_free(path);
	}
	Unlock();
}

void CProject::RemoveFiles(GList *pPaths)
{
	Lock();
	if (pPaths && !m_ReadOnly) {
		GSList *rr_list = NULL;
		for (GList *ptr = pPaths; ptr!=NULL; ptr=ptr->next) {
			GtkTreeRowReference *ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(m_Frames), (GtkTreePath*)ptr->data);
			rr_list = g_slist_prepend(rr_list, ref);
		}
		for (GSList *node = rr_list; node != NULL; node = node->next) {
			GtkTreePath *path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)node->data);
			if (path) {
				RemovePath(path);
				gtk_tree_path_free(path);
			}
		}
		g_slist_foreach(rr_list, (GFunc)gtk_tree_row_reference_free, NULL);
		g_slist_free(rr_list);
	}
	Unlock();
}


//
// Clear the reference file
//
void CProject::ClearReference(void)
{
	Lock();
	m_RefID = -1;
	gtk_tree_row_reference_free(m_Ref);
	m_Ref = NULL;
	m_RefType = REF_UNDEFINED;
	m_OrigCatFile.Clear();
	m_Selection.Clear();
	m_SelectionList.Clear();
	m_Tags.Clear();
	m_Location.Clear();
	m_Coords.Clear();
	g_free(m_Observer);
	m_Observer = NULL;
	Unlock();
}


//
// Delete all temporary files
//
void CProject::ClearTempFiles(void)
{
	gboolean ok;
	int state;
	char *ftemp, *fphot;
	GtkTreeIter iter;

	Lock();
	if (m_DataDir && !m_ReadOnly) {
		ok = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter);
		while (ok) {
			gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter,	FRAME_STATE, &state, 
				FRAME_TEMPFILE, &ftemp, FRAME_PHOTFILE, &fphot, -1);
			// Delete temporary CCD image and photometry file
			if (ftemp && (state & CFILE_CONVERSION)!=0) {
				gchar *path = g_build_filename(m_DataDir, ftemp, NULL);
				g_remove(path);
				g_free(path);
			}
			if (fphot && (state & CFILE_PHOTOMETRY)!=0) {
				gchar *path = g_build_filename(m_DataDir, fphot, NULL);
				g_remove(path);
				g_free(path);
			}
			g_free(ftemp);
			g_free(fphot);
			// We have to mark all files as new
			gtk_list_store_set(m_Frames, &iter, FRAME_STATE, 0, 
				FRAME_JULDAT, 0.0, FRAME_TEMPFILE, NULL, FRAME_PHOTFILE, NULL, 
				FRAME_REPORT, NULL, FRAME_BIASFILE, NULL, FRAME_DARKFILE, NULL, 
				FRAME_FLATFILE, NULL, FRAME_FILTER, NULL, FRAME_TIMECORR, 0.0, 
				FRAME_EXPTIME, -1.0, FRAME_AVGFRAMES, 0, FRAME_SUMFRAMES, 0, 
				FRAME_CCDTEMP, -999.9, FRAME_STARS, 0, FRAME_MSTARS, 0, 
				FRAME_OFFSET_X, 0.0, FRAME_OFFSET_Y, 0.0, FRAME_THUMBNAIL, NULL, -1);
			ok = gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter);
		}
		// Copy temporary correction and reference files
		if (m_TempBiasFile.Valid()) {
			g_remove(m_TempBiasFile.FullPath());
			m_TempBiasFile.Clear();
		}
		if (m_TempDarkFile.Valid()) {
			g_remove(m_TempDarkFile.FullPath());
			m_TempDarkFile.Clear();
		}
		if (m_TempFlatFile.Valid()) {
			g_remove(m_TempFlatFile.FullPath());
			m_TempFlatFile.Clear();
		}
		if (m_RefType==REF_CATALOG_FILE && m_TempCatFile.Valid()) {
			g_remove(m_TempCatFile.FullPath());
			m_TempCatFile.Clear();
		}
	}
	Unlock();
}


//
// Clear thumbnails
//
void CProject::ClearThumbnails(void)
{
	int state;
	gboolean ok;
	GtkTreeIter iter;

	Lock();
	ok = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter);
	while (ok) {
		gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter, FRAME_STATE, &state, -1);
		state &= ~CFILE_THUMBNAIL;
		gtk_list_store_set(m_Frames, &iter, FRAME_STATE, state, FRAME_THUMBNAIL, NULL, -1);
		ok = gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter);
	}
	Unlock();
}

//
// Set reference frame
//
void CProject::SetReference(GtkTreePath *path, const gchar *cat_file)
{
	Lock();
	gtk_tree_row_reference_free(m_Ref);
	m_RefID = -1;
	m_Ref = NULL;
	m_OrigCatFile.Clear();
	if (path) {
		m_RefType = REF_FRAME;
		m_RefID = GetFrameID(path);
		m_Ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(m_Frames), path);
	} else if (cat_file) {
		m_RefType = REF_CATALOG_FILE;
		m_OrigCatFile = CFileInfo(cat_file);
	} else {
		m_RefType = REF_UNDEFINED;
	}
	Unlock();
}

//
// Get reference frame
//
GtkTreePath *CProject::GetReferencePath(void)
{
	GtkTreePath *path;
	Lock();
	path = (m_Ref ? gtk_tree_row_reference_get_path(m_Ref) : NULL);
	Unlock();
	return path;
}

//
// Get tree path to frame 
//
GtkTreePath *CProject::GetFilePath(int id)
{
	tGetFileInfo info;

	if (id>0) {
		info.id = id;
		info.path = NULL;
		Lock();
		gtk_tree_model_foreach(GTK_TREE_MODEL(m_Frames), FindFile, &info);
		Unlock();
		return info.path;
	}
	return NULL;
}

// Get tree path to the first file
GtkTreePath *CProject::GetFirstFile(void)
{
	GtkTreePath *path;
	GtkTreeIter iter;

	Lock();
	if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter))
		path = gtk_tree_model_get_path(GTK_TREE_MODEL(m_Frames), &iter);
	else
		path = NULL;
	Unlock();

	return path;
}

// Get tree path to the last file
GtkTreePath *CProject::GetLastFile(void)
{
	int count;
	GtkTreePath *path = NULL;
	GtkTreeIter iter;

	Lock();
	count = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(m_Frames), NULL);
	if (count>0) {
		gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(m_Frames), &iter, NULL, count-1);
		path = gtk_tree_model_get_path(GTK_TREE_MODEL(m_Frames), &iter);
	} 
	Unlock();
	return path;
}

// Get path to the previous file
GtkTreePath *CProject::GetPreviousFile(GtkTreePath *path)
{
	GtkTreePath *retval;

	Lock();
	GtkTreePath *pPrev = gtk_tree_path_copy(path);
	if (gtk_tree_path_prev(pPrev))
		retval = pPrev;
	else
		retval = NULL;
	Unlock();

	return retval;
}

// Get path to the next file
GtkTreePath *CProject::GetNextFile(GtkTreePath *path)
{
	GtkTreePath *retval;
	GtkTreeIter iter;

	Lock();
	gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, path);
	if (gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter))
		retval = gtk_tree_model_get_path(GTK_TREE_MODEL(m_Frames), &iter);
	else
		retval = NULL;
	Unlock();

	return retval;
}

// Get frame state
unsigned CProject::GetState(int frame_id)
{
	unsigned state = 0;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		state = GetState(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return state;
}
unsigned CProject::GetState(GtkTreePath *pPath)
{
	unsigned state = 0;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		state = GetState(&iter);
		Unlock();
	}

	return state;
}
unsigned CProject::GetState(GtkTreeIter *pIter)
{
	unsigned state = 0;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_STATE, &state, -1);
	Unlock();

	return state;
}

// Get frame flags
unsigned CProject::GetFlags(int frame_id)
{
	unsigned state = 0;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		state = GetFlags(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return state;
}
unsigned CProject::GetFlags(GtkTreePath *pPath)
{
	unsigned state = 0;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		state = GetFlags(&iter);
		Unlock();
	}

	return state;
}
unsigned CProject::GetFlags(GtkTreeIter *pIter)
{
	unsigned state = 0;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_FLAGS, &state, -1);
	Unlock();

	return state;
}

// Get date of observation
double CProject::GetJulDate(int frame_id)
{
	double jd = 0;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		jd = GetJulDate(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return jd;
}
double CProject::GetJulDate(GtkTreePath *pPath)
{
	double jd = 0;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		jd = GetJulDate(&iter);
		Unlock();
	}

	return jd;
}
double CProject::GetJulDate(GtkTreeIter *pIter)
{
	double jd = 0;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_JULDAT, &jd, -1);
	Unlock();

	return jd;
}

// Get color filter
gchar *CProject::GetColorFilter(int frame_id)
{
	gchar *retval = NULL;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		retval = GetColorFilter(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return retval;
}
gchar *CProject::GetColorFilter(GtkTreePath *pPath)
{
	gchar *retval = NULL;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		retval = GetColorFilter(&iter);
		Unlock();
	}

	return retval;
}
gchar *CProject::GetColorFilter(GtkTreeIter *pIter)
{
	gchar *retval;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_FILTER, &retval, -1);
	Unlock();

	return retval;
}

// Get exposure duration
double CProject::GetExposure(int frame_id)
{
	double exptime = -1.0;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		exptime = GetExposure(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return exptime;
}
double CProject::GetExposure(GtkTreePath *pPath)
{
	double exptime = -1.0;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		exptime = GetExposure(&iter);
		Unlock();
	}

	return exptime;
}
double CProject::GetExposure(GtkTreeIter *pIter)
{
	double exptime = -1.0;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_EXPTIME, &exptime, -1);
	Unlock();

	return exptime;
}

// Get frame ID
int CProject::GetFrameID(GtkTreePath *pPath)
{
	int frame_id = 0;

	GtkTreeIter iter;
	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		frame_id = GetFrameID(&iter);
		Unlock();
	}
	return frame_id;
}
int CProject::GetFrameID(GtkTreeIter *pIter)
{
	int id = 0;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_ID, &id, -1);
	Unlock();

	return id;
}

// Get full path to the original file
gchar *CProject::GetSourceFile(int frame_id)
{
	gchar *file = NULL;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		file = GetSourceFile(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return file;
}
gchar *CProject::GetSourceFile(GtkTreePath *pPath)
{
	gchar *file = NULL;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		file = GetSourceFile(&iter);
		Unlock();
	}

	return file;
}
gchar *CProject::GetSourceFile(GtkTreeIter *pIter)
{
	gchar *file = NULL;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_ORIGFILE, &file, -1);
	Unlock();

	return file;
}


// Get name of the temporary file
gchar *CProject::GetImageFileName(int frame_id)
{
	gchar *file = NULL;
	GtkTreePath *path;
	
	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		file = GetImageFileName(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return file;
}
gchar *CProject::GetImageFileName(GtkTreePath *pPath)
{
	gchar *file = NULL;
	GtkTreeIter iter;
	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		file = GetImageFileName(&iter);
		Unlock();
	}
	return file;
}
gchar *CProject::GetImageFileName(GtkTreeIter *pIter)
{
	gchar *file = NULL;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_TEMPFILE, &file, -1);
	Unlock();

	return file;
}


// Get full path to the temporary file
gchar *CProject::GetImageFile(int frame_id)
{
	gchar *name, *file;
	
	Lock();
	name = GetImageFileName(frame_id);
	Unlock();
	if (name) {
		file = g_build_filename(m_DataDir, name, NULL);
		g_free(name);
		return file;
	}
	return NULL;
}
gchar *CProject::GetImageFile(GtkTreePath *pPath)
{
	gchar *name, *file;

	Lock();
	name = GetImageFileName(pPath);
	Unlock();
	if (name) {
		file = g_build_filename(m_DataDir, name, NULL);
		g_free(name);
		return file;
	}
	return NULL;
}
gchar *CProject::GetImageFile(GtkTreeIter *pIter)
{
	gchar *name, *file;

	Lock();
	name = GetImageFileName(pIter);
	Unlock();
	if (name) {
		file = g_build_filename(m_DataDir, name, NULL);
		g_free(name);
		return file;
	}
	return NULL;
}


// Get name of the phootmetry file
gchar *CProject::GetPhotFileName(int frame_id)
{
	gchar *file = NULL;
	GtkTreePath *path;

	Lock();
	path = GetFilePath(frame_id);
	if (path) {
		file = GetPhotFileName(path);
		gtk_tree_path_free(path);
	}
	Unlock();

	return file;
}
gchar *CProject::GetPhotFileName(GtkTreePath *pPath)
{
	gchar *file = NULL;
	GtkTreeIter iter;

	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		file = GetPhotFileName(&iter);
		Unlock();
	}
	return file;
}
gchar *CProject::GetPhotFileName(GtkTreeIter *pIter)
{
	gchar *file = NULL;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_PHOTFILE, &file, -1);
	Unlock();

	return file;
}

// Get full path to the photometry file
gchar *CProject::GetPhotFile(int frame_id)
{
	gchar *name, *file;
	
	Lock();
	name = GetPhotFileName(frame_id);
	Unlock();
	if (name) {
		file = g_build_filename(m_DataDir, name, NULL);
		g_free(name);
		return file;
	}
	return NULL;
}
gchar *CProject::GetPhotFile(GtkTreePath *pPath)
{
	gchar *name, *file;
	
	Lock();
	name = GetPhotFileName(pPath);
	Unlock();
	if (name) {
		file = g_build_filename(m_DataDir, name, NULL);
		g_free(name);
		return file;
	}
	return NULL;
}
gchar *CProject::GetPhotFile(GtkTreeIter *pIter)
{
	gchar *name, *file;
	
	Lock();
	name = GetPhotFileName(pIter);
	Unlock();
	if (name) {
		file = g_build_filename(m_DataDir, name, NULL);
		g_free(name);
		return file;
	}
	return NULL;
}

// Set error message, change the state and set the error flag
void CProject::SetError(int frame_id, const char *message, int state_mask, int state_value, bool update_icon)
{
	Lock();
	GtkTreePath *pPath = GetFilePath(frame_id);
	if (pPath) {
		SetError(pPath, message, state_mask, state_value, update_icon);
		gtk_tree_path_free(pPath);
	}
	Unlock();
}
void CProject::SetError(GtkTreePath *pPath, const char *message, int state_mask, int state_value, bool update_icon)
{
	GtkTreeIter iter;
	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		SetError(&iter, message, state_mask, state_value, update_icon);
		Unlock();
	}
}
void CProject::SetError(GtkTreeIter *pIter, const char *message, int state_mask, int state_value, bool update_icon)
{
	int state;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_STATE, &state, -1);
	state = (state & ~state_mask) | (CFILE_ERROR | state_value);
	if (update_icon)
		state &= ~CFILE_THUMBNAIL;
	gtk_list_store_set(m_Frames, pIter, 
		FRAME_STATE, state, FRAME_REPORT, message, -1);
	Unlock();
}

// Set report, change the state and clear the error flag
void CProject::SetResult(int frame_id, const char *message, int state_mask, int state_value, bool update_icon)
{
	Lock();
	GtkTreePath *pPath = GetFilePath(frame_id);
	if (pPath) {
		SetResult(pPath, message, state_mask, state_value, update_icon);
		gtk_tree_path_free(pPath);
	}
	Unlock();
}
void CProject::SetResult(GtkTreePath *pPath, const char *message, int state_mask, int state_value, bool update_icon)
{
	GtkTreeIter iter;
	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		SetResult(&iter, message, state_mask, state_value, update_icon);
		Unlock();
	}
}
void CProject::SetResult(GtkTreeIter *pIter, const char *message, int state_mask, int state_value, bool update_icon)
{
	int state;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_STATE, &state, -1);
	state = (state & ~(CFILE_ERROR | state_mask)) | (state_value & state_mask);
	if (update_icon)
		state &= ~CFILE_THUMBNAIL;
	gtk_list_store_set(m_Frames, pIter, 
		FRAME_STATE, state, FRAME_REPORT, message, -1);
	Unlock();
}

// Set report, change the state and clear the error flag
void CProject::SetThumbnail(int frame_id, GdkPixbuf *pImage)
{
	Lock();
	GtkTreePath *pPath = GetFilePath(frame_id);
	if (pPath) {
		SetThumbnail(pPath, pImage);
		gtk_tree_path_free(pPath);
	}
	Unlock();
}
void CProject::SetThumbnail(GtkTreePath *pPath, GdkPixbuf *pImage)
{
	GtkTreeIter iter;
	if (pPath) {
		Lock();
		gtk_tree_model_get_iter(GTK_TREE_MODEL(m_Frames), &iter, pPath);
		SetThumbnail(&iter, pImage);
		Unlock();
	}
}
void CProject::SetThumbnail(GtkTreeIter *pIter, GdkPixbuf *pImage)
{
	int state;

	Lock();
	gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), pIter, FRAME_STATE, &state, -1);
	state = state | CFILE_THUMBNAIL;
	gtk_list_store_set(m_Frames, pIter, FRAME_STATE, state, 
		FRAME_THUMBNAIL, pImage, -1);
	Unlock();
}

void CProject::SetOrigBiasFile(const CFileInfo &info)
{
	Lock();
	m_OrigBiasFile = info;
	Unlock();
}

void CProject::SetTempBiasFile(const CFileInfo &info)
{
	Lock();
	m_TempBiasFile = info;
	Unlock();
}

void CProject::SetOrigDarkFile(const CFileInfo &info)
{
	Lock();
	m_OrigDarkFile = info;
	Unlock();
}

void CProject::SetTempDarkFile(const CFileInfo &info)
{
	Lock();
	m_TempDarkFile = info;
	Unlock();
}

void CProject::SetOrigFlatFile(const CFileInfo &info)
{
	Lock();
	m_OrigFlatFile = info;
	Unlock();
}

void CProject::SetTempFlatFile(const CFileInfo &info)
{
	Lock();
	m_TempFlatFile = info;
	Unlock();
}

void CProject::SetTempCatFile(const CFileInfo &info)
{
	Lock();
	m_TempCatFile = info;
	Unlock();
}

void CProject::GetStatus(tStatus *status) const
{
	memset(status, 0, sizeof(tStatus));
	Lock();
	gtk_tree_model_foreach(GTK_TREE_MODEL(m_Frames),
		(GtkTreeModelForeachFunc)MakeStatus, status);
	status->open = isOpen();
	status->readonly = isReadOnly();
	Unlock();
}

tProjectType CProject::ProjectType(void) const
{
	tProjectType retval;
	Lock();
	if (isOpen())
		retval = m_Profile.ProjectType();
	else
		retval = EndOfProjectTypes;
	Unlock();
	return retval;
}

// Copy data files to given folder
bool CProject::CopyFiles(const gchar *datadir, GError **error) const
{
	bool ok = true;
	int id, state;
	gchar *ftemp, *fphot;
	GtkTreeIter iter;

	if (!force_directory(datadir)) {
		g_set_error(error, g_AppError, 0, "Failed to create the directory: %s", datadir);
		return false;
	}

	Lock();

	if (m_OrigBiasFile.Valid() && m_TempBiasFile.Valid()) {
		gchar *dstpath = g_build_filename(datadir, g_basename(m_TempBiasFile.FullPath()), NULL);
		ok = copy_file(m_TempBiasFile.FullPath(), dstpath, false, error);
		g_free(dstpath);
	}
	if (ok && m_OrigDarkFile.Valid() && m_TempDarkFile.Valid()) {
		gchar *dstpath = g_build_filename(datadir, g_basename(m_TempDarkFile.FullPath()), NULL);
		ok = copy_file(m_TempDarkFile.FullPath(), dstpath, false, error);
		g_free(dstpath);
	}
	if (ok && m_OrigFlatFile.Valid() && m_TempFlatFile.Valid()) {
		gchar *dstpath = g_build_filename(datadir, g_basename(m_TempFlatFile.FullPath()), NULL);
		ok = copy_file(m_TempFlatFile.FullPath(), dstpath, false, error);
		g_free(dstpath);
	}
	if (ok && m_RefType==REF_CATALOG_FILE && m_OrigCatFile.Valid() && m_TempCatFile.Valid()) {
		gchar *dstpath = g_build_filename(datadir, g_basename(m_TempCatFile.FullPath()), NULL);
		ok = copy_file(m_TempCatFile.FullPath(), dstpath, false, error);
		g_free(dstpath);
	}

	gboolean iterok = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(m_Frames), &iter);
	while (iterok && ok) {
		gtk_tree_model_get(GTK_TREE_MODEL(m_Frames), &iter,
			FRAME_ID, &id, FRAME_STATE, &state, FRAME_TEMPFILE, &ftemp, 
			FRAME_PHOTFILE, &fphot, -1);
		if ((state & CFILE_CONVERSION)!=0 && ftemp) {
			gchar *srcpath = g_build_filename(m_DataDir, ftemp, NULL);
			gchar *dstpath = g_build_filename(datadir, ftemp, NULL);
			ok = copy_file(srcpath, dstpath, false, error);
			g_free(srcpath);
			g_free(dstpath);
		}
		if (ok && (state & CFILE_PHOTOMETRY)!=0 && fphot) {
			gchar *srcpath = g_build_filename(m_DataDir, fphot, NULL);
			gchar *dstpath = g_build_filename(datadir, fphot, NULL);
			ok = copy_file(srcpath, dstpath, false, error);
			g_free(srcpath);
			g_free(dstpath);
		}
		iterok = gtk_tree_model_iter_next(GTK_TREE_MODEL(m_Frames), &iter);
	}

	Unlock();

	return ok;
}

//
// Set configuration parameters
//
void CProject::SetProfile(const CProfile &profile)
{
	Lock();
	m_Profile = profile;
	Unlock();
}

//
// Set dialog settings
//
void CProject::SetStr(const char *group, const char *key, const char *val)
{
	Lock();
	g_key_file_set_string(m_Params, group, key, (val!=NULL ? val : ""));
	Unlock();
}
void CProject::SetDbl(const char *group, const char *key, const double val)
{
	Lock();
	g_key_file_set_double(m_Params, group, key, val);
	Unlock();
}
void CProject::SetInt(const char *group, const char *key, const int val)
{
	Lock();
	g_key_file_set_integer(m_Params, group, key, val);
	Unlock();
}
void CProject::SetBool(const char *group, const char *key, bool val)
{
	Lock();
	g_key_file_set_boolean(m_Params, group, key, val);
	Unlock();
}

//
// Get dialog settings
//
char *CProject::GetStr(const char *group, const char *key, const char *defval) const
{
	char *res;
	GError *error = NULL;

	Lock();
	res = g_key_file_get_string(m_Params, group, key, &error);
	if (error) {
		res = (defval ? g_strdup(defval) : NULL);
		g_error_free(error);
	}
	Unlock();
	return res;
}
double CProject::GetDbl(const char *group, const char *key, double defval) const
{
	double res;
	GError *error = NULL;

	Lock();
	res = g_key_file_get_double(m_Params, group, key, &error);
	if (error) {
		res = defval;
		g_error_free(error);
	}
	Unlock();
	return res;
}
int CProject::GetInt(const char *group, const char *key, const int defval) const
{
	int res;
	GError *error = NULL;

	Lock();
	res = g_key_file_get_integer(m_Params, group, key, &error);
	if (error) {
		res = defval;
		g_error_free(error);
	}
	Unlock();
	return res;
}
bool CProject::GetBool(const char *group, const char *key, const bool defval) const
{
	bool res;
	GError *error = NULL;

	Lock();
	res = g_key_file_get_boolean(m_Params, group, key, &error)!=0;
	if (error) {
		res = defval;
		g_error_free(error);
	}
	Unlock();
	return res;
}
