/**************************************************************

preview_dlg.cpp (C-Munipack project)
The preview dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>

#include "showheader_dlg.h"
#include "utils.h"
#include "main.h"
#include "ctxhelp.h"

const static struct {
	const char *caption;		// Column name
} TreeViewColumns[] = {
	{ "Name" },
	{ "Value" },
	{ "Comment" },
	{ NULL }
};

//-------------------------   CHOOSE STARS DIALOG   --------------------------------

CShowHeaderDlg::CShowHeaderDlg(GtkWindow *pParent)
{
	int i;

	// Dialog with buttons
	m_pDlg = gtk_dialog_new_with_buttons("", pParent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR),
		GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
	gtk_dialog_widget_standard_tooltips(GTK_DIALOG(m_pDlg));

	// Make tree store
	m_Data = gtk_tree_store_new(3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

	// Make tree view
	m_View = gtk_tree_view_new();
	for (i=0; TreeViewColumns[i].caption!=NULL; i++) {
		GtkTreeViewColumn *col = gtk_tree_view_column_new();
		// Set column name and alignment
		gtk_tree_view_column_set_title(col, TreeViewColumns[i].caption);
		gtk_tree_view_append_column(GTK_TREE_VIEW(m_View), col);
		// Add text renderer
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(col, renderer, TRUE);
		gtk_tree_view_column_add_attribute(col, renderer, "text", i);
		gtk_tree_view_set_model(GTK_TREE_VIEW(m_View), GTK_TREE_MODEL(m_Data));
	}
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(m_View), TRUE);

	// Scrolled window
	m_ScrWnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(m_ScrWnd),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(m_ScrWnd), 
		GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(m_ScrWnd), m_View);
	gtk_widget_set_size_request(m_ScrWnd, 480, 320);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(m_pDlg)->vbox), m_ScrWnd);

	gtk_widget_show_all(GTK_DIALOG(m_pDlg)->vbox);
}

CShowHeaderDlg::~CShowHeaderDlg()
{
	gtk_widget_destroy(m_pDlg);
	g_object_unref(m_Data);
}

void CShowHeaderDlg::Execute(const CCCDFile *img, const gchar *name)
{
	int i;
	gchar *key, *val, *com;
	
	if (!img->Valid())
		return;

	BeginUpdate();
	SetCaption(name);
	for (i=0; img->GetParam(i, &key, &val, &com); i++) {
		Add(key, val, com);
		g_free(key);
		g_free(val);
		g_free(com);
	}
	EndUpdate();

	gtk_dialog_run(GTK_DIALOG(m_pDlg));
}

void CShowHeaderDlg::Execute(const CPhot *pht, const gchar *name)
{
	CmpackDateTime dt;
	CmpackPhtInfo info;
	const struct tm *t;
	char datestr[64], timestr[64], actdate[256];

	// Fill the table
	if (!pht->Valid())
		return;

	pht->GetParams(CMPACK_PI_FRAME_PARAMS | CMPACK_PI_ORIGIN_CRDATE | CMPACK_PI_PHOT_PARAMS, info);
	BeginUpdate();
	SetCaption(name);
	Add("width", info.width, "Image width in pixels");
	Add("height", info.height, "Image height in pixels");
	Add("jd", info.jd, 12, "Julian date of observation");
	cmpack_decodejd(info.jd, &dt);
	cmpack_datetostr(&dt.date, datestr, 64);
	Add("date", datestr, "Date of observation");
	cmpack_timetostr(&dt.time, timestr, 64);
	Add("time", timestr, "Time of observation");
	Add("filter", info.filter, "Optical filter");
	Add("exptime", info.exptime, 6, "Exposure duration in seconds");
	if (info.ccdtemp>-900 && info.ccdtemp<900)
		Add("ccdtemp", info.ccdtemp, 6, "CCD temperature in deg.C");
	Add("origin", info.origin, "Reduction software");
	t = &info.crtime;
	sprintf(actdate, "%04d-%02d-%02d %02d:%02d:%02d", t->tm_year+1900, 
		t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
	Add("crdate", actdate, "Date of creation");

	// Photometry process information 
	Add("phot_stars", pht->ObjectCount(), "Number of stars");
	Add("phot_apertures", pht->ApertureCount(), "Number of apertures");
	Add("phot_datalo", info.range[0], 6, "Lowest good pixel value");
	Add("phot_datahi", info.range[1], 6, "Highest good pixel value");
	Add("phot_gain", info.gain, 6, "Gain in electrons per ADU");
	Add("phot_rnoise", info.rnoise, 6, "Readout noise in ADU");
	Add("phot_fwhm_exp", info.fwhm_exp, 6, "Expected FWHM of objects");
	Add("phot_fwhm_mean", info.fwhm_mean, 6, "Mean FWHM of objects");
	Add("phot_fwhm_err", info.fwhm_err, 6, "Std. error of FWHM of objects");
	Add("phot_thresh", info.threshold, 6, "Detection threshold");
	Add("phot_losharp", info.sharpness[0], 6, "Low sharpness cutoff");
	Add("phot_hisharp", info.sharpness[1], 6, "High sharpness cutoff");
	Add("phot_loround", info.roundness[0], 6, "Low roundness cutoff");
	Add("phot_hiround", info.roundness[1], 6, "High roundness cutoff");
	EndUpdate();

	gtk_dialog_run(GTK_DIALOG(m_pDlg));
}

void CShowHeaderDlg::Execute(const CCatalog *cat, const gchar *name)
{
	int		i;
	const char	*key, *val, *com;

	if (!cat->Valid())
		return;

	BeginUpdate();
	SetCaption(name);
	i = 0;
	while (cat->GetParam(i++, &key, &val, &com)) 
		Add(key, val, com);
	EndUpdate();

	gtk_dialog_run(GTK_DIALOG(m_pDlg));
}

void CShowHeaderDlg::Execute(const CTable *tab, const gchar *name)
{
	int		i;
	char	*key, *val;

	if (!tab->Valid())
		return;

	BeginUpdate();
	SetCaption(name);
	for (i=0; tab->GetParam(i, &key, &val); i++) {
		Add(key, val, NULL);
		g_free(key);
		g_free(val);
	}
	EndUpdate();

	gtk_dialog_run(GTK_DIALOG(m_pDlg));
}

//
// Set window caption
//
void CShowHeaderDlg::SetCaption(const char *name)
{
	char buf[512];

	if (name) 
		sprintf(buf, "%s - header", name);
	else
		strcpy(buf, "Header");
	gtk_window_set_title(GTK_WINDOW(m_pDlg), buf);
}

//
// Start update of multiple rows
//
void CShowHeaderDlg::BeginUpdate(void)
{
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_View), NULL);
	gtk_tree_store_clear(m_Data);
}

//
// Finish update of multiple files
//
void CShowHeaderDlg::EndUpdate(void)
{
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_View), GTK_TREE_MODEL(m_Data));
	gtk_tree_view_expand_all(GTK_TREE_VIEW(m_View));
}

//
// Add field
//
void CShowHeaderDlg::Add(const gchar *keyword, const gchar *value, const gchar *comment)
{
	GtkTreeIter iter;

	gtk_tree_store_append(m_Data, &iter, NULL);
	if (keyword)
		gtk_tree_store_set(m_Data, &iter, 0, keyword, -1);
	if (value)
		gtk_tree_store_set(m_Data, &iter, 1, value, -1);
	if (comment)
		gtk_tree_store_set(m_Data, &iter, 2, comment, -1);
}

void CShowHeaderDlg::Add(const gchar *keyword, int value, const gchar *comment)
{
	char buf[64];

	sprintf(buf, "%d", value);
	Add(keyword, buf, comment);
}

void CShowHeaderDlg::Add(const gchar *keyword, double value, int prec, const gchar *comment)
{
	char buf[64];

	sprintf(buf, "%.*f", prec, value);
	Add(keyword, buf, comment);
}
