/**************************************************************

table_class.cpp (C-Munipack project)
Table class interface
Copyright (C) 2011 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>

#include "table_class.h"
#include "utils.h"
#include "main.h"

#define ALLOC_BY 64

//--------------------------   HELPER FUNCTIONS   ----------------------------------

static gchar *FromLocale(const char *str)
{
	if (str)
		return g_locale_to_utf8(str, -1, NULL, NULL, NULL);
	return NULL;
}

class CTablePrinter
{
public:
	CTablePrinter(int col):m_col(col) {}
	virtual ~CTablePrinter() {}

	static void print(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
		GtkTreeModel *tree_model, GtkTreeIter *iter, CTablePrinter *data) 
	{
		char buf[256];
		buf[0] = '\0';
		data->fprint(tree_model, iter, buf);
		g_object_set(cell, "text", buf, NULL);
	}

	static void free(CTablePrinter *data) { delete data; }

protected:
	int m_col;

	virtual void fprint(GtkTreeModel *tree_model, GtkTreeIter *iter, char *buf) = 0;
};

class CTablePrintInt:public CTablePrinter
{
public:
	CTablePrintInt(int col, int min, int max):CTablePrinter(col), m_min(min), m_max(max) {}

private:
	int m_min, m_max;

protected:
	virtual void fprint(GtkTreeModel *tree_model, GtkTreeIter *iter, char *buf)
	{
		int value;
		gtk_tree_model_get(tree_model, iter, m_col, &value, -1);
		if (value>=m_min && value<=m_max)
			sprintf(buf, "%d", value);
	}
};

class CTablePrintDbl:public CTablePrinter
{
public:
	CTablePrintDbl(int col, double min, double max, int prec):CTablePrinter(col), m_min(min), 
		m_max(max), m_prec(prec) {}

private:
	double m_min, m_max;
	int m_prec;

protected:
	virtual void fprint(GtkTreeModel *tree_model, GtkTreeIter *iter, char *buf)
	{
		double value;
		gtk_tree_model_get(tree_model, iter, m_col, &value, -1);
		if (value>=m_min && value<=m_max) 
			sprintf(buf, "%.*f", m_prec, value);
	}
};

class CTablePrintUTC:public CTablePrinter
{
public:
	CTablePrintUTC(int col):CTablePrinter(col) {}

protected:
	virtual void fprint(GtkTreeModel *tree_model, GtkTreeIter *iter, char *buf)
	{
		double jd;
		CmpackDateTime dt;
		gtk_tree_model_get(tree_model, iter, m_col, &jd, -1);
		if (jd>=JD_MIN && jd<=JD_MAX && cmpack_decodejd(jd, &dt)==0) {
			sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d", dt.date.year, dt.date.month, dt.date.day,
				dt.time.hour, dt.time.minute, dt.time.second);
		}
	}
};

class CTableWriter
{
public:
	CTableWriter(int tabcol, int outcol):m_tabcol(tabcol), m_outcol(outcol) {}
	virtual ~CTableWriter();

	virtual void write(CmpackTable *table, CCSVWriter &writer) = 0;

	static void free(CTableWriter *data) { delete data; }

protected:
	int m_tabcol, m_outcol;
};

class CTableWriteInt:public CTableWriter
{
public:
	CTableWriteInt(int tabcol, int outcol, int min, int max):CTableWriter(tabcol, outcol), 
		m_min(min), m_max(max) {}

	virtual void write(CmpackTable *table, CCSVWriter &writer)
	{
		int value;
		if (cmpack_tab_gtdi(table, m_tabcol, &value) && value>=m_min && value<=m_max)
			writer.SetInt(m_outcol, value);
	}

private:
	int m_min, m_max;
};

class CTableWriteDbl:public CTableWriter
{
public:
	CTableWriteDbl(int tabcol, int outcol, double min, double max, int prec):CTableWriter(tabcol, outcol),
		m_min(min), m_max(max), m_prec(prec) {}

	virtual void write(CmpackTable *table, CCSVWriter &writer)
	{
		double value;
		if (cmpack_tab_gtdd(table, m_tabcol, &value) && value>=m_min && value<=m_max)
			writer.SetDbl(m_outcol, value, m_prec);
	}

private:
	double m_min, m_max;
	int m_prec;
};

class CTableWriteUTC:public CTableWriter
{
public:
	CTableWriteUTC(int tabcol, int outcol):CTableWriter(tabcol, outcol) {}

	virtual void write(CmpackTable *table, CCSVWriter &writer)
	{
		double jd;
		CmpackDateTime dt;
		if (cmpack_tab_gtdd(table, m_tabcol, &jd) && jd>=JD_MIN && jd<=JD_MAX) {
			char buf[256];
			sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d", dt.date.year, dt.date.month, dt.date.day,
				dt.time.hour, dt.time.minute, dt.time.second);
			writer.SetStr(m_outcol, buf);
		}
	}
};

static void CheckLimits(double *a, double *b, double eps, double min, double max)
{
	double z, g;

	*a = LimitValue(*a, min, max);
	*b = LimitValue(*b, min, max);
	if ((*b-*a)<eps && (max-min)>=eps) {
		z = 0.5 * (*a + *b);
		g = z - (z - min)/(max - min)*eps;
		g = LimitValue(g, MAX(min, *b-eps), MIN(*a, max-eps));
		*a = g;
		*b = g + eps;
	}
}

static void AddViewCol(GtkTreeView *view, CTablePrinter *print_data, const gchar *name, double xalign)
{
	GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "xalign", xalign, NULL);
	gtk_tree_view_insert_column_with_data_func(view, -1, name, renderer, (GtkTreeCellDataFunc)CTablePrinter::print, 
		print_data, (GDestroyNotify)CTablePrinter::free);
}

static void AddViewCol_Int(GtkTreeView *view, int column, int min, int max, const gchar *name)
{
	AddViewCol(view, new CTablePrintInt(column, min, max), name, 1.0);
}

static void AddViewCol_Dbl(GtkTreeView *view, int column, double min, double max, int prec, const gchar *name)
{
	AddViewCol(view, new CTablePrintDbl(column, min, max, prec), name, 1.0);
}

static void AddViewCol_UTC(GtkTreeView *view, int column, const gchar *name)
{
	AddViewCol(view, new CTablePrintUTC(column), name, 1.0);
}

//-----------------------   LIST OF CHANNELS   ---------------------

// Constructor with initialization
CChannel::CChannel(CTable *tab, int ycol, int ucol, tChannelInfo info, tChannelFlags flags):m_Table(tab), 
	m_Column(ycol), m_ColumnU(ucol), m_Name(NULL), m_Exported(true), m_Info(info),
	m_Flags(flags), m_DataType(CMPACK_TYPE_VOID), m_Precision(0)
{
	CmpackTabColumn iinfo;

	if (m_Table && m_Column>=0) {
		cmpack_tab_get_column(m_Table->Handle(), m_Column, CMPACK_TM_NAME | CMPACK_TM_TYPE_PREC, &iinfo);
		if (iinfo.name)
			m_Name = FromLocale(iinfo.name);
		else
			m_Name = g_strdup("");
		m_DataType = iinfo.dtype;
		m_Precision = iinfo.prec;
	}
}

// Destructor
CChannel::~CChannel(void)
{
	g_free(m_Name);
}

void CChannel::SetExported(bool enabled)
{
	m_Exported = enabled;
}

const gchar *CChannel::Unit(void) const
{
	switch (m_Info)
	{
	case DATA_MAGNITUDE:
	case DATA_DEVIATION:
		return "mag";
	case DATA_HELCOR:
		return "days";
	case DATA_OFFSET:
		return "px";
	case DATA_ALTITUDE:
		return "deg";
	default:
		return NULL;
	}
}

// Get minimum value
double CChannel::Min(void) const
{
	CmpackTabColumn info;

	if (m_Table && m_Column>=0) {
		cmpack_tab_get_column(m_Table->Handle(), m_Column, CMPACK_TM_RANGE, &info);
		if (info.range_valid)
			return info.range_min;
	}
	return 0.0;
}

// Get minimum value
double CChannel::Max(void) const
{
	CmpackTabColumn info;

	if (m_Table && m_Column>=0) {
		cmpack_tab_get_column(m_Table->Handle(), m_Column, CMPACK_TM_RANGE, &info);
		if (info.range_valid)
			return info.range_max;
	}
	return 0.0;
}

// Get minimum value
double CChannel::MaxU(void) const
{
	CmpackTabColumn info;

	if (m_Table && m_ColumnU>=0) {
		cmpack_tab_get_column(m_Table->Handle(), m_ColumnU, CMPACK_TM_RANGE, &info);
		if (info.range_valid)
			return info.range_max;
	}
	return 0.0;
}

// Default Constructor
CChannels::CChannels(void):m_Count(0), m_Capacity(0), m_List(NULL)
{
}

// Destructor
CChannels::~CChannels()
{
	int i;

	for (i=0; i<m_Count; i++)
		delete m_List[i];
	g_free(m_List);
}

// Clear the table
void CChannels::Clear(void)
{
	int i;

	for (i=0; i<m_Count; i++)
		delete m_List[i];
	m_Count = 0;
}

// Add an aperture, if the aperture with the same id is 
// in the table, it changes its parameters
void CChannels::Add(CChannel *item)
{
	if (m_Count>=m_Capacity) {
		m_Capacity += ALLOC_BY;
		m_List = (CChannel**)g_realloc(m_List, m_Capacity*sizeof(CChannel));
	}
	m_List[m_Count++] = item;
}

// Find aperture by id 
int CChannels::FindColumn(int col) const
{
	int i;

	for (i=0; i<m_Count; i++) {
		if (m_List[i]->Column() == col || m_List[i]->ColumnU() == col)
			return i;
	}
	return -1;
}

// Find aperture by id 
int CChannels::FindFirst(const gchar *name) const
{
	int i;

	for (i=0; i<m_Count; i++) {
		if (strcmp(m_List[i]->Name(), name)==0)
			return i;
	}
	return -1;
}

// Find aperture by id 
int CChannels::FindFirst(tChannelInfo info) const
{
	int i;

	for (i=0; i<m_Count; i++) {
		if (m_List[i]->Info() == info)
			return i;
	}
	return -1;
}

// Get aperture by index
const CChannel *CChannels::Get(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index];
	return NULL;
}
CChannel *CChannels::Get(int index) 
{
	if (index>=0 && index<m_Count)
		return m_List[index];
	return NULL;
}

// Get column by index
int CChannels::GetColumn(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Column();
	return -1;
}

// Get column by index
int CChannels::GetColumnU(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->ColumnU();
	return -1;
}

// Get name by index
const gchar *CChannels::GetName(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Name();
	return NULL;
}

// Get name by index
int CChannels::GetPrecision(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Precision();
	return 0;
}

// Get channel info
tChannelInfo CChannels::GetInfo(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Info();
	return DATA_UNDEFINED;
}

// Get channel info
const gchar *CChannels::GetUnit(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Unit();
	return NULL;
}

// Get channel info
double CChannels::GetMin(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Min();
	return DATA_UNDEFINED;
}

// Get channel info
double CChannels::GetMax(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Max();
	return DATA_UNDEFINED;
}

// Get channel info
double CChannels::GetMaxU(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->MaxU();
	return DATA_UNDEFINED;
}

// Get channel info
bool CChannels::GetExported(int index) const
{
	if (index>=0 && index<m_Count)
		return m_List[index]->Exported();
	return false;
}

//-----------------------   TABLE   --------------------------------

//
// Default constructor
//
CTable::CTable(void):m_Handle(NULL), m_CacheFlags(0), m_PKColumn(-1)
{
}

//
// Constructor with initialization
//
CTable::CTable(CmpackTable *handle):m_Handle(handle), m_CacheFlags(0), 
	m_PKColumn(-1)
{
}

//
// Constructor with initialization
//
CTable::CTable(CmpackTableType type):m_CacheFlags(0), m_PKColumn(-1)
{
	m_Handle = cmpack_tab_init(type);
}

//
// Destructor
//
CTable::~CTable(void)
{
	if (m_Handle)
		cmpack_tab_destroy(m_Handle);
}

//
// Create a new table
//
void CTable::Init(CmpackTableType type)
{
	Clear();
	m_Handle = cmpack_tab_init(type);
	InvalidateCache();
}

//
// Assign a table
//
void CTable::Assign(CmpackTable *table)
{
	Clear();
	m_Handle = table;
	InvalidateCache();
}

//
// Deep copy
//
void CTable::MakeCopy(const CTable &orig)
{
	Clear();
	if (orig.Handle()) {
		m_Handle = cmpack_tab_init(CMPACK_TABLE_UNSPECIFIED);
		cmpack_tab_copy(m_Handle, orig.Handle());
	}
}

//
// Load table
//
bool CTable::Load(const gchar *fpath, int *result)
{
	int res;

	Clear();

	char *f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	res = cmpack_tab_load(&m_Handle, f, 0);
	g_free(f);
	if (res!=0) {
		if (result)
			*result = res;
		return false;
	}
	InvalidateCache();
	if (result)
		*result = 0;
	return true;
}

//
// Save table
//
bool CTable::Save(const gchar *fpath, int *result)
{
	int res;

	if (!m_Handle) {
		if (result)
			*result = CMPACK_ERR_CLOSED_FILE;
		return false;
	}
	char *f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	res = cmpack_tab_save(m_Handle, f, 0, 0, 0);
	g_free(f);
	if (res!=0) {
		if (result)
			*result = res;
		return false;
	}

	InvalidateCache();
	if (result)
		*result = 0;
	return true;
}


//
// Close handle
//
void CTable::Clear(void)
{
	if (m_Handle) {
		cmpack_tab_destroy(m_Handle);
		m_Handle = NULL;
	}
	m_ChannelsX.Clear();
	m_ChannelsY.Clear();
	m_PKColumn = -1;
	InvalidateCache();
}


//
// Get type identifier
//
CmpackTableType CTable::Type(void) const
{
	if (m_Handle)
		return cmpack_tab_get_type(m_Handle);
	return CMPACK_TABLE_UNSPECIFIED;
}

//
// Get header field by index
//
bool CTable::GetParam(int index, char **key, char **val) const
{
	const char *keyword, *value;

	if (m_Handle && cmpack_tab_gkyn(m_Handle, index, &keyword, &value)) {
		if (key)
			*key = (keyword ? g_strdup(keyword) : NULL);
		if (val)
			*val = (value ? g_strdup(value) : NULL);
		return true;
	}
	return false;
}


// Set header field
void CTable::SetParam(const char *key, const char *value)
{
	if (m_Handle)
		cmpack_tab_pkys(m_Handle, key, value);
}

//
// Get number of rows
//
int CTable::Rows(void) const
{
	if (m_Handle)
		return cmpack_tab_nrows(m_Handle);
	return 0;
}

//
// Browsing
//
bool CTable::Rewind(void)
{
	return (m_Handle && cmpack_tab_rewind(m_Handle)==0);
}
bool CTable::Next(void)
{
	return (m_Handle && cmpack_tab_next(m_Handle)==0);
}
bool CTable::EndOfTable(void)
{
	return (m_Handle && cmpack_tab_eof(m_Handle)!=0);
}

//
// Get color filter name
//
const char *CTable::Filter(void) const
{
	if (m_Handle)
		return cmpack_tab_gkys(m_Handle, "Filter");
	return NULL;
}

//
// Get aperture index
//
int CTable::Aperture(void) const
{
	int value;
	if (m_Handle) {
		if (cmpack_tab_gkyi(m_Handle, "Aperture", &value)==0)
			return value;
	}
	return 0;
}

//
// Make graph data
//
CmpackGraphData *CTable::ToGraphData(int xch, int ych, bool no_errors)
{
	int		row_count, param, xcol, ycol, ucol;
	double	x, y, error;
	CmpackGraphItem item;

	if (!m_Handle)
		return NULL;

	memset(&item, 0, sizeof(CmpackGraphItem));

	CmpackGraphData *data = cmpack_graph_data_new();
	if (data) {
		const CChannel *ch_x = ChannelsX()->Get(xch), *ch_y = ChannelsY()->Get(ych);
		row_count = cmpack_tab_nrows(m_Handle);
		if (ch_x && ch_y && row_count>0) {
			cmpack_graph_data_alloc(data, row_count);
			xcol = ch_x->Column();
			ycol = ch_y->Column();
			ucol = ch_y->ColumnU();
			int index = 0;
			cmpack_tab_rewind(m_Handle);
			while (!cmpack_tab_eof(m_Handle)) {
				if (cmpack_tab_gtdd(m_Handle, xcol, &x)==0 && 
					cmpack_tab_gtdd(m_Handle, ycol, &y)==0) {
						if (ucol<0 || no_errors || cmpack_tab_gtdd(m_Handle, ucol, &error)!=0)
							error = 0.0;
						item.x = x;
						item.y = y;
						item.error = error;
						if (m_PKColumn>=0)
							cmpack_tab_gtdi(m_Handle, m_PKColumn, &param);
						else
							param = index;
						item.param = param;
						cmpack_graph_data_add(data, &item, sizeof(item));
				}
				cmpack_tab_next(m_Handle);
				index++;
			}
		}
	}
	return data;
}

//
// Make tree model
//
GtkTreeModel *CTable::ToTreeModel(void)
{
	const static int mask = CMPACK_TM_NAME | CMPACK_TM_TYPE_PREC;

	if (m_Handle) {
		int ncol = cmpack_tab_ncolumns(m_Handle);
		GType *cols = (GType*)g_malloc(ncol*sizeof(GType));
		for (int i=0; i<ncol; i++) {
			CmpackTabColumn col;
			cmpack_tab_get_column(m_Handle, i, mask, &col);
			if (col.dtype==CMPACK_TYPE_DBL) 
				cols[i] = G_TYPE_DOUBLE;
			else if (col.dtype==CMPACK_TYPE_INT)
				cols[i] = G_TYPE_INT;
			else
				cols[i] = G_TYPE_STRING;
		}
		GtkListStore *data = gtk_list_store_newv(ncol, cols);
		cmpack_tab_rewind(m_Handle);
		while (!cmpack_tab_eof(m_Handle)) {
			GtkTreeIter iter;
			gtk_list_store_append(data, &iter);
			for (int i=0; i<ncol; i++) {
				if (cols[i] == G_TYPE_DOUBLE) {
					double value;
					if (cmpack_tab_gtdd(m_Handle, i, &value)==0)
						gtk_list_store_set(data, &iter, i, value, -1);
					else
						gtk_list_store_set(data, &iter, i, DBL_MAX, -1);
				} else
				if (cols[i] == G_TYPE_INT) {
					int param;
					if (cmpack_tab_gtdi(m_Handle, i, &param)==0)
						gtk_list_store_set(data, &iter, i, param, -1);
				} else
				if (cols[i] == G_TYPE_STRING) {
					char *string;
					if (cmpack_tab_gtds(m_Handle, i, &string)==0) {
						gtk_list_store_set(data, &iter, i, string, -1);
						cmpack_free(string);
					}
				}
			}
			cmpack_tab_next(m_Handle);
		}
		g_free(cols);
		return GTK_TREE_MODEL(data);
	}
	return NULL;
}

//
// Invalidate cache
//
void CTable::InvalidateCache(void)
{
	m_CacheFlags = 0;
}

//
// Get double value from actual row
//
bool CTable::GetDbl(int column, double *val) const
{
	if (m_Handle)
		return cmpack_tab_gtdd(m_Handle, column, val)==0;
	return false;
}

// Set nul value to given column
void CTable::SetNulValue(int column, double nulval)
{
	CmpackTabColumn info;
	if (m_Handle) {
		info.nul_value = nulval;
		cmpack_tab_set_column(m_Handle, column, CMPACK_TM_NULVAL, &info);
	}
}

//
// Get column with primary keys
//
int CTable::PKColumn(void)
{
	if ((m_CacheFlags & CF_CHANNELS)==0)
		UpdateChannels();
	return m_PKColumn;
}

//
// Update list of idependent channels
//
CChannels *CTable::ChannelsX(void)
{
	if ((m_CacheFlags & CF_CHANNELS)==0)
		UpdateChannels();
	return &m_ChannelsX;
}

//
// Update list of idependent channels
//
CChannels *CTable::ChannelsY(void)
{
	if ((m_CacheFlags & CF_CHANNELS)==0)
		UpdateChannels();
	return &m_ChannelsY;
}

//
// Update channels
//
void CTable::UpdateChannels(void)
{
	const static int mask = CMPACK_TM_NAME | CMPACK_TM_TYPE_PREC;
	CmpackTabColumn col, col2;

	m_ChannelsX.Clear();
	m_ChannelsY.Clear();
	m_PKColumn = -1;
	if (!m_Handle)
		return;

	int ncol = cmpack_tab_ncolumns(m_Handle);
	CmpackTableType ttype = cmpack_tab_get_type(m_Handle);
	for (int i=0; i<ncol; i++) {
		cmpack_tab_get_column(m_Handle, i, mask, &col);
		switch (ttype)
		{
		case CMPACK_TABLE_LIGHTCURVE:
		case CMPACK_TABLE_TRACKLIST:
		case CMPACK_TABLE_AIRMASS:
			if (col.dtype == CMPACK_TYPE_INT && strcmp(col.name, "FRAME")==0) {
				m_PKColumn = i;
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "JD")==0 || strcmp(col.name, "GJD")==0 || strcmp(col.name, "JDGEO")==0)) {
				m_ChannelsX.Add(new CChannel(this, i, -1, DATA_JD));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "HJD")==0 || strcmp(col.name, "JDHEL")==0)) {
				m_ChannelsX.Add(new CChannel(this, i, -1, DATA_JD, DATA_HELIOCENTRIC));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "AIRMASS")==0) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_AIRMASS));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "ALTITUDE")==0) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_ALTITUDE));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "OFFSETX")==0 || strcmp(col.name, "OFFSETY")==0)) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_OFFSET));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "HELCOR")==0) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_HELCOR));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (col.name[0]=='V' || col.name[0]=='C' || col.name[0]=='K')) {
				int err = -1;
				if (cmpack_tab_get_column(m_Handle, i+1, mask, &col2)==0 && col2.name[0]=='s') 
					err = i+1;
				m_ChannelsY.Add(new CChannel(this, i, err, DATA_MAGNITUDE));
				if (err>=0)
					i++;
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strstr(col.name,"MAG")==col.name) {
				int err = -1;
				if (cmpack_tab_get_column(m_Handle, i+1, mask, &col2)==0 && strstr(col2.name, "ERR")==col2.name)
					err = i+1;
				m_ChannelsY.Add(new CChannel(this, i, err, DATA_MAGNITUDE));
				if (err>=0)
					i++;
			} else
			if (col.dtype==CMPACK_TYPE_DBL || col.dtype==CMPACK_TYPE_INT) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_ANY));
			}
			break;

		case CMPACK_TABLE_MUNIFIND:
			if (col.dtype==CMPACK_TYPE_INT && strcmp(col.name, "INDEX")==0) {
				m_PKColumn = i;
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "MEAN_MAG")==0) {
				m_ChannelsX.Add(new CChannel(this, i, -1, DATA_MAGNITUDE));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "STDEV")==0) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_DEVIATION));
			}
			break;

		case CMPACK_TABLE_APERTURES:
			if (strcmp(col.name, "APERTURE")==0) {
				m_PKColumn = i;
				m_ChannelsX.Add(new CChannel(this, i, -1, DATA_ANY));
			} else
			if (col.name[0]=='C' || col.name[0]=='K') {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_DEVIATION));
			} else
			if (col.dtype==CMPACK_TYPE_DBL || col.dtype==CMPACK_TYPE_INT) {
				m_ChannelsY.Add(new CChannel(this, i, -1, DATA_ANY));
			}
			break;

		default:
			if (col.dtype==CMPACK_TYPE_DBL || col.dtype==CMPACK_TYPE_INT) {
				if (m_ChannelsX.Count()==0) 
					m_ChannelsX.Add(new CChannel(this, i, -1, DATA_ANY));
				else
					m_ChannelsY.Add(new CChannel(this, i, -1, DATA_ANY));
			}
			break;
		}
	}

	m_CacheFlags |= CF_CHANNELS;
}

//
// Set graph view parameters
//
void CTable::SetView(CmpackGraphView *view, int xch, int ych, bool no_errors,
	const gchar *user_name_x, const gchar *user_name_y, tDateFormat datef)
{
	gdouble min, max;
	const gchar *name;

	CChannel *ch_x = ChannelsX()->Get(xch);
	if (ch_x) {
		min = ch_x->Min();
		max = ch_x->Max();
		name = (user_name_x ? user_name_x : ch_x->Name());
		switch (ch_x->Info())
		{
		case DATA_MAGNITUDE:
			CheckLimits(&min, &max, 0.05, -99.0, +99.0);
			cmpack_graph_view_set_x_axis(view, FALSE, FALSE, min, max, 0.05,
				GRAPH_FIXED, 1, 3, name);
			break;
		case DATA_JD:
			CheckLimits(&min, &max, 0.05, JD_MIN, JD_MAX);
			if (datef == JULIAN_DATE) {
				cmpack_graph_view_set_x_axis(view, FALSE, FALSE, min, max, 0.005,
					GRAPH_FIXED, 0, 3, ((ch_x->Flags() & DATA_HELIOCENTRIC)==0 ? "JD" : "JD (hel.)"));
			} else {
				cmpack_graph_view_set_x_axis(view, FALSE, FALSE, min, max, 0.005,
					GRAPH_TIME, 0, 3, ((ch_x->Flags() & DATA_HELIOCENTRIC)==0 ? "UTC" : "UTC (hel.)"));
			}
			break;
		default:
			CheckLimits(&min, &max, 0.005, -1e99, 1e99);
			cmpack_graph_view_set_x_axis(view, FALSE, FALSE, min, max, 0.005,
				GRAPH_FIXED, 0, 3, name);
			break;
		}
	}

	CChannel *ch_y = ChannelsY()->Get(ych);
	if (ch_y) {
		min = ch_y->Min();
		max = ch_y->Max();
		if (!no_errors && ch_y->ColumnU()>0) {
			min -= ch_y->MaxU();
			max += ch_y->MaxU();
		}
		name = (user_name_y ? user_name_y : ch_y->Name());
		switch (ch_y->Info())
		{
		case DATA_MAGNITUDE:
			CheckLimits(&min, &max, 0.05, -99.0, +99.0);
			cmpack_graph_view_set_y_axis(view, FALSE, TRUE, min, max, 0.05,
				GRAPH_FIXED, 1, 3, name);
			break;
		case DATA_DEVIATION:
			CheckLimits(&min, &max, 0.05, 1e-6, 99.0);
			cmpack_graph_view_set_y_axis(view, FALSE, FALSE, min, max, 0.05,
				GRAPH_FIXED, 1, 3, name);
			break;
		case DATA_HELCOR:
			CheckLimits(&min, &max, 0.0001, -1.0, 1.0);
			cmpack_graph_view_set_y_axis(view, FALSE, FALSE, min, max, 0.0001,
				GRAPH_FIXED, 1, 4, name);
			break;
		case DATA_AIRMASS:
			CheckLimits(&min, &max, 0.01, 1.0, 99.0);
			cmpack_graph_view_set_y_axis(view, FALSE, FALSE, min, max, 0.01,
				GRAPH_FIXED, 1, 2, name);
			break;
		case DATA_OFFSET:
			CheckLimits(&min, &max, 0.1, -1e6, 1e6);
			cmpack_graph_view_set_y_axis(view, FALSE, FALSE, min, max, 0.1,
				GRAPH_FIXED, 0, 1, name);
			break;
		case DATA_ALTITUDE:
			CheckLimits(&min, &max, 1.0, -90.0, +90.0);
			cmpack_graph_view_set_y_axis(view, FALSE, FALSE, min, max, 1.0,
				GRAPH_FIXED, 0, 0, name);
			break;
		default:
			CheckLimits(&min, &max, 0.005, -1e99, 1e99);
			cmpack_graph_view_set_y_axis(view, FALSE, FALSE, min, max, 0.005,
				GRAPH_FIXED, 0, 3, name);
			break;
		}
	}
}

//
// Set graph view parameters
//
void CTable::SetView(GtkTreeView *view)
{
	const static int mask = CMPACK_TM_NAME | CMPACK_TM_TYPE_PREC;
	CmpackTabColumn col, col2;
	char buf[256];

	// Delete old columns
	GList *list = gtk_tree_view_get_columns(view);
	for (GList *ptr=list; ptr!=NULL; ptr=ptr->next) 
		gtk_tree_view_remove_column(view, (GtkTreeViewColumn*)(ptr->data));
	g_list_free(list);

	// Create new columns
	if (m_Handle) {
		int ncol = cmpack_tab_ncolumns(m_Handle);
		CmpackTableType ttype = cmpack_tab_get_type(m_Handle);
		for (int i=0; i<ncol; i++) {
			cmpack_tab_get_column(m_Handle, i, mask, &col);
			switch (ttype)
			{
			case CMPACK_TABLE_LIGHTCURVE:
			case CMPACK_TABLE_TRACKLIST:
			case CMPACK_TABLE_AIRMASS:
				if (col.dtype == CMPACK_TYPE_INT && strcmp(col.name, "FRAME")==0) {
					AddViewCol_Int(view, i, 1, INT_MAX, "Frame #");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "JD")==0 || strcmp(col.name, "GJD")==0 || strcmp(col.name, "JDGEO")==0)) {
					AddViewCol_Dbl(view, i, JD_MIN, JD_MAX, col.prec, "JD");
					AddViewCol_UTC(view, i, "UTC");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "HJD")==0 || strcmp(col.name, "JDHEL")==0)) {
					AddViewCol_Dbl(view, i, JD_MIN, JD_MAX, col.prec, "JD (hel.)");
					AddViewCol_UTC(view, i, "UTC (hel.)");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "AIRMASS")==0) {
					AddViewCol_Dbl(view, i, 1.0, 99.0, col.prec, "Airmass");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "ALTITUDE")==0) {
					AddViewCol_Dbl(view, i, -90.0, 90.0, col.prec, "Altitude");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "OFFSETX")==0 || strcmp(col.name, "OFFSETY")==0)) {
					sprintf(buf, "Offset %s", col.name+6);
					AddViewCol_Dbl(view, i, -1e6, +1e6, col.prec, buf);
				} else
				if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "HELCOR")==0) {
					AddViewCol_Dbl(view, i, -1.0, 1.0, col.prec, "Hel. corr.");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && (col.name[0]=='V' || col.name[0]=='C' || col.name[0]=='K')) {
					AddViewCol_Dbl(view, i, -99.0, 99.0, col.prec, col.name);
					if (cmpack_tab_get_column(m_Handle, i+1, mask, &col2)==0 && col2.name[0]=='s') {
						sprintf(buf, "\xCF\x83%s", col2.name+1);
						AddViewCol_Dbl(view, i+1, 1e-6, 99.0, col2.prec, buf);
						i++;
					}
				} else
				if (col.dtype==CMPACK_TYPE_DBL && strstr(col.name,"MAG")==col.name) {
					AddViewCol_Dbl(view, i, -99.0, 99.0, col.prec, col.name);
					if (cmpack_tab_get_column(m_Handle, i+1, mask, &col2)==0 && strstr(col2.name, "ERR")==col2.name) {
						AddViewCol_Dbl(view, i+1, 1e-6, 99.0, col2.prec, col2.name);
						i++;
					}
				} else
				if (col.dtype==CMPACK_TYPE_DBL) {
					AddViewCol_Dbl(view, i, DBL_MIN, DBL_MAX, col.prec, col.name);
				} else
				if (col.dtype==CMPACK_TYPE_INT) {
					AddViewCol_Int(view, i, INT_MIN, INT_MAX, col.name);
				}
				break;

			case CMPACK_TABLE_MUNIFIND:
				if (col.dtype==CMPACK_TYPE_INT && strcmp(col.name, "INDEX")==0) {
					AddViewCol_Int(view, i, 1, INT_MAX, "Object #");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "MEAN_MAG")==0) {
					AddViewCol_Dbl(view, i, -99.0, 99.0, col.prec, "mag");
				} else
				if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "STDEV")==0) {
					AddViewCol_Dbl(view, i, 1e-6, 99.0, col.prec, "dev");
				}
				break;

			case CMPACK_TABLE_APERTURES:
				if (strcmp(col.name, "APERTURE")==0) {
					AddViewCol_Int(view, i, 1, INT_MAX, "Aperture #");
				} else
				if (col.name[0]=='C' || col.name[0]=='K') {
					AddViewCol_Dbl(view, i, -99.0, 99.0, col.prec, col.name);
				} else
				if (col.dtype==CMPACK_TYPE_DBL) {
					AddViewCol_Dbl(view, i, DBL_MIN, DBL_MAX, col.prec, col.name);
				} else
				if (col.dtype==CMPACK_TYPE_INT) {
					AddViewCol_Int(view, i, INT_MIN, INT_MAX, col.name);
				}
				break;

			default:
				if (col.dtype==CMPACK_TYPE_DBL) {
					AddViewCol_Dbl(view, i, DBL_MIN, DBL_MAX, col.prec, col.name);
				} else
				if (col.dtype==CMPACK_TYPE_INT) {
					AddViewCol_Int(view, i, INT_MIN, INT_MAX, col.name);
				}
				break;
			}
		}
	}

	// Insert last empty column that will stretch to the rest of the table
	gtk_tree_view_insert_column_with_data_func(view, -1, NULL, gtk_cell_renderer_text_new(), NULL, NULL, NULL);
}

bool CTable::Export(const gchar *filepath, const gchar *format, int sort_column_id,
		GtkSortType sort_order, GError **error)
{
	g_assert(format != NULL);
	g_assert(filepath != NULL);

	if ((m_CacheFlags & CF_CHANNELS)==0)
		UpdateChannels();

	if (strcmp(format, "csv")) {
		g_set_error(error, g_AppError, 0, "Unsupported file format (%s)", format);
		return false;
	}

	FILE *f = open_file(filepath, "w");
	if (!f) {
		g_set_error(error, g_AppError, 0, "Failed to create the file: %s", filepath);
		return false;
	}

	CCSVWriter writer(f);
	ExportTable(writer, sort_column_id, sort_order);
	fclose(f);
	return true;
}

// Export table to a file
bool CTable::SaveAs(const gchar *filepath, bool ignore_ids, bool no_header, bool no_errors, int *result)
{
	return false;
}


// Export table to file
void CTable::ExportTable(CCSVWriter &writer, int sort_column_id, GtkSortType sort_order)
{
	/*const static int mask = CMPACK_TM_NAME | CMPACK_TM_TYPE_PREC;
	char buf[256];

	GSList *outcols = NULL;

	int tabcols = cmpack_tab_ncolumns(m_Handle);	// Table columns
	CmpackTableType ttype = cmpack_tab_get_type(m_Handle);
	tSortType sort_type = SORT_TYPE_INT;
	int sort_col = -1, out_col = 0;
	for (int i=0; i<tabcols; i++) {
		CmpackTabColumn col;
		cmpack_tab_get_column(m_Handle, i, mask, &col);
		switch (ttype)
		{
		case CMPACK_TABLE_LIGHTCURVE:
		case CMPACK_TABLE_TRACKLIST:
		case CMPACK_TABLE_AIRMASS:
			if (col.dtype == CMPACK_TYPE_INT && strcmp(col.name, "FRAME")==0) {
				writer.AddColumn("Frame #");
				outcols = g_slist_prepend(outcols, new CTableWriteInt(i, out_col++, 1, INT_MAX));
				if (sort_column_id == SORT_FRAME_ID && sort_col<0) {
					sort_type = SORT_TYPE_INT;
					sort_col = i;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "JD")==0 || strcmp(col.name, "GJD")==0 || strcmp(col.name, "JDGEO")==0)) {
				writer.AddColumn("JD");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, JD_MIN, JD_MAX, col.prec));
				writer.AddColumn("UTC");
				outcols = g_slist_prepend(outcols, new CTableWriteUTC(i, out_col++));
				if (sort_column_id == SORT_DATE && sort_col<0) {
					sort_type = SORT_TYPE_DOUBLE;
					sort_col = i;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "HJD")==0 || strcmp(col.name, "JDHEL")==0)) {
				writer.AddColumn("JD (hel.)");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, JD_MIN, JD_MAX, col.prec));
				writer.AddColumn("UTC (hel.)");
				outcols = g_slist_prepend(outcols, new CTableWriteUTC(i, out_col++));
				if (sort_column_id == SORT_DATE && sort_col<0) {
					sort_type = SORT_TYPE_DOUBLE;
					sort_col = i;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "AIRMASS")==0) {
				writer.AddColumn("Airmass");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, 1.0, 99.0, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "ALTITUDE")==0) {
				writer.AddColumn("Altitude");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -90.0, 90.0, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (strcmp(col.name, "OFFSETX")==0 || strcmp(col.name, "OFFSETY")==0)) {
				sprintf(buf, "Offset %s", col.name+6);
				writer.AddColumn(buf);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -1e6, +1e6, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "HELCOR")==0) {
				writer.AddColumn("Hel. corr.");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -1.0, 1.0, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_DBL && (col.name[0]=='V' || col.name[0]=='C' || col.name[0]=='K')) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -99.0, 99.0, col.prec));
				CmpackTabColumn col2;
				if (cmpack_tab_get_column(m_Handle, i+1, mask, &col2)==0 && col2.name[0]=='s') {
					sprintf(buf, "\xCF\x83%s", col2.name+1);
					writer.AddColumn(buf);
					outcols = g_list_prepend(outcols, new CTableWriteDbl(i+1, out_col++, 1e-6, 99.0, col2.prec));
					i++;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strstr(col.name,"MAG")==col.name) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -99.0, 99.0, col.prec));
				CmpackTabColumn col2;
				if (cmpack_tab_get_column(m_Handle, i+1, mask, &col2)==0 && strstr(col2.name, "ERR")==col2.name) {
					writer.AddColumn(col2.name);
					outcols = g_slist_prepend(outcols, new CTableWriteDbl(i+1, out_col++, 1e-6, 99.0, col2.prec));
					i++;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, DBL_MIN, DBL_MAX, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_INT) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteInt(i, out_col++, INT_MIN, INT_MAX));
			}
			break;

		case CMPACK_TABLE_MUNIFIND:
			if (col.dtype==CMPACK_TYPE_INT && strcmp(col.name, "INDEX")==0) {
				writer.AddColumn("Object #");
				outcols = g_slist_prepend(outcols, new CTableWriteInt(i, out_col++, 1, INT_MAX));
				if (sort_column_id == SORT_OBJECT_ID && sort_col<0) {
					sort_type = SORT_TYPE_INT;
					sort_col = i;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "MEAN_MAG")==0) {
				writer.AddColumn("mag");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -99.0, 99.0, col.prec));
				if (sort_column_id == SORT_MEAN_MAG && sort_col<0) {
					sort_type = SORT_TYPE_INT;
					sort_col = i;
				}
			} else
			if (col.dtype==CMPACK_TYPE_DBL && strcmp(col.name, "STDEV")==0) {
				writer.AddColumn("dev");
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, 1e-6, 99.0, col.prec));
			}
			break;

		case CMPACK_TABLE_APERTURES:
			if (strcmp(col.name, "APERTURE")==0) {
				writer.AddColumn("Aperture #");
				outcols = g_slist_prepend(outcols, new CTableWriteInt(i, out_col++, 1, INT_MAX));
				if (sort_column_id == SORT_APERTURE_ID && sort_col<0) {
					sort_type = SORT_TYPE_INT;
					sort_col = i;
				}
			} else
			if (col.name[0]=='C' || col.name[0]=='K') {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, -99.0, 99.0, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_DBL) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, DBL_MIN, DBL_MAX, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_INT) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteInt(i, out_col++, INT_MIN, INT_MAX));
			}
			break;

		default:
			if (col.dtype==CMPACK_TYPE_DBL) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteDbl(i, out_col++, DBL_MIN, DBL_MAX, col.prec));
			} else
			if (col.dtype==CMPACK_TYPE_INT) {
				writer.AddColumn(col.name);
				outcols = g_slist_prepend(outcols, new CTableWriteInt(i, out_col++, INT_MIN, INT_MAX));
			}
			break;
		}
	}
	
	int count = cmpack_tab_nrows(m_Handle);
	if (count>1 && sort_col>=0) {
		tSortItem *items = (tSortItem*)g_malloc(count*sizeof(tSortItem));
		if (items) {
			switch (sort_type) 
			{
			case SORT_TYPE_INT:
				for (int i=0; i<count; i++) {
					items[i].row = i;
					cmpack_tab_setpos(m_Handle, i);
					if (!cmpack_tab_gtdi(m_Handle, sort_col, &items[i].value.i)) 
						items[i].value.i = -1;
				}
				break;
			case SORT_TYPE_DOUBLE:
				for (int i=0; i<count; i++) {
					items[i].row = i;
					cmpack_tab_setpos(m_Handle, i);
					if (!cmpack_tab_gtdd(m_Handle, sort_col, &items[i].value.d)) 
						items[i].value.d = DBL_MAX;
				}
				break;
			}
			SortItems(items, count, sort_type, sort_order);
			for (int j=0; j<count; j++) {
				cmpack_tab_setpos(m_Handle, items[j].row);
				int i = 0;
				for (GSList *ptr=outcols; ptr!=NULL; ptr=ptr->next) {
					CTableWriter *w = (CTableWriter*)ptr->data;


					if (cols[i] == G_TYPE_DOUBLE) {
						double value;
						if (cmpack_tab_gtdd(m_Handle, i, &value)==0)
							gtk_list_store_set(data, &iter, i, value, -1);
						else
							gtk_list_store_set(data, &iter, i, DBL_MAX, -1);
					} else
					if (cols[i] == G_TYPE_INT) {
						int param;
						if (cmpack_tab_gtdi(m_Handle, i, &param)==0)
							gtk_list_store_set(data, &iter, i, param, -1);
					} else
					if (cols[i] == G_TYPE_STRING) {
						char *string;
						if (cmpack_tab_gtds(m_Handle, i, &string)==0) {
							gtk_list_store_set(data, &iter, i, string, -1);
							cmpack_free(string);
						}
					}
				}



	cmpack_tab_rewind(m_Handle);
	while (!cmpack_tab_eof(m_Handle)) {
		GtkTreeIter iter;
		gtk_list_store_append(data, &iter);
		for (int i=0; i<ncol; i++) {
			if (cols[i] == G_TYPE_DOUBLE) {
				double value;
				if (cmpack_tab_gtdd(m_Handle, i, &value)==0)
					gtk_list_store_set(data, &iter, i, value, -1);
				else
					gtk_list_store_set(data, &iter, i, DBL_MAX, -1);
			} else
			if (cols[i] == G_TYPE_INT) {
				int param;
				if (cmpack_tab_gtdi(m_Handle, i, &param)==0)
					gtk_list_store_set(data, &iter, i, param, -1);
			} else
			if (cols[i] == G_TYPE_INT) {
				char *string;
				if (cmpack_tab_gtds(m_Handle, i, &string)==0) {
					gtk_list_store_set(data, &iter, i, string, -1);
					cmpack_free(string);
				}
			}
		}
		cmpack_tab_next(m_Handle);
	}

	g_slist_foreach(outcols, (GFunc)CTableWriter::free, NULL);
	g_slist_free(outcols);*/
}

// Show/hide this row from exports
void CTable::Show(bool show)
{
	if (m_Handle)
		cmpack_tab_show(m_Handle, show);
}

// Find row by primary key
bool CTable::Find(int value) const
{
	if (m_Handle) {
		if (m_PKColumn>=0) {
			cmpack_tab_rewind(m_Handle);
			return cmpack_tab_ftdi(m_Handle, m_PKColumn, value)!=0;
		} else {
			// Note: Return codes of cmpack_tab_ftdi and cmpack_tab_setpos are different!
			return cmpack_tab_setpos(m_Handle, value)==0;
		}
	}
	return false;
}

void CTable::Delete(void)
{
	if (m_Handle) 
		cmpack_tab_delete(m_Handle);
}
