/**************************************************************

selection.h (C-Munipack project)
Table of selected stars
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_TABLE_CLASSE_H
#define MUNIWIN_TABLE_CLASSE_H

#include "helper_classes.h"
#include "file_classes.h"

//
// Channel info
//
enum tChannelInfo {
	DATA_UNDEFINED,				// Nothing
	DATA_JD,					// Julian date
	DATA_MAGNITUDE,				// Magnitude
	DATA_HELCOR,				// Heliocentric correction
	DATA_AIRMASS,				// Airmass coefficient
	DATA_OFFSET,				// Pixel offset
	DATA_DEVIATION,				// Deviation
	DATA_ALTITUDE,				// Altitude
	DATA_ANY					// General purpose FP numbers
};

//
// Channel flags
//
typedef unsigned tChannelFlags;
enum tChannelFlag {
	DATA_HELIOCENTRIC = (1<<0)	// Heliocetric JD
};

//
// Channel descriptor
//
class CChannel
{
public:
	// Constructors
	CChannel(class CTable *tab, int y_col, int u_col, tChannelInfo info, tChannelFlags flags = 0);

	// Destructor
	virtual ~CChannel(void);

	// Get identifier
	int Column(void) const
	{ return m_Column; }

	// Get identifier
	int ColumnU(void) const
	{ return m_ColumnU; }

	// Is the channel valid?
	bool Valid(void) const
	{ return m_Column>=0; }

	// Channel information
	tChannelInfo Info(void) const
	{ return m_Info; }

	// Channel flags
	tChannelFlags Flags(void) const
	{ return m_Flags; }

	// Get radius in pixels
	const gchar *Name(void) const
	{ return m_Name; }

	// Get unit name (m, s, days, deg, ...)
	// Returns pointer to static buffer, you don't need to make copy
	const gchar *Unit(void) const;

	// Minimum value
	double Min(void) const;

	// Maximum value
	double Max(void) const;

	// Maximum error value
	double MaxU(void) const;

	// Enable/disable exporting of this column
	void SetExported(bool enable);

	// Is this column exported?
	bool Exported(void) const
	{ return m_Exported; }
	
	// Get default number of decimal places
	int Precision(void) const 
	{ return m_Precision; }

private:
	CTable			*m_Table;
	int				m_Column, m_ColumnU;
	char			*m_Name;
	bool			m_Exported;
	tChannelInfo	m_Info;
	tChannelFlags	m_Flags;
	CmpackType		m_DataType;
	int				m_Precision;

	// Assignment operator
	CChannel &operator=(const CChannel&);

	CChannel(const CChannel&);
};

//
// Table of channels
//
class CChannels
{
public:
	// Constructor(s)
	CChannels(void);
	
	// Destructor
	virtual ~CChannels();

	// Get number of apertures
	int Count(void) const
	{ return m_Count; }

	// Returns true if the table is empty
	bool Empty(void) const
	{ return m_Count==0; }

	// Clear the table
	void Clear(void);

	// Add a channel, if the channel with the column index is 
	// already in the table, it changes its parameters
	void Add(CChannel *item);

	// Find first channel by column index
	int FindColumn(int column) const;

	// Find first channel by name
	int FindFirst(const gchar *name) const;

	// Find first channel by given info
	int FindFirst(tChannelInfo info) const;

	// Get aperture by index
	const CChannel *Get(int index) const;
	CChannel *Get(int index);

	// Get column by index
	int GetColumn(int index) const;

	// Get column by index
	int GetColumnU(int index) const;

	// Get name by index
	const gchar *GetName(int index) const;

	// Get channel info
	tChannelInfo GetInfo(int index) const;

	// Get unit name (m, s, days, deg, ...)
	// Returns pointer to static buffer, you don't need to make copy
	const gchar *GetUnit(int index) const;

	// Minimum value
	double GetMin(int index) const;

	// Maximum value
	double GetMax(int index) const;

	// Maximum error value
	double GetMaxU(int index) const;

	// Get "exported" flag
	bool GetExported(int index) const;

	// Get default number of decimal places
	int GetPrecision(int index) const;

protected:
	int			m_Count, m_Capacity;
	CChannel	**m_List;

private:
	// Disable copy constructor and assigment operator
	CChannels &operator=(const CChannels&);
	CChannels(const CChannels&);
};

//
// Table class interface
//
class CTable
{
public:
	enum tSortCol
	{
		SORT_FRAME_ID, SORT_DATE, SORT_OBJECT_ID, SORT_MEAN_MAG, SORT_APERTURE_ID
	};

	// Default constructor
	CTable(void);

	// Constructor with initialization
	CTable(CmpackTableType type);

	// Constructor with initialization (steals the reference)
	CTable(CmpackTable *handle);

	// Destructor
	virtual ~CTable(void);

	// Makes deep copy of the table
	void MakeCopy(const CTable &table);

	// Create a new table
	void Init(CmpackTableType type);

	// Assign a table (steals the reference)
	void Assign(CmpackTable *tab);

	// Load table from a file
	bool Load(const gchar *filepath, int *result = NULL);

	// Save table to a file
	bool Save(const gchar *filepath, int *result = NULL);

	// Export table to a file
	bool SaveAs(const gchar *filepath, bool ignore_ids, bool no_header, 
		bool no_errors, int *result = NULL);

	// Export table to a file
	bool Export(const gchar *filepath, const gchar *format, int sort_column_id,
		GtkSortType sort_order, GError **error);

	// Free internal data
	void Clear(void);

	// Is table valid?
	bool Valid(void) const
	{ return m_Handle!=NULL; }

	// Get handle to the table
	CmpackTable *Handle(void) const
	{ return m_Handle; }

	// Get type identifier of the table
	CmpackTableType Type(void) const;

	// Get header field by index
	bool GetParam(int index, char **key, char **val) const;

	// Set header field
	void SetParam(const char *key, const char *value);

	// Get number of rows
	int Rows(void) const;

	// Browsing
	bool Rewind(void);
	bool Next(void);
	bool EndOfTable(void);

	// Primary key column
	int PKColumn(void);

	// Get list of dependent channels
	CChannels *ChannelsX(void);

	// Get list of independent channels
	CChannels *ChannelsY(void);

	// Get color filter name
	const char *Filter(void) const;

	// Get aperture index
	int Aperture(void) const;

	// Make graph data
	CmpackGraphData *ToGraphData(int x_channel, int y_channel, bool no_errors);

	// Make table model
	GtkTreeModel *ToTreeModel(void);

	// Set view parameters
	void SetView(CmpackGraphView *view, int x_channel, int y_channel, bool no_errors,
		const gchar *user_name_x = NULL, const gchar *user_name_y = NULL,
		tDateFormat datef = JULIAN_DATE);

	// Set view parameters
	void SetView(GtkTreeView *view);

	// Get double value from actual row
	bool GetDbl(int column, double *val) const;
	//double GetDbl(int column, double defval) const;

	// Set nul value to given column
	void SetNulValue(int column, double nulval);

	// Show/hide this row from exports
	void Show(bool show);

	// Find row by primary key
	bool Find(int pk_value) const;

	// Delete current row
	void Delete(void);

	// Invalidate cached data
	void InvalidateCache(void);

private:
	enum tCacheFlags {
		CF_CHANNELS = (1<<0)
	};
	
	CmpackTable		*m_Handle;
	unsigned		m_CacheFlags;
	CChannels		m_ChannelsX, m_ChannelsY;
	int				m_PKColumn;

	// Update list of channels
	void UpdateChannels(void);

	// Export table to file
	void ExportTable(CCSVWriter &f, int sort_column_id, GtkSortType sort_order);

private:
	// Disable copy constructor and assigment operator
	CTable(const CTable &tab);
	CTable &operator=(const CTable &orig);
};

#endif
