/**************************************************************

munilist.c (C-Munipack project)
Making output data from photometry files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "cmunipack.h"
#include "cmdline.h"
#include "cfgfile.h"
#include "utils.h"

/* Program */
AppName prog = 
{ 
	"munilist", 
	"utility for making output data from photometry files"
};

/* Options */
CmdOption opt[] =
{
	{ "diff-mag", 0, "make table of differential instrumental magnitudes (default)" },
	{ "inst-mag", 0, "make table of absolute instrumental magnitudes" },
	{ "aperture", 'a', "aperture identifier", "index", CMD_PTYPE_INT },
	{ "variable-stars", 'v', "identifier(s) of the variable star(s)", "star", CMD_PTYPE_INT_V },
	{ "comparison-stars", 'c', "indentifier(s) of the comparison star(s)", "star", CMD_PTYPE_INT_V },
	{ "check-stars", 'e', "indentifier(s) of the check star(s)", "star", CMD_PTYPE_INT_V },
	{ "configuration-file", 'p', "read parameters from configuration file", "filepath", CMD_PTYPE_FILEPATH },
	{ "set", 's', "set configuration parameter", "name=value", CMD_PTYPE_PARAMLIST },
	{ "read-dirfile", 'i', "read list of source photometry files from specified file", "filepath", CMD_PTYPE_FILEPATH },
	{ "help", 'h', "print list of command-line parameters" },
	{ "quiet", 'q', "quiet mode; suppress all messages" },
	{ "version", 0, "print software version" },
	{ "licence", 0, "print software licence" },
	{ "verbose", 0, "verbose mode; print debug messages" },
	{ NULL }
};

/* Configuration parameters */
CfgParam cfg[] = 
{
	{ "aperture", CFG_PTYPE_INT, "aperture identifier" },
	{ "var", CFG_PTYPE_INT_V, "identifier(s) of the variable star(s)" },
	{ "comp", CFG_PTYPE_INT_V, "indentifier(s) of the comparison star(s)" },
	{ "check", CFG_PTYPE_INT_V, "indentifier(s) of the check star(s)" },
	{ NULL }
};

/* Input files */
CmdParam flist[] = 
{
	{ "output-file", "output file name", CMD_PTYPE_FILEPATH },
	{ "files", "list of input photometry files", CMD_PTYPE_FILELIST },
	{ NULL }
};

/* Command-line definition */
AppInfo cmd = 
{ 
	&prog, opt, flist, cfg
};

static void error(int status, const char *file)
{
	PrintError(&cmd, file, 0, status);
	exit(status);
}

static int list(CmpackFrameSet *fset, const char *srcpath, int frame_id)
{
	int res;
	CmpackPhtFile *pht;

	printf("%s\n", srcpath);

	res = cmpack_pht_open(&pht, srcpath, CMPACK_OPEN_READONLY, 0);
	if (res!=0) {
		error(res, srcpath);
		return res;
	}
	res = cmpack_fset_append_frame(fset, pht, frame_id, srcpath);
	if (res!=0) {
		error(res, srcpath);
		cmpack_pht_destroy(pht);
		return res;
	}
	cmpack_pht_destroy(pht);
	return 0;
}

int main(int argc, char **argv)
{
	FILE	*din = NULL;
	int		i, res, instmag, frame_id = 0, aperture, *intv, in_files = 0, count;
	int		ap_valid = 0;
	char 	line[MAXLINE], buf1[MAXLINE];
	char	*dirfile_in, *output, *cfgfile, **files;
	CmpackLCurve *lc;
	CmpackFrameSet *fset;
	CmpackConsole *con;
	CmpackTable *tab;
	CmpackPhtAperture apinfo;
	CmpackCatObject stinfo;

	Init(&cmd);

	tab = NULL;

	/* Parse command line */
	if (!cmdline_parse(argc, argv, &cmd))
		return CMPACK_ERR_CMDLINE_ERROR;

	/* Process standard options */
	if (cmdline_isdef(&cmd, "version")) {
		PrintVersion(&cmd);
		return 0;
	}
	if (cmdline_isdef(&cmd, "licence")) {
		PrintLicense(&cmd);
		return 0;
	}
	if (cmdline_isdef(&cmd, "help")) {
		PrintHelp(&cmd);
		return 0;
	}

	/* Print program info */
	PrintPrologue(&cmd);

	fset = cmpack_fset_init();
	if (!fset) error(CMPACK_ERR_MEMORY, NULL);
	lc = cmpack_lcurve_init();
	if (!lc) error(CMPACK_ERR_MEMORY, NULL);
	
	/* Set output format */
	instmag = 0;
	if (cmdline_isdef(&cmd, "inst-mag"))
		instmag = 1;

	if (!cmdline_getpars(&cmd, "output-file", &output))
		error(CMPACK_ERR_NO_OUTPUT_FILE, NULL);
	if (!CheckFilePath(output))
		error(CMPACK_ERR_INVALID_OUT_FILE, NULL);

	/* Set callback function and output level */
	con = cmpack_con_init();
	if (cmdline_isdef(&cmd, "quiet")) 
		cmpack_con_set_level(con, CMPACK_LEVEL_QUIET);
	if (cmdline_isdef(&cmd, "verbose"))
		cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
	cmpack_lcurve_set_console(lc, con);

	/* Read configuration file */
	if (cmdline_gets(&cmd, "configuration-file", &cfgfile)) {
		res = cfgfile_read(&cmd, cfgfile);
		if (res!=0) error(res, NULL);
	}

	/* Set configuration paremeters */
	if (cfgfile_geti(&cmd, "aperture", &aperture)) {
		cmpack_lcurve_set_aperture(lc, aperture);
		apinfo.id = aperture;
		apinfo.radius = 0;
		cmpack_fset_add_aperture(fset, CMPACK_PA_ID, &apinfo);
		ap_valid = 1;
	}
	if (cfgfile_getiv(&cmd, "variable-stars", &intv, &count)) {
		cmpack_lcurve_set_var(lc, intv, count);
		for (i=0; i<count; i++) {
			stinfo.id = intv[i];
			cmpack_fset_add_object(fset, CMPACK_OM_ID, &stinfo);
		}
	}
	if (cfgfile_getiv(&cmd, "comparison-stars", &intv, &count)) {
		cmpack_lcurve_set_comp(lc, intv, count);
		for (i=0; i<count; i++) {
			stinfo.id = intv[i];
			cmpack_fset_add_object(fset, CMPACK_OM_ID, &stinfo);
		}
	}
	if (cfgfile_getiv(&cmd, "check-stars", &intv, &count)) {
		cmpack_lcurve_set_check(lc, intv, count);
		for (i=0; i<count; i++) {
			stinfo.id = intv[i];
			cmpack_fset_add_object(fset, CMPACK_OM_ID, &stinfo);
		}
	}
	
	/* Aperture selection */
	if (cmdline_geti(&cmd, "aperture", &aperture)) {
		cmpack_lcurve_set_aperture(lc, aperture);
		apinfo.id = aperture;
		apinfo.radius = 0;
		cmpack_fset_add_aperture(fset, CMPACK_PA_ID, &apinfo);
		ap_valid = 1;
	}

	/* Set selection */
	if (cmdline_getiv(&cmd, "variable-stars", &intv, &count)) {
		cmpack_lcurve_set_var(lc, intv, count);
			for (i=0; i<count; i++) {
			stinfo.id = intv[i];
			cmpack_fset_add_object(fset, CMPACK_OM_ID, &stinfo);
		}
	}
	if (cmdline_getiv(&cmd, "comparison-stars",  &intv, &count)) {
		cmpack_lcurve_set_comp(lc, intv, count);
		for (i=0; i<count; i++) {
			stinfo.id = intv[i];
			cmpack_fset_add_object(fset, CMPACK_OM_ID, &stinfo);
		}
	}
	if (cmdline_getiv(&cmd, "check-stars", &intv, &count)) {
		cmpack_lcurve_set_check(lc, intv, count);
		for (i=0; i<count; i++) {
			stinfo.id = intv[i];
			cmpack_fset_add_object(fset, CMPACK_OM_ID, &stinfo);
		}
	}

	if (!ap_valid) {
		fprintf(stderr, "Missing aperture selection.");
		return CMPACK_ERR_AP_NOT_FOUND;
	}
	
	/* Create a frame set */
	if (cmdline_gets(&cmd, "read-dirfile", &dirfile_in)) {
		/* Image names are read from a file */
		if (strcmp(dirfile_in, "@")==0) {
			din = stdin;
		} else {
			din = fopen(dirfile_in, "r");
			if (din==NULL) error(CMPACK_ERR_CANT_OPEN_DIRFILE, dirfile_in);
		}
		while (fgets(buf1, MAXLINE, din)) {
			sscanf(buf1, "%1023s", line);
			if (CheckFilePath(line)) {
				in_files++;
				list(fset, line, frame_id++);
			}
		}
		if (din && din!=stdin)
			fclose(din);
	} else {
		/* images names are given through the command line */
		if (cmdline_getparsv(&cmd, "files", &files, &count)) {
			for (i=0; i<count; i++) {
				if (CheckFilePath(files[i])) {
#ifdef _WIN32
					struct _finddata_t fd;
					intptr_t fh = _findfirst(files[i], &fd);
					if (fh!=-1) {
						do {
							if ((fd.attrib & _A_SUBDIR)==0)  {
								in_files++;
								MakeInputName(line, MAXLINE, files[i], fd.name);
								list(fset, line, frame_id++);
							}
						} while (_findnext(fh, &fd)==0);
						_findclose(fh);
					}
#else
					in_files++;
					list(fset, files[i], frame_id++);
#endif
				}
			}
		}
	}

	if (instmag) 
		res = cmpack_lcurve(lc, fset, &tab, CMPACK_LCURVE_INSTMAG);
	else
		res = cmpack_lcurve(lc, fset, &tab, CMPACK_LCURVE_DEFAULT);
	if (res!=0) error(res, NULL);

	/* Write table to file */
	res = cmpack_tab_save(tab, output, 0, NULL, 0);
	if (res!=0) error(res, output);

	PrintEpilogue(&cmd, in_files);

	cmpack_tab_destroy(tab);
	cmpack_fset_destroy(fset);
	cmpack_lcurve_destroy(lc);
	cmpack_con_destroy(con);
	Clean(&cmd);

	return 0;
}
